Type: Package
Package: rprimer
Title: Design Degenerate Oligos from a Multiple DNA Sequence Alignment
Version: 1.17.1
Description: 
    Functions, workflow, and a Shiny application for visualizing sequence 
    conservation and designing degenerate primers, probes, and (RT)-(q/d)PCR 
    assays from a multiple DNA sequence alignment. The results can be presented 
    in data frame format and visualized as dashboard-like plots. For more information, 
    please see the package vignette.
Authors@R: 
    person(given = "Sofia",
           family = "Persson",
           role = c("aut", "cre"),
           email = "sofiapersson27@gmail.com",
           comment = c(ORCID = "0000-0003-2611-3030"))
License: GPL-3
URL: https://github.com/sofpn/rprimer
BugReports: https://github.com/sofpn/rprimer/issues
Depends: R (>= 4.1), Biostrings, MultipleAlignment,
Imports: bslib, DT, ggplot2, IRanges, mathjaxr, methods, patchwork,
        reshape2, S4Vectors, shiny, shinycssloaders, shinyFeedback
Suggests: BiocStyle, covr, kableExtra, knitr, rmarkdown, styler,
        testthat (>= 3.0.0)
VignetteBuilder: knitr
Encoding: UTF-8
RdMacros: mathjaxr
biocViews: Alignment, ddPCR, Coverage, MultipleSequenceAlignment,
        SequenceMatching, qPCR
biocType: Software
Config/testthat/edition: 3
Config/roxygen2/version: 8.0.0
git_url: https://git.bioconductor.org/packages/rprimer
git_branch: devel
git_last_commit: 32bb697
git_last_commit_date: 2026-07-29
Repository: Bioconductor 3.24
Date/Publication: 2026-07-30
NeedsCompilation: no
Packaged: 2026-07-31 00:01:25 UTC; biocbuild
Author: Sofia Persson [aut, cre] (ORCID:
    <https://orcid.org/0000-0003-2611-3030>)
Maintainer: Sofia Persson <sofiapersson27@gmail.com>

# Generated by roxygen2: do not edit by hand

export(RprimerAssay)
export(RprimerMatchAssay)
export(RprimerMatchOligo)
export(RprimerOligo)
export(RprimerProfile)
export(checkMatch)
export(consensusProfile)
export(designAssays)
export(designOligos)
export(plotData)
export(runRprimerApp)
exportMethods(checkMatch)
exportMethods(plotData)
import(mathjaxr)
import(methods)
import(shiny)
importClassesFrom(Biostrings,DNAStringSet)
importClassesFrom(S4Vectors,DataFrame)
importFrom(MultipleAlignment,colmask)
importFrom(MultipleAlignment,readDNAMultipleAlignment)
importFrom(MultipleAlignment,rowmask)
importFrom(patchwork,wrap_plots)

<!-- README.md is generated from README.Rmd. Please edit that file -->

# rprimer <img src='man/figures/rprimer.png' align="right" height="139" />

<!-- badges: start -->

[![R-CMD-check](https://github.com/sofpn/rprimer/workflows/R-CMD-check/badge.svg)](https://github.com/sofpn/rprimer/actions)
<!-- badges: end -->

rprimer is an R package that designs degenerate oligos and PCR assays
from a multiple DNA sequence alignment of target sequences of interest.
The package is specifically designed for sequence variable viruses.

## Installation

To install rprimer from
[Bioconductor](https://bioconductor.org/packages/devel/bioc/html/rprimer.html),
start R (version 4.2 or later) and enter:

``` r
if (!requireNamespace("BiocManager", quietly = TRUE))
  install.packages("BiocManager")
BiocManager::install("rprimer")
```

Attach the package by calling:

``` r
library(rprimer)
```

## Overview

The package contains five main functions:

- `consensusProfile()`
- `designOligos()`
- `designAssays()`
- `checkMatch()`
- `plotData()`

## Shiny application

The package can be run through a Shiny application (a graphical user
interface). To start the application, type `runRprimerApp()` from within
R upon installing and attaching the package.

The application can also be found online,
[here](https://sofpn.shinyapps.io/rprimer).

## Workflow

### Import alignment

The first step is to import an alignment with target sequences of
interest. This is done by using `readDNAMultipleAlignment()`.

The file “example_alignment.txt” contains an alignment of 50 hepatitis E
virus sequences.

``` r
infile <- system.file("extdata", "example_alignment.txt", package = "rprimer")

myAlignment <- readDNAMultipleAlignment(infile, format = "fasta")
```

### Step 1: `consensusProfile`

`consensusProfile()` takes a `DNAMultipleAlignment` as input and returns
all the information needed for the subsequent design process.

``` r
myConsensusProfile <- consensusProfile(myAlignment, ambiguityThreshold = 0.05)
```

Results (row 100-106):

| position |    a |    c |    g |    t | other | gaps | majority | identity | iupac | coverage |
|---------:|-----:|-----:|-----:|-----:|------:|-----:|:---------|---------:|:------|---------:|
|      100 | 0.00 | 1.00 | 0.00 | 0.00 |  0.00 |    0 | C        |     1.00 | C     |     1.00 |
|      101 | 1.00 | 0.00 | 0.00 | 0.00 |  0.00 |    0 | A        |     1.00 | A     |     1.00 |
|      102 | 0.16 | 0.00 | 0.84 | 0.00 |  0.00 |    0 | G        |     0.84 | R     |     1.00 |
|      103 | 0.00 | 0.00 | 1.00 | 0.00 |  0.00 |    0 | G        |     1.00 | G     |     1.00 |
|      104 | 0.00 | 0.98 | 0.00 | 0.00 |  0.02 |    0 | C        |     1.00 | C     |     1.00 |
|      105 | 0.20 | 0.00 | 0.02 | 0.78 |  0.00 |    0 | T        |     0.78 | W     |     0.98 |
|      106 | 0.00 | 0.00 | 1.00 | 0.00 |  0.00 |    0 | G        |     1.00 | G     |     1.00 |

The results can be visualized with `plotData()`:

``` r
plotData(myConsensusProfile)
#> Warning in ggplot2::geom_segment(color = "#93A8AC", ggplot2::aes(x = min(position), : All aesthetics have length 1, but the data has 7597 rows.
#> ℹ Please consider using `annotate()` or provide this layer with data containing
#>   a single row.
```

<img src="man/figures/README-unnamed-chunk-6-1.png" width="100%" />

### Step 2: `designOligos`

The next step is to design oligos. You can either use the default
settings as below, or adjust them as preferred (see the package vignette
or `?designOligos` for more information). The default settings allow a
maximum degeneracy of four, which means that only the most conserved
regions of the genome will be considered as oligo binding sites.

``` r
myOligos <- designOligos(myConsensusProfile)
```

Results (first six rows):

| type   | fwd   | rev   | start | end | length | iupacSequence        | iupacSequenceRc      | identity | coverage | degeneracy | gcContentMean | gcContentRange | tmMean | tmRange | deltaGMean | deltaGRange | sequence   | sequenceRc | gcContent  | tm         | deltaG     | method    | score | roiStart | roiEnd |
|:-------|:------|:------|------:|----:|-------:|:---------------------|:---------------------|---------:|---------:|-----------:|--------------:|---------------:|-------:|--------:|-----------:|------------:|:-----------|:-----------|:-----------|:-----------|:-----------|:----------|------:|---------:|-------:|
| probe  | TRUE  | TRUE  |   124 | 143 |     20 | TCYGCCYTGGCGAATGCTGT | ACAGCATTCGCCARGGCRGA |     0.95 |     0.99 |          4 |          0.60 |           0.10 |  63.17 |    4.33 |     -21.61 |        2.08 | TCCGCCCT…. | ACAGCATT…. | 0.65, 0….. | 65.33623…. | -22.6538…. | ambiguous |     2 |        1 |   7597 |
| probe  | FALSE | TRUE  |   127 | 146 |     20 | GCCYTGGCGAATGCTGTGGT | ACCACAGCATTCGCCARGGC |     0.98 |     0.99 |          2 |          0.62 |           0.05 |  63.18 |    1.84 |     -21.73 |        0.83 | GCCCTGGC…. | ACCACAGC…. | 0.65, 0.6  | 64.10586…. | -22.1475…. | ambiguous |     3 |        1 |   7597 |
| primer | TRUE  | FALSE |   128 | 146 |     19 | CCYTGGCGAATGCTGTGGT  | ACCACAGCATTCGCCARGG  |     0.97 |     0.99 |          2 |          0.61 |           0.05 |  61.48 |    1.95 |     -19.84 |        0.83 | CCCTGGCG…. | ACCACAGC…. | 0.631578…. | 62.45335…. | -20.2570…. | ambiguous |     3 |        1 |   7597 |
| primer | TRUE  | FALSE |   128 | 147 |     20 | CCYTGGCGAATGCTGTGGTR | YACCACAGCATTCGCCARGG |     0.96 |     0.99 |          4 |          0.60 |           0.10 |  61.61 |    3.37 |     -20.55 |        1.76 | CCCTGGCG…. | TACCACAG…. | 0.6, 0.5…. | 61.77166…. | -20.5089…. | ambiguous |     2 |        1 |   7597 |
| probe  | TRUE  | TRUE  |   128 | 146 |     19 | CCYTGGCGAATGCTGTGGT  | ACCACAGCATTCGCCARGG  |     0.97 |     0.99 |          2 |          0.61 |           0.05 |  60.45 |    1.94 |     -19.84 |        0.83 | CCCTGGCG…. | ACCACAGC…. | 0.631578…. | 61.41686…. | -20.2570…. | ambiguous |     3 |        1 |   7597 |
| probe  | TRUE  | TRUE  |   128 | 147 |     20 | CCYTGGCGAATGCTGTGGTR | YACCACAGCATTCGCCARGG |     0.96 |     0.99 |          4 |          0.60 |           0.10 |  60.63 |    3.37 |     -20.55 |        1.76 | CCCTGGCG…. | TACCACAG…. | 0.6, 0.5…. | 60.78676…. | -20.5089…. | ambiguous |     2 |        1 |   7597 |

The results can be visualized as a dashboard, using `plotData()`:

``` r
plotData(myOligos)
```

<img src="man/figures/README-unnamed-chunk-9-1.png" width="100%" />

### Step 3: `designAssays`

`designAssays()` finds pairs of forward and reverse primers and combine
them with probes, if probes are present in the input dataset. You can
either use the default settings as below, or adjust the design
constraints (see the package vignette or `?designAssays` for more
information).

``` r
myAssays <- designAssays(myOligos)
```

Results (first six rows):

| start |  end | length | totalDegeneracy | score | startFwd | endFwd | lengthFwd | iupacSequenceFwd     | identityFwd | coverageFwd | degeneracyFwd | gcContentMeanFwd | gcContentRangeFwd | tmMeanFwd | tmRangeFwd | deltaGMeanFwd | deltaGRangeFwd | sequenceFwd | gcContentFwd | tmFwd      | deltaGFwd  | methodFwd | startRev | endRev | lengthRev | iupacSequenceRev     | identityRev | coverageRev | degeneracyRev | gcContentMeanRev | gcContentRangeRev | tmMeanRev | tmRangeRev | deltaGMeanRev | deltaGRangeRev | sequenceRev | gcContentRev | tmRev      | deltaGRev  | methodRev | plusPr | minusPr | startPr | endPr | lengthPr | iupacSequencePr        | iupacSequenceRcPr      | identityPr | coveragePr | degeneracyPr | gcContentMeanPr | gcContentRangePr | tmMeanPr | tmRangePr | deltaGMeanPr | deltaGRangePr | sequencePr | sequenceRcPr | gcContentPr | tmPr       | deltaGPr   | methodPr  | roiStart | roiEnd |
|------:|-----:|-------:|----------------:|------:|---------:|-------:|----------:|:---------------------|------------:|------------:|--------------:|-----------------:|------------------:|----------:|-----------:|--------------:|---------------:|:------------|:-------------|:-----------|:-----------|:----------|---------:|-------:|----------:|:---------------------|------------:|------------:|--------------:|-----------------:|------------------:|----------:|-----------:|--------------:|---------------:|:------------|:-------------|:-----------|:-----------|:----------|:-------|:--------|--------:|------:|---------:|:-----------------------|:-----------------------|-----------:|-----------:|-------------:|----------------:|-----------------:|---------:|----------:|-------------:|--------------:|:-----------|:-------------|:------------|:-----------|:-----------|:----------|---------:|-------:|
|  5605 | 5673 |     69 |               6 |  2.00 |     5605 |   5624 |        20 | GGCRGTGGTTTCTGGGGTGA |        0.98 |           1 |             2 |             0.62 |              0.05 |     62.84 |       2.51 |        -20.86 |           1.25 | GGCAGTGG….  | 0.6, 0.65    | 61.57995…. | -20.2350…. | ambiguous |     5654 |   5673 |        20 | GTTGGTTGGATGAASATAGG |           1 |           1 |             2 |              0.4 |                 0 |     50.71 |        1.1 |        -15.27 |           0.52 | GTTGGTTG….  | 0.4, 0.4     | 50.15469…. | -15.0078…. | ambiguous | TRUE   | FALSE   |    5625 |  5642 |       18 | CMGGGTTGATTCTCAGCC     | GGCTGAGAATCAACCCKG     |       0.97 |       0.99 |            2 |            0.58 |             0.06 |    55.14 |      2.79 |       -17.06 |          1.25 | CAGGGTTG…. | GGCTGAGA….   | 0.555555….  | 53.74554…. | -16.4324…. | ambiguous |        1 |   7597 |
|  5605 | 5673 |     69 |               6 |  2.33 |     5605 |   5624 |        20 | GGCRGTGGTTTCTGGGGTGA |        0.98 |           1 |             2 |             0.62 |              0.05 |     62.84 |       2.51 |        -20.86 |           1.25 | GGCAGTGG….  | 0.6, 0.65    | 61.57995…. | -20.2350…. | ambiguous |     5654 |   5673 |        20 | GTTGGTTGGATGAASATAGG |           1 |           1 |             2 |              0.4 |                 0 |     50.71 |        1.1 |        -15.27 |           0.52 | GTTGGTTG….  | 0.4, 0.4     | 50.15469…. | -15.0078…. | ambiguous | TRUE   | FALSE   |    5625 |  5643 |       19 | CMGGGTTGATTCTCAGCCC    | GGGCTGAGAATCAACCCKG    |       0.97 |       0.99 |            2 |            0.61 |             0.05 |    57.63 |      2.64 |       -18.54 |          1.25 | CAGGGTTG…. | GGGCTGAG….   | 0.578947….  | 56.30713…. | -17.9185…. | ambiguous |        1 |   7597 |
|  5605 | 5673 |     69 |               6 |  2.00 |     5605 |   5624 |        20 | GGCRGTGGTTTCTGGGGTGA |        0.98 |           1 |             2 |             0.62 |              0.05 |     62.84 |       2.51 |        -20.86 |           1.25 | GGCAGTGG….  | 0.6, 0.65    | 61.57995…. | -20.2350…. | ambiguous |     5654 |   5673 |        20 | GTTGGTTGGATGAASATAGG |           1 |           1 |             2 |              0.4 |                 0 |     50.71 |        1.1 |        -15.27 |           0.52 | GTTGGTTG….  | 0.4, 0.4     | 50.15469…. | -15.0078…. | ambiguous | TRUE   | TRUE    |    5625 |  5644 |       20 | CMGGGTTGATTCTCAGCCCT   | AGGGCTGAGAATCAACCCKG   |       0.97 |       1.00 |            2 |            0.58 |             0.05 |    58.87 |      2.55 |       -19.43 |          1.25 | CAGGGTTG…. | AGGGCTGA….   | 0.55, 0.6   | 57.59836…. | -18.8035…. | ambiguous |        1 |   7597 |
|  5605 | 5673 |     69 |               6 |  1.67 |     5605 |   5624 |        20 | GGCRGTGGTTTCTGGGGTGA |        0.98 |           1 |             2 |             0.62 |              0.05 |     62.84 |       2.51 |        -20.86 |           1.25 | GGCAGTGG….  | 0.6, 0.65    | 61.57995…. | -20.2350…. | ambiguous |     5654 |   5673 |        20 | GTTGGTTGGATGAASATAGG |           1 |           1 |             2 |              0.4 |                 0 |     50.71 |        1.1 |        -15.27 |           0.52 | GTTGGTTG….  | 0.4, 0.4     | 50.15469…. | -15.0078…. | ambiguous | TRUE   | TRUE    |    5625 |  5645 |       21 | CMGGGTTGATTCTCAGCCCTT  | AAGGGCTGAGAATCAACCCKG  |       0.98 |       1.00 |            2 |            0.55 |             0.05 |    59.21 |      2.43 |       -20.08 |          1.25 | CAGGGTTG…. | AAGGGCTG….   | 0.523809….  | 57.99472…. | -19.4553…. | ambiguous |        1 |   7597 |
|  5605 | 5673 |     69 |               6 |  2.00 |     5605 |   5624 |        20 | GGCRGTGGTTTCTGGGGTGA |        0.98 |           1 |             2 |             0.62 |              0.05 |     62.84 |       2.51 |        -20.86 |           1.25 | GGCAGTGG….  | 0.6, 0.65    | 61.57995…. | -20.2350…. | ambiguous |     5654 |   5673 |        20 | GTTGGTTGGATGAASATAGG |           1 |           1 |             2 |              0.4 |                 0 |     50.71 |        1.1 |        -15.27 |           0.52 | GTTGGTTG….  | 0.4, 0.4     | 50.15469…. | -15.0078…. | ambiguous | TRUE   | FALSE   |    5625 |  5646 |       22 | CMGGGTTGATTCTCAGCCCTTC | GAAGGGCTGAGAATCAACCCKG |       0.98 |       0.99 |            2 |            0.57 |             0.05 |    59.91 |      2.28 |       -21.11 |          1.25 | CAGGGTTG…. | GAAGGGCT….   | 0.545454….  | 58.77533…. | -20.4881…. | ambiguous |        1 |   7597 |
|  5605 | 5673 |     69 |               6 |  2.00 |     5605 |   5624 |        20 | GGCRGTGGTTTCTGGGGTGA |        0.98 |           1 |             2 |             0.62 |              0.05 |     62.84 |       2.51 |        -20.86 |           1.25 | GGCAGTGG….  | 0.6, 0.65    | 61.57995…. | -20.2350…. | ambiguous |     5654 |   5673 |        20 | GTTGGTTGGATGAASATAGG |           1 |           1 |             2 |              0.4 |                 0 |     50.71 |        1.1 |        -15.27 |           0.52 | GTTGGTTG….  | 0.4, 0.4     | 50.15469…. | -15.0078…. | ambiguous | TRUE   | FALSE   |    5626 |  5643 |       18 | MGGGTTGATTCTCAGCCC     | GGGCTGAGAATCAACCCK     |       0.97 |       0.99 |            2 |            0.58 |             0.06 |    55.71 |      1.65 |       -17.21 |          0.94 | AGGGTTGA…. | GGGCTGAG….   | 0.555555….  | 54.88806…. | -16.7409…. | ambiguous |        1 |   7597 |

The assays can be visualized using `plotData()`:

``` r
plotData(myAssays)
```

<img src="man/figures/README-unnamed-chunk-12-1.png" width="100%" />

### Additional step: `checkMatch`

`checkMatch()` shows the proportion and names of the target sequences in
the input alignment that match with the generated oligos or assays. See
the package vignette or `?checkMatch` for more information.

``` r
## Randomly select six oligos to illustrate an example 
selection <- sample(seq_len(nrow(myOligos)), size = 6)

matchTableOligos <- checkMatch(myOligos[selection, ], target = myAlignment)
```

Results:

| iupacSequence          | perfectMatch | idPerfectMatch | oneMismatch | idOneMismatch | twoMismatches | idTwoMismatches | threeMismatches | idThreeMismatches | fourOrMoreMismatches | idFourOrMoreMismatches | offTargetMatch | idOffTargetMatch |
|:-----------------------|-------------:|:---------------|------------:|:--------------|--------------:|:----------------|----------------:|:------------------|---------------------:|:-----------------------|---------------:|:-----------------|
| ACMGGGTTGATTCTCAGCCCTT |         0.90 | AB073912….     |        0.10 | AB481228….    |          0.00 |                 |               0 |                   |                    0 |                        |              0 |                  |
| TATWTTCATCCAACCAACCCC  |         0.98 | AB073912….     |        0.02 | HM439284.1    |          0.00 |                 |               0 |                   |                    0 |                        |              0 |                  |
| CRGTGGTTTCTGGGGTGAC    |         0.96 | AB073912….     |        0.04 | BD378055….    |          0.00 |                 |               0 |                   |                    0 |                        |              0 |                  |
| TTCATCCAACCAACCCCT     |         0.98 | AB073912….     |        0.02 | HM439284.1    |          0.00 |                 |               0 |                   |                    0 |                        |              0 |                  |
| GGTGACMGGGTTGATTCT     |         0.90 | AB073912….     |        0.08 | BD378055….    |          0.02 | JQ953665.1      |               0 |                   |                    0 |                        |              0 |                  |
| TGGTTTCTGGGGTGACMG     |         0.96 | AB073912….     |        0.04 | BD378055….    |          0.00 |                 |               0 |                   |                    0 |                        |              0 |                  |

The match table can be visualized using `plotData()`:

``` r
plotData(matchTableOligos)
```

<img src="man/figures/README-unnamed-chunk-15-1.png" width="100%" />

## More information

Please see the [package
vignette](https://bioconductor.org/packages/devel/bioc/vignettes/rprimer/inst/doc/getting-started-with-rprimer.html)
for more information on how to use the package.

## Citation

Persson S., Larsson C., Simonsson M., Ellström P. (2022) rprimer: an
R/bioconductor package for design of degenerate oligos for sequence
variable viruses. [*BMC Bioinformatics*
23:239](https://bmcbioinformatics.biomedcentral.com/articles/10.1186/s12859-022-04781-0)

## ----include = FALSE----------------------------------------------------------
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>"
)

## ----internal setup, echo=FALSE, warning=FALSE, message=FALSE-----------------
library(kableExtra)

## ----warning=FALSE, message=FALSE---------------------------------------------
library(rprimer)
library(Biostrings)

## -----------------------------------------------------------------------------
filepath <- system.file("extdata", "example_alignment.txt", package = "rprimer")

myAlignment <- readDNAMultipleAlignment(filepath, format = "fasta")

## -----------------------------------------------------------------------------
## Mask everything but position 3000 to 4000 and 5000 to 6000
myMaskedAlignment <- myAlignment

colmask(myMaskedAlignment, invert = TRUE) <- c(3000:4000, 5000:6000)

## -----------------------------------------------------------------------------
myConsensusProfile <- consensusProfile(myAlignment, ambiguityThreshold = 0.05)

## ----echo=FALSE---------------------------------------------------------------
myConsensusProfile[100:105, ] %>%
  kbl(., digits = 2) %>%
  kable_material(c("striped", "hover"), position = "right") %>%
  scroll_box(width = "650px")

## ----fig.width=12, fig.height=6-----------------------------------------------
plotData(myConsensusProfile)

## ----fig.width=12, fig.height=6-----------------------------------------------
## Select position 5000 to 5800 in the consensus profile 
selection <- myConsensusProfile[
  myConsensusProfile$position >= 5000 & myConsensusProfile$position <= 5800, 
  ]

plotData(selection)

## -----------------------------------------------------------------------------
myOligos <- designOligos(myConsensusProfile)

## ----echo=FALSE---------------------------------------------------------------
myOligos[1:5, ] %>%
  kbl(., digits = 2) %>%
  kable_material(c("striped", "hover"), position = "right") %>%
  scroll_box(width = "650px")

## -----------------------------------------------------------------------------
myOligos$sequence[[1]] ## All sequence variants of the first oligo (i.e., first row) 
myOligos$gcContent[[1]] ## GC-content of all variants of the first oligo 
myOligos$tm[[1]] ## Tm of all variants of the first oligo 

## ----fig.align="center", fig.width=12, fig.height=8---------------------------
plotData(myOligos)

## -----------------------------------------------------------------------------
## I first need to make a consensus profile of the masked alignment 
myMaskedConsensusProfile <- consensusProfile(myMaskedAlignment, ambiguityThreshold = 0.05)

myMaskedMixedPrimers <- designOligos(myMaskedConsensusProfile,
                                     maxDegeneracyPrimer = 8,
                                     lengthPrimer = c(24:28),
                                     tmPrimer = c(65,70),
                                     designStrategyPrimer = "mixed", 
                                     probe = FALSE)

## ----fig.align="center", fig.width=12, fig.height=8, warn=FALSE---------------
plotData(myMaskedMixedPrimers)

## ----fig.align="center", fig.width=12, fig.height=8---------------------------
## Get the minimum score from myOligos 
bestOligoScore <- min(myOligos$score)
bestOligoScore

## Make a subset that only oligos with the best score are included 
oligoSelection <- myOligos[myOligos$score == bestOligoScore, ]

## -----------------------------------------------------------------------------
## Get the binding region of the first oligo in the selection above (first row): 
bindingRegion <- myConsensusProfile[
  myConsensusProfile$position >= oligoSelection[1, ]$start & 
    myConsensusProfile$position <= oligoSelection[1, ]$end,
  ]

## -----------------------------------------------------------------------------
plotData(bindingRegion, type = "nucleotide")

## -----------------------------------------------------------------------------
myAssays <- designAssays(myOligos)  

## ----echo=FALSE---------------------------------------------------------------
myAssays[1:5, ] %>%
  kbl(., digits = 2) %>%
  kable_material(c("striped", "hover"), position = "right") %>%
  scroll_box(width = "650px")

## ----fig.width=12, fig.height=6-----------------------------------------------
plotData(myAssays)

## -----------------------------------------------------------------------------
## Get the minimum (best) score from myAssays
bestAssayScore <- min(myAssays$score)
bestAssayScore

## Make a subset that only contains the assays with the best score 
myAssaySelection <- myAssays[myAssays$score == bestAssayScore, ]

## -----------------------------------------------------------------------------
from <- myAssaySelection[1, ]$start
to <- myAssaySelection[1, ]$end

## ----fig 3, fig.width=12, fig.height=6----------------------------------------
plotData(myConsensusProfile, highlight = c(from, to))

## -----------------------------------------------------------------------------
myAssayRegion <- myConsensusProfile[
  myConsensusProfile$position >= from & 
    myConsensusProfile$position <= to, 
] 

## ----fig.width=12, fig.height=6, fig.align="center"---------------------------
plotData(myAssayRegion, type = "nucleotide")

## -----------------------------------------------------------------------------
paste(myAssayRegion$iupac, collapse = "")

## -----------------------------------------------------------------------------
## Check the first three candidates in the oligoSelection dataset
oligoSelectionMatch <- checkMatch(oligoSelection[1:3, ], target = myAlignment)

## ----echo=FALSE---------------------------------------------------------------
oligoSelectionMatch %>%
  kbl(., digits = 2) %>%
  kable_material(c("striped", "hover"), position = "right") %>%
  scroll_box(width = "650px")

## -----------------------------------------------------------------------------
## Get the id of all sequences that has one mismatch to the first oligo in the input dataset
oligoSelectionMatch$idOneMismatch[[1]] 

## ----fig.width=12, fig.height=8-----------------------------------------------
plotData(oligoSelectionMatch)

## -----------------------------------------------------------------------------
## Convert the first two oligos 
as(myOligos[1:2, ], "DNAStringSet")

## -----------------------------------------------------------------------------
## Convert the first two assays 
as(myAssays[1:2, ], "DNAStringSet")

## ----warning=FALSE------------------------------------------------------------
data("exampleRprimerProfile")
data("exampleRprimerOligo")
data("exampleRprimerAssay")
data("exampleRprimerMatchOligo")
data("exampleRprimerMatchAssay")

## -----------------------------------------------------------------------------
sessionInfo()

## Import all genotyping results
results <- read.table("typing_results.txt", sep = "\t", header = TRUE)

## Identify all Ortohepevirus A sequences
valid <- results$BLAST.result == "Hepeviridae Orthohepevirus A"
hev <- results$name[valid]

## Randomly select 50 sequences to include as example data in the package
set.seed(1)
selection <- sample(hev, 50)

## Save accession numbers for the selected sequences so that they can be
## downloaded using batch entrez
write.table(
    selection, file = "accession_valid_sequences.txt", quote = FALSE,
    row.names = FALSE, col.names = FALSE, sep = "\t"
)

#' Plot an Rprimer object
#'
#' \code{plotData} visualizes objects from all different
#' Rprimer classes.
#'
#' @param x
#' An \code{RprimerProfile}, \code{RprimerOligo} \code{RprimerAssay},
#' \code{rprimerMatchOligo} or \code{RprimerMatchAssay} object.
#'
#' @param ...
#' Optional arguments for \code{RprimerProfile} objects.
#'
#' @param type
#' For \code{Rprimeroligo} objects:
#' Type of plot: \code{"overview"}, or
#' \code{"nucleotide"}, defaults to \code{"overview"}.
#'
#' @param highlight
#' For \code{Rprimeroligo} objects:
#' If a specific region within an overview plot should be highlighted.
#' A numeric vector indicating the start and end position,
#' e.g. \code{c(100, 1000)}, defaults to \code{NULL}
#' (i.e., no highlight).
#'
#' @param rc
#' For \code{Rprimeroligo} objects, and \code{type = "nucleotide"}:
#' If the plotted sequence should be displayed
#' as reverse complement or not.
#' \code{TRUE} or {FALSE}, defaults to \code{FALSE}.
#'
#' See examples below.
#'
#' @return A plot.
#'
#' @export
setGeneric("plotData", \(x, ...) standardGeneric("plotData"))

# Methods ======================================================================

#' @describeIn plotData
#'
#' @importFrom patchwork wrap_plots
#'
#' @export
#'
#' @examples
#' #### Plot an RprimerProfile object
#'
#' data("exampleRprimerProfile")
#' prof <- exampleRprimerProfile
#'
#' ## Plot an overview
#' plotData(prof, highlight = c(500, 1000))
#'
#' ## Select a region of interest
#' roi <- prof[prof$position >= 500 & prof$position <= 550, ]
#'
#' ## Plot an overview of the roi
#' plotData(roi)
#'
#' ## Plot the nucleotide distribution of the roi
#' plotData(roi, type = "nucleotide")
setMethod("plotData", "RprimerProfile", \(x,
    type = "overview",
    highlight = NULL,
    rc = FALSE) {
    if (nrow(x) == 0L) {
        stop("'x' does not contain any observations.", call. = FALSE)
    }
    if (type == "overview") {
        if (is.null(highlight)) {
            highlight <- -Inf
        }
        if (!is.numeric(highlight)) {
            stop("'highlight' must be numeric, e.g. c(1, 10).", call. = FALSE)
        }
        .plotOverview(x, highlight)
    } else if (type == "nucleotide") {
        if (!(is.logical(rc))) {
            stop("'rc' must be set to TRUE or FALSE.", call. = FALSE)
        }
        .plotNucleotides(x, rc)
    } else {
        stop("'type' must be either 'overview' or 'nucleotide'", call. = FALSE)
    }
})

#' @describeIn plotData
#'
#' @aliases plotData
#'
#' @importFrom patchwork wrap_plots
#'
#' @export
#'
#' @examples
#'
#' #### Plot an RprimerOligo object
#'
#' data("exampleRprimerOligo")
#' plotData(exampleRprimerOligo)
setMethod("plotData", "RprimerOligo", \(x) {
    if (nrow(x) == 0L) {
        stop("'x' does not contain any observations.", call. = FALSE)
    }
    x <- as.data.frame(x)
    patchwork::wrap_plots(
        list(.oligoPlot(x), .oligoFeaturePlot(x)),
        ncol = 1
    )
})

#' @describeIn plotData
#'
#' @importFrom patchwork wrap_plots
#'
#' @export
#'
#' @examples
#'
#' #### Plot an RprimerAssay object
#'
#' data("exampleRprimerAssay")
#' plotData(exampleRprimerAssay)
setMethod("plotData", "RprimerAssay", \(x) {
    if (nrow(x) == 0L) {
        stop("'x' does not contain any observations.", call. = FALSE)
    }
    x <- as.data.frame(x)
    patchwork::wrap_plots(
        list(.assayPlot(x), .assayFeaturePlot(x)),
        ncol = 1
    )
})

#' @describeIn plotData
#'
#' @export
#'
#' @examples
#'
#' #### Plot an RprimerMatchOligo object
#'
#' data("exampleRprimerMatchOligo")
#' plotData(exampleRprimerMatchOligo)
setMethod("plotData", "RprimerMatchOligo", \(x) {
    if (nrow(x) == 0L) {
        stop("'x' does not contain any observations.", call. = FALSE)
    }
    x <- as.data.frame(x)
    .plotMatch(x)
})

#' @describeIn plotData
#'
#' @export
#'
#' @examples
#'
#' #### Plot an RprimerMatchAssay object
#'
#' data("exampleRprimerMatchAssay")
#' plotData(exampleRprimerMatchAssay)
setMethod("plotData", "RprimerMatchAssay", \(x) {
    if (nrow(x) == 0L) {
        stop("'x' does not contain any observations.", call. = FALSE)
    }
    x <- as.data.frame(x)
    .plotMatchAssay(x)
})

# Helpers for plotting an RprimerOligo/Assay ===================================

.addEmptyRow <- function(x) {
    x[1, ] <- NA
    x$start <- 1
    x$end <- 1
    x
}

.splitOligoDf <- function(x) {
    all <- list()
    all$fwd <- x[x$type == "primer" & x$fwd, , drop = FALSE]
    all$rev <- x[x$type == "primer" & x$rev, , drop = FALSE]
    if (any(x$type == "probe")) {
        all$prPos <- x[x$type == "probe" & x$fwd, , drop = FALSE]
        all$prNeg <- x[x$type == "probe" & x$rev, , drop = FALSE]
    }
    nRows <- vapply(all, nrow, integer(1), USE.NAMES = FALSE)
    all[nRows == 0] <- lapply(all[nRows == 0], .addEmptyRow)
    all
}

.oligoPlot <- function(x) {
    if (all(x$type == "primer")) {
        .primerPlot(x)
    } else if (all(x$type == "probe")) {
        .probePlot(x)
    } else {
        .primerProbePlot(x)
    }
}

.primerPlot <- function(x) {
    start <- end <- NULL
    roiStart <- x$roiStart[[1]]
    roiEnd <- x$roiEnd[[1]]
    colors <- c(fwd = "#424B54", rev = "#93A8AC")
    oligos <- .splitOligoDf(x)
    ggplot2::ggplot() +
        ggplot2::xlim(roiStart, roiEnd) +
        ggplot2::ylim(0, 1) +
        ggplot2::labs(x = "Position", y = "") +
        ggplot2::geom_segment(
            color = "grey60", lwd = 2, ggplot2::aes(
                x = roiStart, xend = roiEnd, y = 0, yend = 0
            )
        ) +
        ggplot2::geom_rect(data = oligos$fwd, ggplot2::aes(
            xmin = start, xmax = end, ymin = 0.35, ymax = 0.65
        ), fill = colors["fwd"]) +
        ggplot2::geom_rect(data = oligos$rev, ggplot2::aes(
            xmin = start, xmax = end, ymin = 0.05, ymax = 0.35
        ), fill = colors["rev"]) +
        ggplot2::annotate(
            "label",
            x = roiStart,
            y = seq(0.89, length.out = 2, by = 0.07), label = c(
                paste(
                    "Reverse primer n =",
                    nrow(oligos$rev[!is.na(oligos$rev$length), ])
                ),
                paste(
                    "Forward primer n =",
                    nrow(oligos$fwd[!is.na(oligos$fwd$length), ])
                )
            ), size = 3, hjust = 0, fontface = 2,
            color = rev(colors), fill = "white", label.size = NA
        ) +
        .themeRprimer(showYAxis = FALSE)
}

.probePlot <- function(x) {
    start <- end <- NULL
    roiStart <- x$roiStart[[1]]
    roiEnd <- x$roiEnd[[1]]
    colors <- c(prPos = "#9B6A6C", prNeg = "#E2B4BD")
    oligos <- .splitOligoDf(x)
    ggplot2::ggplot() +
        ggplot2::xlim(roiStart, roiEnd) +
        ggplot2::ylim(0, 1) +
        ggplot2::labs(x = "Position", y = "") +
        ggplot2::geom_segment(
            color = "grey60", lwd = 2, ggplot2::aes(
                x = roiStart, xend = roiEnd, y = 0, yend = 0
            )
        ) +
        ggplot2::geom_rect(data = oligos$prPos, ggplot2::aes(
            xmin = start, xmax = end, ymin = 0.35, ymax = 0.65
        ), fill = colors["prPos"]) +
        ggplot2::geom_rect(data = oligos$prNeg, ggplot2::aes(
            xmin = start, xmax = end, ymin = 0.05, ymax = 0.35
        ), fill = colors["prNeg"]) +
        ggplot2::annotate(
            "label",
            x = roiStart,
            y = seq(0.89, length.out = 2, by = 0.07), label = c(
                paste(
                    "Probe (-) n =",
                    nrow(oligos$prNeg[!is.na(oligos$prNeg$length), ])
                ),
                paste(
                    "Probe (+) n =",
                    nrow(oligos$prPos[!is.na(oligos$prPos$length), ])
                )
            ), size = 3, hjust = 0, fontface = 2,
            color = rev(colors), fill = "white", label.size = NA
        ) +
        .themeRprimer(showYAxis = FALSE)
}

.primerProbePlot <- function(x) {
    start <- end <- NULL
    roiStart <- x$roiStart[[1]]
    roiEnd <- x$roiEnd[[1]]
    oligos <- .splitOligoDf(x)
    colors <- c(
        fwd = "#424B54", rev = "#93A8AC", prPos = "#9B6A6C", prNeg = "#E2B4BD"
    )
    ggplot2::ggplot() +
        ggplot2::xlim(roiStart, roiEnd) +
        ggplot2::ylim(0, 1) +
        ggplot2::labs(x = "Position", y = "") +
        ggplot2::geom_segment(
            color = "grey60", lwd = 2, ggplot2::aes(
                x = roiStart, xend = roiEnd, y = 0, yend = 0
            )
        ) +
        ggplot2::geom_rect(data = oligos$fwd, ggplot2::aes(
            xmin = start, xmax = end, ymin = 0.5, ymax = 0.65
        ), fill = colors["fwd"]) +
        ggplot2::geom_rect(data = oligos$rev, ggplot2::aes(
            xmin = start, xmax = end, ymin = 0.35, ymax = 0.5
        ), fill = colors["rev"]) +
        ggplot2::geom_rect(data = oligos$prPos, ggplot2::aes(
            xmin = start, xmax = end, ymin = 0.20, ymax = 0.35
        ), fill = colors["prPos"]) +
        ggplot2::geom_rect(data = oligos$prNeg, ggplot2::aes(
            xmin = start, xmax = end, ymin = 0.05, ymax = 0.20
        ), fill = colors["prNeg"]) +
        ggplot2::annotate(
            "label",
            x = roiStart, y = seq(0.75, length.out = 4, by = 0.07),
            label = c(
                paste(
                    "Probe (-) n =",
                    nrow(oligos$prNeg[!is.na(oligos$prNeg$length), ])
                ),
                paste(
                    "Probe (+) n =",
                    nrow(oligos$prPos[!is.na(oligos$prPos$length), ])
                ),
                paste(
                    "Reverse primer n =",
                    nrow(oligos$rev[!is.na(oligos$rev$length), ])
                ),
                paste(
                    "Forward primer n =",
                    nrow(oligos$fwd[!is.na(oligos$fwd$length), ])
                )
            ), size = 3, hjust = 0, fontface = 2,
            color = rev(colors), fill = "white", label.size = NA
        ) +
        .themeRprimer(showYAxis = FALSE)
}

.violinPlot <- function(data, y, title = "", color = "grey20") {
    ggplot2::ggplot() +
        ggplot2::geom_violin(
            data = data, ggplot2::aes(x = 1, y = y),
            fill = color, alpha = 0.4
        ) +
        ggplot2::geom_point(
            data = data, ggplot2::aes(x = 1, y = y), alpha = 0.05, color = color
        ) +
        ggplot2::geom_boxplot(
            data = data, ggplot2::aes(x = 1, y = y), width = 0.1,
            color = color, fill = color, alpha = 0.2
        ) +
        ggplot2::ylab("") +
        ggplot2::labs(title = title) +
        .themeRprimer(showXAxis = FALSE)
}

.boxPlot <- function(data, y, title = "", color = "grey20") {
    ggplot2::ggplot() +
        ggplot2::geom_point(
            data = data, ggplot2::aes(x = 1, y = y), alpha = 0.05, color = color
        ) +
        ggplot2::geom_boxplot(
            data = data, ggplot2::aes(x = 1, y = y), width = 0.1,
            color = color, fill = color, alpha = 0.2
        ) +
        ggplot2::ylab("") +
        ggplot2::labs(title = title) +
        .themeRprimer(showXAxis = FALSE)
}

.dotPlot <- function(data, y, title = "", color = "grey20") {
    ggplot2::ggplot() +
        ggplot2::geom_point(
            data = data, ggplot2::aes(x = 1, y = y), alpha = 0.5, color = color
        ) +
        ggplot2::ylab("") +
        ggplot2::labs(title = title) +
        .themeRprimer(showXAxis = FALSE)
}

.barPlot <- function(data, y, title = "", color = "grey20") {
    ggplot2::ggplot(data, ggplot2::aes(factor(y))) +
        ggplot2::geom_bar(
            fill = color, color = color, alpha = 0.4
        ) +
        ggplot2::xlab("") +
        ggplot2::ylab("") +
        ggplot2::labs(title = title) +
        .themeRprimer(showXAxis = TRUE)
}

.gcTmCoveragePlot <- function(x, color = "grey20", type = "Primers") {
    if (nrow(x) >= 10) {
        patchwork::wrap_plots(
            list(
                .violinPlot(
                    x, x$gcContentMean,
                    paste0("\n", type, "\n\nGC-content (mean)"),
                    color = color
                ),
                .violinPlot(x, x$tmMean, "\n\n\nTm (mean)", color = color),
                .violinPlot(
                    x, x$identity, "\n\n\nIdentity (mean)",
                    color = color
                ),
                .violinPlot(
                    x, x$coverage, "\n\n\nCoverage (mean)",
                    color = color
                ),
                .barPlot(x, x$length, "\n\n\nLength", color = color),
                .barPlot(x, x$degeneracy, "\n\n\nDegeneracy", color = color)
            ),
            ncol = 6
        )
    } else {
        patchwork::wrap_plots(
            list(
                .dotPlot(
                    x, x$gcContentMean,
                    paste0("\n", type, "\n\nGC-content (mean)"),
                    color = color
                ),
                .dotPlot(x, x$tmMean, "\n\n\nTm (mean)", color = color),
                .dotPlot(
                    x, x$identity, "\n\n\nIdentity (mean)",
                    color = color
                ),
                .dotPlot(
                    x, x$coverage, "\n\n\nCoverage (mean)",
                    color = color
                ),
                .barPlot(x, x$length, "\n\n\nLength", color = color),
                .barPlot(x, x$degeneracy, "\n\n\nDegeneracy", color = color)
            ),
            ncol = 6
        )
    }
}

.oligoFeaturePlot <- function(x) {
    if (all(x$type == "probe")) {
        patchwork::wrap_plots(list(
            .gcTmCoveragePlot(
                x[x$type == "probe", ],
                color = "grey20",
                type = "Probes"
            )
        ), ncol = 1)
    } else if (all(x$type == "primer")) {
        patchwork::wrap_plots(list(
            .gcTmCoveragePlot(
                x[x$type == "primer", ],
                color = "grey20",
                type = "Primers"
            )
        ), ncol = 1)
    } else {
        patchwork::wrap_plots(list(
            .gcTmCoveragePlot(
                x[x$type == "primer", ],
                color = "grey20", type = "Primers"
            ),
            .gcTmCoveragePlot(
                x[x$type == "probe", ],
                color = "grey20", type = "Probes"
            )
        ), ncol = 1)
    }
}

.assayPlot <- function(x) {
    start <- x$roiStart[[1]]
    end <- x$roiEnd[[1]]
    ggplot2::ggplot() +
        ggplot2::xlim(start, end) +
        ggplot2::ylim(0, 1) +
        ggplot2::labs(x = "Position", y = "") +
        ggplot2::geom_segment(
            color = "grey", lwd = 2,
            ggplot2::aes(x = start, xend = end, y = 0, yend = 0)
        ) +
        ggplot2::geom_rect(
            data = x[!duplicated(data.frame(x$start, x$end)), ],
            ggplot2::aes(
                xmin = start, xmax = end, ymin = 0.05, ymax = 0.65
            ), fill = "#424B54", alpha = 0.5
        ) +
        ggplot2::annotate(
            "label",
            x = start, y = 0.8,
            label = paste(
                "Assays n =", nrow(x[!is.na(x$length), ])
            ), size = 3, hjust = 0, fontface = 2,
            color = "#424B54", fill = "white", label.size = NA
        ) +
        .themeRprimer(showYAxis = FALSE)
}

.assayFeaturePlot <- function(x, color = "grey20") {
    if (nrow(x) >= 10) {
        patchwork::wrap_plots(
            list(
                .boxPlot(
                    x, x$length, "\n\nAmplicon length",
                    color = color
                ),
                .barPlot(
                    x, x$totalDegeneracy, "\n\nTotal degeneracy",
                    color = color
                )
            ),
            ncol = 2
        )
    } else {
        patchwork::wrap_plots(
            list(
                .barPlot(
                    x, x$length, "\n\nAmplicon length",
                    color = color
                ),
                .barPlot(
                    x, x$totalDegeneracy, "\n\nTotal degeneracy",
                    color = color
                )
            ),
            ncol = 2
        )
    }
}

# Helpers for plotting an RprimerProfile =======================================

.identifyMask <- function(x) {
    x <- as.data.frame(x)
    x <- x[c("a", "c", "g", "t", "other", "gaps")]
    x <- as.matrix(x)
    masked <- apply(x, 1, \(y) all(is.na(y)))
    which(masked)
}

#' Calculate running average
#'
#' @param x A numeric vector.
#'
#' @param size
#' The number of observations in each average.
#' If \code{NULL}, the size will be set to the nearest positive, nonzero
#' integer to \code{length(x) / 100}.
#'
#' @return A data frame with position and running average of \code{x}.
#'
#' @noRd
.runningAverage <- function(x, size = NULL) {
    if (is.null(size)) {
        size <- round(length(x) / 100)
        if (size == 0) size <- 1
    }
    sums <- c(0, cumsum(x))
    from <- seq_len(length(sums) - size)
    to <- seq(size + 1, length(sums))
    average <- (sums[to] - sums[from]) / size
    position <- seq(size, length(x))
    if (size > 1) {
        midpoint <- size / 2
        position <- position - midpoint
    }
    data.frame(position, average)
}

.identityPlot <- function(x, highlight = NULL, mask) {
    position <- identity <- average <- NULL
    averages <- .runningAverage(x$identity)
    xadj <- unique(x$position - seq_along(x$position))
    averages$position <- averages$position + xadj
    ggplot2::ggplot(
        data = x, ggplot2::aes(x = position, y = identity)
    ) +
        .highlightRegion(highlight) +
        ggplot2::geom_point(alpha = 1 / 3, shape = 1, color = "#93A8AC") +
        ggplot2::geom_line(
            data = averages, color = "#1B1C22",
            ggplot2::aes(x = position, y = average)
        ) +
        .maskRegion(mask + xadj) +
        ggplot2::ylim(0, 1) +
        ggplot2::ylab("Identity") +
        ggplot2::xlab("") +
        .themeRprimer(showXAxis = FALSE)
}

.gcPlot <- function(x, highlight = NULL, mask) {
    position <- average <- NULL
    gc <- ifelse(x$majority == "C" | x$majority == "G", 1, 0)
    gc[x$majority == "-"] <- 0.5
    gc[is.na(gc)] <- 0.5
    averages <- .runningAverage(gc)
    xadj <- unique(x$position - seq_along(x$position))
    averages$position <- averages$position + xadj
    ggplot2::ggplot(data = x, ggplot2::aes(x = position)) +
        ggplot2::geom_segment(
            color = "#93A8AC",
            ggplot2::aes(
                x = min(position), xend = max(position), y = 0.5, yend = 0.5
            )
        ) +
        .highlightRegion(highlight) +
        ggplot2::geom_line(
            data = averages, color = "#1B1C22",
            ggplot2::aes(x = position, y = average)
        ) +
        .maskRegion(mask + xadj) +
        ggplot2::xlab("") +
        ggplot2::ylab("GC-content") +
        ggplot2::ylim(0, 1) +
        .themeRprimer(showXAxis = FALSE)
}

.gapPlot <- function(x, highlight = NULL, mask) {
    position <- gaps <- NULL
    xadj <- unique(x$position - seq_along(x$position))
    ggplot2::ggplot(
        data = x, ggplot2::aes(x = position, y = gaps)
    ) +
        .highlightRegion(highlight) +
        ggplot2::geom_point(alpha = 1 / 3, shape = 1, color = "#93A8AC") +
        .maskRegion(mask + xadj) +
        ggplot2::ylim(0, 1) +
        ggplot2::xlab("Position") +
        ggplot2::ylab("Gaps") +
        .themeRprimer()
}

.maskRegion <- function(x) {
    ggplot2::geom_vline(xintercept = x, color = "grey80")
}

.highlightRegion <- function(highlight = NULL) {
    ggplot2::annotate(
        "rect",
        xmin = min(highlight), xmax = max(highlight), ymin = -Inf, ymax = Inf,
        color = "white", alpha = 0.4, fill = "#9B6A6C"
    )
}

.plotOverview <- function(x, highlight = NULL) {
    x <- as.data.frame(x)
    x$identity[is.na(x$identity)] <- 1
    x$coverage[is.na(x$coverage)] <- 1
    mask <- .identifyMask(x)
    patchwork::wrap_plots(
        list(
            .identityPlot(x, highlight, mask),
            .gcPlot(x, highlight, mask),
            .gapPlot(x, highlight, mask)
        ),
        ncol = 1
    )
}

.plotNucleotides <- function(x, rc = FALSE) {
    position <- x$position
    x <- as.data.frame(x)
    select <- c("a", "c", "g", "t", "other")
    x <- x[names(x) %in% select]
    x <- t(as.matrix(x))
    colnames(x) <- position
    rownames(x)[-5] <- toupper(rownames(x))[-5]
    if (rc) {
        rownames(x) <- unname(lookup$complement[rownames(x)])
        x <- x[order(rownames(x)), ]
        if ("other" %in% rownames(x)) {
            swap <- c(which(rownames(x) == "other"), nrow(x))
            x[swap, ] <- x[rev(swap), ]
            rownames(x)[swap] <- rownames(x)[rev(swap)]
        }
    }
    Base <- Position <- Frequency <- NULL
    x <- reshape2::melt(x)
    names(x) <- c("Base", "Position", "Frequency")
    x$Base <- factor(x$Base)
    x$Position <- factor(x$Position)
    if (rc) {
        x$Position <- factor(x$Position, levels = rev(levels(x$Position)))
    }
    basePalette <- c(
        "other" = "#D1D6DB",
        "A" = "#424B54", "C" = "#93A8AC", "G" = "#E2B4BD", "T" = "#9B6A6C"
    )
    ggplot2::ggplot(
        data = x, ggplot2::aes(x = Position, y = Frequency, fill = Base)
    ) +
        ggplot2::geom_bar(stat = "identity") +
        ggplot2::scale_fill_manual(values = basePalette) +
        ggplot2::ylab("Proportion") +
        .themeRprimer(showLegend = TRUE)
}

# Helpers for plotting an RprimerMatchOligo object =============================

.plotMatch <- function(x, showLegend = TRUE, type = "oligo") {
    mismatches <- value <- NULL
    id <- as.factor(seq_len(nrow(x)))
    x <- x[!(grepl("id", names(x)))]
    x <- cbind(id, x)
    x <- suppressMessages({reshape2::melt(x)})
    names(x)[3] <- "mismatches"
    levels(x$mismatches) <- c(
        "0 mismatches", "1 mismatch", "2 mismatches", "3 mismatches",
        "4 or more mismatches", "Off-target match (< 3 mismatches)"
    )
    if (type == "oligo") {
        yLabels <- x$iupacSequence
    } else {
        yLabels <- paste0(
            "assay ", id, ", ", type,
            " (length: ", nchar(x$iupacSequence), ")"
        )
    }
    offTarget <- x[x$mismatches == "Off-target match (< 3 mismatches)", ]
    onTarget <- x[x$mismatches != "Off-target match (< 3 mismatches)", ]
    ggplot2::ggplot(data = onTarget, ggplot2::aes(x = id)) +
        ggplot2::geom_bar(
            ggplot2::aes(x = id, y = value, fill = mismatches),
            stat = "identity", position = "stack"
        ) +
        ggplot2::ylab("Proportion") +
        ggplot2::xlab("") +
        ggplot2::scale_x_discrete(
            limits = rev(levels(factor(x$id))),
            labels = rev(yLabels)
        ) +
        ggplot2::coord_flip() +
        ggplot2::geom_bar(
            data = offTarget,
            ggplot2::aes(x = id, y = value, fill = mismatches),
            stat = "identity", position = "dodge", width = 0.2,
            alpha = 0.9
        ) +
        ggplot2::scale_fill_manual(
            values = c(
                "#BDC9CC", "#B4B1B4", "#AC999C",
                "#A38183", "#9B6A6C", "#424B54"
            )
        ) +
        .themeRprimer(showLegend = showLegend)
}

# Helpers for plotting an RprimerMatchAssay object =============================

.plotMatchAssay <- function(x) {
    fwd <- x[grepl("Fwd$", names(x))]
    names(fwd) <- gsub("Fwd", "", names(fwd))
    rev <- x[grepl("Rev$", names(x))]
    names(rev) <- gsub("Rev", "", names(rev))
    if (any(grepl("Pr$", names(x)))) {
        pr <- x[grepl("Pr$", names(x))]
        names(pr) <- gsub("Pr", "", names(pr))
        patchwork::wrap_plots(
            list(
                .plotMatch(fwd, showLegend = FALSE, type = "fwd"),
                .plotMatch(pr, type = "pr", showLegend = FALSE),
                .plotMatch(rev, type = "rev")
            ),
            ncol = 3
        )
    } else {
        patchwork::wrap_plots(
            list(
                .plotMatch(fwd, showLegend = FALSE, type = "fwd"),
                .plotMatch(rev, type = "rev")
            ),
            ncol = 2
        )
    }
}

# Plot theme ===================================================================

.themeRprimer <- function(showXAxis = TRUE,
                          showYAxis = TRUE,
                          showLegend = FALSE) {
    if (showXAxis && showYAxis) {
        .showXYAxes(showLegend)
    } else if (showXAxis && !showYAxis) {
        .showXAxisHideYAxis(showLegend)
    } else if (!showXAxis && showYAxis) {
        .showYAxisHideXAxis(showLegend)
    } else {
        .hideXYAxes(showLegend)
    }
}

.showXYAxes <- function(showLegend = TRUE) {
    ggplot2::theme_bw() +
        ggplot2::theme(
            legend.title = ggplot2::element_blank(),
            legend.text = ggplot2::element_text(size = 9),
            legend.position = ifelse(showLegend, "right", "none"),
            axis.title.y = ggplot2::element_text(
                size = 9,
                margin = ggplot2::margin(r = 10)
            ),
            axis.title.x = ggplot2::element_text(
                size = 9,
                margin = ggplot2::margin(t = 10)
            ),
            plot.title = ggplot2::element_text(size = 9),
            plot.margin = ggplot2::unit(rep(0.1, 4), "cm")
        )
}

.showXAxisHideYAxis <- function(showLegend = TRUE) {
    ggplot2::theme_bw() +
        ggplot2::theme(
            legend.title = ggplot2::element_blank(),
            legend.text = ggplot2::element_text(size = 9),
            legend.position = ifelse(showLegend, "right", "none"),
            axis.title.x = ggplot2::element_text(
                size = 9,
                margin = ggplot2::margin(t = 10)
            ),
            panel.grid.major.y = ggplot2::element_blank(),
            panel.grid.minor.y = ggplot2::element_blank(),
            axis.title.y = ggplot2::element_blank(),
            axis.text.y = ggplot2::element_blank(),
            axis.ticks.y = ggplot2::element_blank(),
            plot.title = ggplot2::element_text(size = 9),
            plot.margin = ggplot2::unit(rep(0.1, 4), "cm")
        )
}

.showYAxisHideXAxis <- function(showLegend = TRUE) {
    ggplot2::theme_bw() +
        ggplot2::theme(
            legend.title = ggplot2::element_blank(),
            legend.text = ggplot2::element_text(size = 9),
            legend.position = ifelse(showLegend, "right", "none"),
            axis.title.y = ggplot2::element_text(
                size = 9,
                margin = ggplot2::margin(r = 10)
            ),
            axis.title.x = ggplot2::element_blank(),
            axis.text.x = ggplot2::element_blank(),
            axis.ticks.x = ggplot2::element_blank(),
            plot.title = ggplot2::element_text(size = 9),
            plot.margin = ggplot2::unit(rep(0.1, 4), "cm")
        )
}

.hideXYAxes <- function(showLegend = TRUE) {
    ggplot2::theme_bw() +
        ggplot2::theme(
            legend.title = ggplot2::element_blank(),
            legend.text = ggplot2::element_text(size = 9),
            legend.position = ifelse(showLegend, "right", "none"),
            axis.title.y = ggplot2::element_blank(),
            axis.text.y = ggplot2::element_blank(),
            axis.ticks.y = ggplot2::element_blank(),
            axis.title.x = ggplot2::element_blank(),
            axis.text.x = ggplot2::element_blank(),
            axis.ticks.x = ggplot2::element_blank(),
            plot.title = ggplot2::element_text(size = 9),
            plot.margin = ggplot2::unit(rep(0.1, 4), "cm")
        )
}

## Import data to test on
data("exampleRprimerProfile")
data("exampleRprimerOligo")
data("exampleRprimerAssay")

# .plotData ====================================================================

test_that("plotData works", {
    p <- plotData(exampleRprimerProfile)
    expect_true(ggplot2::is.ggplot(p))
    p <- plotData(exampleRprimerProfile[1:100, ], highlight = c(1, 10))
    expect_true(ggplot2::is.ggplot(p))
    expect_error(plotData(exampleRprimerProfile, highlight = TRUE))
    p <- plotData(exampleRprimerOligo)
    expect_true(ggplot2::is.ggplot(p))
    p <- plotData(exampleRprimerOligo[exampleRprimerOligo$type == "primer", ])
    expect_true(ggplot2::is.ggplot(p))
    p <- plotData(exampleRprimerOligo[1, ])
    expect_true(ggplot2::is.ggplot(p))
    p <- plotData(exampleRprimerOligo[exampleRprimerOligo$type == "probe", ])
    expect_true(ggplot2::is.ggplot(p))
    p <- plotData(exampleRprimerAssay)
    expect_true(ggplot2::is.ggplot(p))
    p <- plotData(exampleRprimerAssay[1, ])
    expect_true(ggplot2::is.ggplot(p))
    p <- plotData(exampleRprimerProfile[1:10, ], type = "nucleotide")
    expect_true(ggplot2::is.ggplot(p))
    p <- plotData(exampleRprimerProfile[1:10, ], rc = TRUE, type = "nucleotide")
    expect_true(ggplot2::is.ggplot(p))
    expect_error(plotData(exampleRprimerProfile, rc = "t", type = "nucleotide"))
    expect_error(plotData(exampleRprimerProfile, type = "nt"))
    p <- ggplot2::ggplot() +
        .themeRprimer(showXAxis = FALSE, showYAxis = FALSE)
    expect_true(ggplot2::is.ggplot(p))
})

test_that("plotData returns an error when it should", {
    expect_error(plotData(unclass(exampleRprimerProfile)))
    expect_error(plotData(exampleRprimerProfile, shadeFrom = FALSE))
    expect_error(plotData(exampleRprimerProfile, shadeTo = FALSE))
    expect_error(plotData(
        exampleRprimerProfile,
        type = "nucleotide", rc = "FALSE"
    ))
    expect_error(plotData(exampleRprimerProfile, type = ""))
    expect_error(plotData(unclass(exampleRprimerOligo)))
    expect_error(plotData(unclass(exampleRprimerAssay)))
})

test_that(".runningAverage works", {
    toTest <- runif(100)
    expect_identical(nrow(.runningAverage(toTest)), length(toTest))
    toTest <- runif(10)
    expect_identical(nrow(.runningAverage(toTest)), length(toTest))
    toTest <- runif(1000)
    expect_equal(.runningAverage(toTest, size = 100)$position[1], 50)
})

#' Get sequence information from an alignment
#'
#' \code{consensusProfile()} takes a DNA multiple alignment as input and
#' returns all the data needed for subsequent primer and probe design process.
#' The function is a wrapper to
#' \code{Biostrings::consensusMatrix()} (Pages et al., 2020).
#'
#' @param x
#' A \code{MultipleAlignment::DNAMultipleAlignment} object.
#'
#' @param ambiguityThreshold
#' "Detection level" for ambiguous bases.
#' All DNA bases that occur with a relative frequency higher than the
#' specified value will be included when the IUPAC consensus character
#' is determined.
#' Can range from 0 to 0.2, defaults to \code{0}.
#'
#' @return
#' An \code{RprimerProfile} object, which contains the following information:
#'
#' \describe{
#'   \item{position}{Position in the alignment.}
#'   \item{a}{Proportion of A.}
#'   \item{c}{Proportion of C.}
#'   \item{g}{Proportion of G.}
#'   \item{t}{Proportion of T.}
#'   \item{other}{Proportion of bases other than A, C, G, T.}
#'   \item{gaps}{Proportion of gaps (recognized as "-" in the alignment).}
#'   \item{majority}{Majority consensus sequence.
#'   Denotes the most frequently occurring nucleotide.
#'   If two or more bases occur with the same frequency,
#'   the consensus nucleotide will be randomly selected among these.}
#'   \item{identity}{Proportion of sequences, among all sequences with a
#'   DNA base (i.e., A, C, G or T), that has the majority consensus base.}
#'   \item{iupac}{The consensus sequence expressed in IUPAC format.
#'   The IUPAC consensus sequence only
#'   takes 'A', 'C', 'G', 'T' and '-' as input. Degenerate bases
#'   will be skipped. If a position only contains
#'   degenerate bases, the IUPAC consensus will be \code{NA} at that
#'   position.}
#'   \item{coverage}{Proportion of sequences in the target alignment,
#'   among all sequences with a DNA base, that are covered the IUPAC consensus
#'   character.
#'   The value will be 1 if there are no "remaining" DNA bases (and/or if
#'   \code{ambiguityThreshold = 0}).}
#' }
#'
#' @references
#' Pages, H., Aboyoun, P., Gentleman R., and DebRoy S. (2020). Biostrings:
#' Efficient manipulation of biological strings. R package version
#' 2.57.2.
#'
#' @export
#'
#' @examples
#' data("exampleRprimerAlignment")
#' consensusProfile(exampleRprimerAlignment)
#'
#' consensusProfile(exampleRprimerAlignment, ambiguityThreshold = 0.05)
consensusProfile <- function(x, ambiguityThreshold = 0) {
    if (!methods::is(x, "DNAMultipleAlignment")) {
        stop("'x' must be a DNAMultipleAlignment object.", call. = FALSE)
    }
    if (!(ambiguityThreshold >= 0 && ambiguityThreshold <= 0.2)) {
        stop(
            "'ambiguityThreshold' must be a number from 0 to 0.2.",
            call. = FALSE
        )
    }
    x <- .consensusMatrix(x)
    profile <- list()
    profile$position <- seq_len(ncol(x))
    profile$a <- unname(x["A", ])
    profile$c <- unname(x["C", ])
    profile$g <- unname(x["G", ])
    profile$t <- unname(x["T", ])
    profile$other <- unname(x["other", ])
    profile$gaps <- unname(x["-", ])
    profile$majority <- .majorityConsensus(x)
    profile$identity <- .nucleotideIdentity(x)
    profile$iupac <- .iupacConsensus(x, ambiguityThreshold)
    profile$iupac[profile$majority == "-"] <- "-"
    profile$coverage <- .coverage(x, ambiguityThreshold)
    profile$coverage[profile$majority == "-"] <- NA
    RprimerProfile(profile)
}

# Helpers ======================================================================

#' @noRd
#'
#' @examples
#' data("exampleRprimerAlignment")
#' .consensusMatrix(exampleRprimerAlignment)
.consensusMatrix <- function(x) {
    x <- Biostrings::consensusMatrix(x, as.prob = TRUE)
    colnames(x) <- seq_len(ncol(x))
    bases <- c("A", "C", "G", "T", "-")
    other <- colSums(x[!rownames(x) %in% bases, , drop = FALSE])
    x <- x[rownames(x) %in% bases, , drop = FALSE]
    rbind(x, other)
}

#' @noRd
#'
#' @examples
#' data("exampleRprimerAlignment")
#' x <- .consensusMatrix(exampleRprimerAlignment)
#' .findMostCommonBase(x[, 1])
.findMostCommonBase <- function(x) {
    mostCommon <- names(x)[x == max(x)]
    if (length(mostCommon) > 1) mostCommon <- sample(mostCommon, 1)
    mostCommon
}

#' @noRd
#'
#' @examples
#' data("exampleRprimerAlignment")
#' x <- .consensusMatrix(exampleRprimerAlignment)
#' .majorityConsensus(x)
.majorityConsensus <- function(x) {
    consensus <- apply(x, 2, .findMostCommonBase)
    unname(consensus)
}

#' @noRd
#'
#' @examples
#' .asIUPAC("A,G,C")
.asIUPAC <- function(x) {
    x <- unlist(strsplit(x, split = ","), use.names = FALSE)
    x <- x[order(x)]
    x <- unique(x)
    bases <- c("A", "C", "G", "T", "-")
    x <- x[x %in% bases]
    x <- paste(x, collapse = ",")
    unname(lookup$iupac[x])
}

#' @noRd
#'
#' @examples
#' data("exampleRprimerAlignment")
#' x <- .consensusMatrix(exampleRprimerAlignment)
#' .dnaBasesOnly(x)
.dnaBasesOnly <- function(x) {
    bases <- c("A", "C", "G", "T")
    s <- x[rownames(x) %in% bases, , drop = FALSE]
    apply(s, 2, \(x) x / sum(x))
}

#' @noRd
#'
#' @examples
#' data("exampleRprimerAlignment")
#' x <- .consensusMatrix(exampleRprimerAlignment)
#' .iupacConsensus(x)
.iupacConsensus <- function(x, ambiguityThreshold = 0) {
    x <- .dnaBasesOnly(x)
    basesToInclude <- apply(x, 2, \(y) {
        paste(rownames(x)[y > ambiguityThreshold], collapse = ",")
    })
    consensus <- vapply(
        basesToInclude, .asIUPAC, character(1L),
        USE.NAMES = FALSE
    )
    consensus
}

#' @noRd
#'
#' @examples
#' data("exampleRprimerAlignment")
#' x <- .consensusMatrix(exampleRprimerAlignment)
#' .nucleotideIdentity(x)
.nucleotideIdentity <- function(x) {
    x <- .dnaBasesOnly(x)
    identity <- apply(x, 2, max)
    identity[is.nan(identity)] <- NA
    unname(identity)
}

#' @noRd
#'
#' @examples
#' data("exampleRprimerAlignment")
#' x <- .consensusMatrix(exampleRprimerAlignment)
#' .coverage(x, ambiguityThreshold = 0.05)
.coverage <- function(x, ambiguityThreshold = 0) {
    x <- .dnaBasesOnly(x)
    x[x > ambiguityThreshold] <- 0
    coverage <- 1 - colSums(x)
    unname(coverage)
}

test_that("Data import and workflow works.", {
    infile <- system.file("extdata", "example_alignment.txt", package = "rprimer")
    aln <- MultipleAlignment::readDNAMultipleAlignment(infile, format = "fasta")
    consensus <- consensusProfile(aln, 0.1)
    ols <- designOligos(consensus[1:2000, ])
    expect_s4_class(ols, "RprimerOligo")
})

## Import data to test on
infile <- system.file("extdata", "example_alignment.txt", package = "rprimer")
testdata <- MultipleAlignment::readDNAMultipleAlignment(infile)
testmat <- .consensusMatrix(testdata)

# consensusProfile =============================================================

test_that("consensusProfile returns an error when it should", {
    expect_error(consensusProfile(unclass(testdata)))
    expect_error(consensusProfile(testdata, -0.1))
    expect_error(consensusProfile(testdata, 0.21))
})

test_that("consensusProfile works", {
    expect_s4_class(consensusProfile(testdata), "RprimerProfile")
})

test_that("consensusProfile works with a rowmask", {
    ## Select only one sequence
    MultipleAlignment::rowmask(testdata, invert = TRUE) <- 3
    prof <- consensusProfile(testdata)
    expect_s4_class(prof, "RprimerProfile")
    expect_true(all(prof$gaps == 0 | prof$gaps == 1))
    expect_true(all(prof$coverage == 1 | is.na(prof$coverage)))
})

test_that("consensusProfile works with a colmask", {
    MultipleAlignment::colmask(testdata, invert = TRUE) <- 500:550
    prof <- consensusProfile(testdata)
    expect_s4_class(prof, "RprimerProfile")
})

# .consensusMatrix =============================================================

test_that(".consensusMatrix works", {
    expect_true(is.matrix(testmat))
    expect_equal(nrow(testmat), 6)
    expect_equal(rownames(testmat), c("A", "C", "G", "T", "-", "other"))
    expect_false(any(is.na(testmat)))
})

# .findMostCommonBase ==========================================================

test_that(".findMostCommonbase works", {
    testvec <- testmat[, 2]
    expect_equal(.findMostCommonBase(testvec), "-")
})

# .majorityConsensus ===========================================================

test_that(".majorityConsensus works", {
    testmat <- testmat[, 1:5]
    expect_equal(.majorityConsensus(testmat), rep("-", 5))
})

# .asIUPAC =====================================================================

test_that(".asIUPAC works", {
    expect_true(is.na(.asIUPAC("GC")))
    expect_equal(.asIUPAC("C,G"), "S")
    expect_equal(.asIUPAC("C,G,G"), "S")
    expect_equal(.asIUPAC("-"), "-")
    expect_true(is.na(.asIUPAC("X")))
})

# .dnaBasesOnly ================================================================

test_that(".dnaBasesOnly works", {
    dnaBases <- .dnaBasesOnly(testmat)[, 1:10]
    expect_equal(rownames(dnaBases), c("A", "C", "G", "T"))
    expect_true(
        all(colSums(dnaBases) > 0.9999999 & colSums(dnaBases < 1.00000001))
    )
})

# .iupacConsensus ==============================================================

test_that(".iupacConsensus works", {
    selection <- testmat[, 200:220]
    x <- selection[, 1:4]
    bases <- c("A", "C", "G", "T", "-")
    s <- x[rownames(x) %in% bases, , drop = FALSE]
    s <- apply(s, 2, function(x) x / sum(x))
    expect_equal(colSums(s), rep(1, 4), ignore_attr = TRUE)
})

# .nucleotideIdentity ==========================================================

test_that(".nucleotideIdentity works", {
    selection <- testmat[, 200:220]
    x <- selection[, 1:4]
    bases <- c("A", "C", "G", "T", "-")
    s <- x[rownames(x) %in% bases, , drop = FALSE]
    s <- apply(s, 2, function(x) x / sum(x))
    expect_equal(
        .nucleotideIdentity(x),
        c(max(s[, 1]), max(s[, 2]), max(s[, 3]), max(s[, 4]))
    )
})

# .coverage ====================================================================

# Helpers ======================================================================

#' @noRd
#'
#' @examples
#' .nn(c("A", "G", "C", "T"))
.nn <- function(x) {
    from <- (seq_along(x) - 1)[-1]
    to <- seq_along(x)[-1]
    vapply(seq_along(from), \(i) {
        paste(x[from[[i]]:to[[i]]], collapse = "")
    }, character(1L))
}

#' @noRd
#'
#' @examples
#' x <- t(matrix(.nn(c("A", "C", "G"))))
#' .stack(x)
.stack <- function(x, table = "dH") {
    selected <- lookup$nn[[table]]
    stack <- matrix(nrow = nrow(x), ncol = ncol(x), dimnames = dimnames(x))
    stack[] <- selected[match(x, lookup$nn$bases)]
    stack
}

#' @noRd
#'
#' @examples
#' x <- t(matrix(.nn(c("A", "C", "G"))))
#' .initiate(x)
.initiate <- function(x, table = "dH") {
    initiation <- lookup$nn[lookup$nn$bases == "Initiation", ][[table]]
    penalty <- lookup$nn[lookup$nn$bases == "AT_penalty", ][[table]]
    penaltyFirst <- ifelse(grepl("^(A|T)", x[, 1]), penalty, 0)
    penaltyLast <- ifelse(grepl("(A|T)$", x[, ncol(x)]), penalty, 0)
    initiation + penaltyFirst + penaltyLast
}

#' @noRd
#'
#' @examples
#' .tmParameters(t(matrix(c("A", "G", "T", "T", "C", "G", "G", "T", "C", "G"))))
.tmParameters <- function(x, concNa = 0.05) {
    nn <- t(apply(x, 1, .nn))
    dhStack <- rowSums(.stack(nn, "dH"))
    dsStack <- rowSums(.stack(nn, "dS"))
    dhInit <- .initiate(nn, "dH")
    dsInit <- .initiate(nn, "dS")
    sumdH <- dhStack + dhInit
    sumdS <- dsStack + dsInit
    n <- dim(x)[[2]] - 1 ## Number of phosphates
    sumdS <- sumdS + 0.368 * n * log(concNa) ## Salt correction of deltaS
    m <- matrix(c(sumdH, sumdS), ncol = 2)
    rownames(m) <- rownames(nn)
    colnames(m) <- c("sumdH", "sumdS")
    m
}

#' @noRd
#'
#' @examples
#' x <- .tmParameters(t(matrix(c("A", "G", "T", "T", "C", "G", "G", "T", "C"))))
#' .tm(x)
.tm <- function(x, concOligo = 250) {
    concOligo <- concOligo * 1e-9
    tm <- (x[, "sumdH"] * 1000) / (x[, "sumdS"] + 1.987 * log(concOligo)) - 273.15
    names(tm) <- rownames(x)
    tm
}

#' @noRd
#'
#' @examples
#' x <- .tmParameters(t(matrix(c("C", "G", "T", "T", "G", "A"))), concNa = 1)
#' .deltaG(x)
.deltaG <- function(x, temperature = 37) {
    temperature <- temperature + 273.15
    dG <- (x[, "sumdH"] * 1000 - temperature * x[, "sumdS"]) / 1000
    names(dG) <- rownames(x)
    dG
}

# .nn ==========================================================================

test_that(".nn works", {
    x <- c("A", "C", "G")
    expect_equal(.nn(x), c("AC", "CG"))
})

# .stack =======================================================================

test_that(".stack works", {
    x <- t(matrix(.nn(c("A", "C", "G"))))
    expect_true(is.matrix(.stack(x)))
})

# .initiate ====================================================================

test_that(".initiate works", {
    x1 <- t(matrix(.nn(c("A", "T", "C"))))
    x2 <- t(matrix(.nn(c("T", "T", "C"))))
    x3 <- t(matrix(.nn(c("C", "T", "C"))))
    x4 <- t(matrix(.nn(c("A", "C", "A"))))
    x5 <- t(matrix(.nn(c("A", "C", "T"))))
    initiation <- lookup$nn$dH[lookup$nn$bases == "Initiation"]
    at_penalty <- lookup$nn$dH[lookup$nn$bases == "AT_penalty"]
    expect_equal(.initiate(x1), initiation + at_penalty)
    expect_equal(.initiate(x2), initiation + at_penalty)
    expect_equal(.initiate(x3), initiation)
    expect_equal(.initiate(x4), initiation + 2 * at_penalty)
    expect_equal(.initiate(x5), initiation + 2 * at_penalty)
})

# .tmParameters ================================================================

test_that(".tmParameters works", {
    x <- t(matrix(c("A", "G", "T", "T", "C", "G", "G", "T", "C", "G")))
    test <- .tmParameters(x)
    expect_true(is.matrix(test))
    test <- .tmParameters(rbind(x, x))
    expect_true(all(test[1, ] == test[2, ]))
})

# .tm ==========================================================================

test_that(".tm works", {
    x <- .tmParameters(t(matrix(
        c("A", "G", "T", "T", "C", "G", "G", "T", "C")
    )))
    test1 <- .tm(x, 250)
    expect_true(is.numeric(test1))
    test2 <- .tm(x, 500)
    expect_true(is.numeric(test2))
    expect_true(test2 > test1)
    x <- rbind(x, x)
    test <- .tm(x)
    expect_true(test[1] == test[2])
})

# .deltaG ======================================================================

test_that(".deltaG works", {
    ## Confirm that the calculation agrees with the examples in
    ## SantaLucia and Hicks 2004
    m <- t(matrix(unlist(strsplit("CGTTGA", split = ""))))
    x <- .tmParameters(m, concNa = 1)
    test <- .deltaG(x)
    expect_equal(round(test, 2), -5.36)
    x <- .tmParameters(m, concNa = 0.115)
    test <- .deltaG(x)
    expect_equal(round(test, 2), -4.12)
})

#' Design primers and probes
#'
#' \code{designOligos()} designs oligos (primers and probes)
#' from a consensus profile.
#'
#' @param x An \code{RprimerProfile} object.
#'
#' @param maxGapFrequency
#' Maximum allowed gap frequency at the primer and probe binding sites in
#' the target alignment.
#' A number [0, 1], defaults to \code{0.01}.
#'
#' @param lengthPrimer
#' Primer length range. A numeric vector [15, 30],
#' defaults to \code{c(18, 22)}.
#'
#' @param maxDegeneracyPrimer
#' Maximum number of variants of each primer. A number [1, 64],
#' defaults to \code{4}.
#'
#' @param gcClampPrimer
#' If primers must have a GC clamp.
#' A GC clamp
#' is identified as two to three G or
#' C:s within the last five bases (3' end) of the primer.
#' \code{TRUE} or \code{FALSE}, defaults to \code{TRUE}.
#'
#' @param avoidThreeEndRunsPrimer
#' If primers with more than two runs
#' of the same nucleotide at the terminal 3' end should be avoided.
#' \code{TRUE} or \code{FALSE}, defaults to \code{TRUE}.
#'
#' @param gcPrimer
#' GC-content range for primers.
#' A numeric vector [0, 1], defaults to \code{c(0.40, 0.65)}.
#'
#' @param tmPrimer
#' Tm range for primers (in Celcius degrees).
#' A numeric vector [30, 90], defaults to \code{c(55, 65)}.
#'
#' @param concPrimer
#' Primer concentration in nM, for tm calculation. A number
#' [20, 2000], defaults to \code{500}.
#'
#' @param designStrategyPrimer
#' \code{"ambiguous"} or \code{"mixed"}. Defaults to \code{"ambiguous"}
#' (see details below).
#'
#' @param probe
#' If probes should be designed. \code{TRUE} or \code{FALSE},
#' defaults to \code{TRUE}.
#'
#' @param lengthProbe
#' Probe length range. A numeric vector [15, 40],
#' defaults to \code{c(18, 22)}.
#'
#' @param maxDegeneracyProbe
#' Maximum number of variants of each probe. A number [1, 64],
#' defaults to \code{4}.
#'
#' @param avoidFiveEndGProbe
#' If probes with G
#' at the 5' end should be avoided. \code{TRUE} or \code{FALSE},
#' defaults to \code{TRUE}.
#'
#' @param gcProbe
#' GC-content range for probes. A numeric vector [0, 1],
#' defaults to \code{c(0.40, 0.65)}.
#'
#' @param tmProbe
#' Tm range for probes (in Celcius degrees).
#' A numeric vector [30, 90], defaults to \code{c(55, 70)}.
#'
#' @param concProbe
#' Primer concentration in nM, for tm calculation. A numeric vector
#' [20, 2000], defaults to \code{250}.
#'
#' @param concNa
#' Sodium ion (equivalent) concentration in the PCR reaction (in M).
#' For calculation of tm and delta G.
#' A numeric vector [0.01, 1], defaults to \code{0.05} (50 mM).
#'
#' @return
#' An \code{RprimerOligo} object, containing the following information:
#'
#' \describe{
#'   \item{type}{Whether the oligo is a primer or probe.}
#'   \item{fwd}{\code{TRUE} if the oligo is valid in forward direction,
#'     \code{FALSE} otherwise.}
#'   \item{rev}{\code{TRUE} if the oligo is valid in reverse direction,
#'     \code{FALSE} otherwise.}
#'   \item{start}{Start position of the oligo.}
#'   \item{end}{End positon of the oligo.}
#'   \item{length}{Oligo length.}
#'   \item{iupacSequence}{Oligo sequence in IUPAC format
#'   (i.e. with ambiguous bases).}
#'   \item{iupaSequenceRc}{The reverse complement of the iupacSequence.}
#'   \item{identity}{For ambiguous oligos: average identity of the oligo.
#'     For mixed oligos: average identity of the 5' (consensus) part of the
#'     oligo. The value can range from 0 to 1.}
#'   \item{coverage}{For ambiguous oligos: average coverage of the oligo.
#'     For mixed oligos: average coverage of the 3' (degenerate) part of the
#'     oligo. The value can range from 0 to 1.}
#'   \item{degeneracy}{Number of sequence variants of the oligo.}
#'   \item{gcContentMean}{Mean GC-content of all sequence variants of the oligo.
#'   }
#'   \item{gcContent}{Range in GC-content of all sequence variants of
#'     the oligo.}
#'   \item{tmMean}{Mean tm of all sequence variants of the oligo
#'   (in Celcius degrees).}
#'   \item{tm}{Range in tm of all sequence variants of the oligo
#'   (in Celcius degrees).}
#'   \item{deltaGMean}{Mean delta G of all sequence variants of the oligo
#'   (in kcal/mol).}
#'   \item{deltaG}{Range in delta G of all sequence variants of the oligo
#'   (in kcal/mol).}
#'   \item{sequence}{All sequence variants of the oligo.}
#'   \item{sequenceRc}{Reverse complements of all sequence variants.}
#'   \item{gcContent}{GC-content of all sequence variants.}
#'   \item{tm}{Tm of all sequence variants (in Celcius degrees).}
#'   \item{deltaG}{Delta G of all sequence variants (in kcal/mol).}
#'   \item{method}{Design method used to generate the oligo: "ambiguous",
#'   "mixedFwd" or "mixedRev".}
#'   \item{score}{Oligo score, the lower the better.}
#'   \item{roiStart}{First position of the input \code{RprimerProfile} object
#'     (roi = region of interest).}
#'   \item{roiEnd}{Last position of the input \code{RprimerProfile} object.}
#' }
#'
#' An error message will return if no oligos are found. If so, a good idea
#' could be to re-run the design process with relaxed constraints.
#'
#' @details
#'
#' \strong{Valid oligos}
#'
#' For an oligo to be considered as valid, all sequence variants must fulfill
#' all the specified design constraints.
#'
#' Furthermore, oligos with at least one sequence variant containing
#' more than four consecutive runs
#' of the same
#' nucleotide (e.g. "AAAAA") and/or more than three consecutive runs
#' of the same di-nucleotide (e.g. "TATATATA") will be excluded
#' from consideration.
#'
#' \strong{Calculation of tm and delta G}
#'
#' Melting temperatures
#' are calculated for perfectly matching
#' DNA duplexes using the
#' nearest-neighbor
#' method (SantaLucia and Hicks, 2004), by using the following equation:
#'
#' \loadmathjax
#' \mjsdeqn{Tm = (\Delta H ^o \cdot 1000) / (\Delta S ^o + R \cdot \log [\mathrm{oligo}]) - 273.15}
#'
#' where \mjseqn{\Delta H ^o} is
#' the change in enthalpy (in cal/mol) and \mjseqn{\Delta S ^o} is the
#' change in entropy (in cal/K/mol) when an
#' oligo and a perfectly matching target sequence goes from random coil to
#' duplex formation.
#' \mjseqn{K} is the gas constant (1.9872 cal/mol \emph{K}).
#'
#' Delta G is calculated at 37 Celcius degrees, for when an oligo and a
#' perfectly matching target
#' sequence goes from random coil to duplex state, by using  the following
#' equation:
#'
#' \mjsdeqn{ \Delta G ^o _T = ( \Delta H ^o \cdot 1000 - T \cdot \Delta S ^o ) / 1000}{ASCII representation}
#
#' For both tm and delta G, the following salt correction method is used
#' for \mjseqn{ \Delta S^o }, as
#' described in SantaLucia and Hicks (2004):
#'
#' \mjsdeqn{ \Delta S^o [\mathrm{Na^+}] = \Delta S^o [\mathrm{1 M NaCl}] + 0.368 \cdot N / 2 \cdot \log [\mathrm{Na^+}] }
#'
#' where \mjseqn{N} is the total number of phosphates in the duplex, and [Na+] is
#' the total
#' concentration of monovalent cations.
#'
#' Nearest neighbor table values for \mjseqn{\Delta S^o} and \mjseqn{\Delta H^o} are
#' from SantaLucia and Hicks, 2004, and can be
#' retrieved calling \code{rprimer:::lookup$nn}.
#'
#' \strong{Primer design strategies}
#'
#' Primers can be generated by using one of the two following strategies:
#'
#' \itemize{
#' \item{The \strong{ambiguous strategy} (default) generates primers from the
#' IUPAC consensus sequence alone, which means that ambiguous bases can
#' occur at any position in the primer.}
#'
#' \item{The \strong{mixed strategy} generates primers from both the majority
#' and the IUPAC consensus sequence. These primers consist of a shorter
#' degenerate part at the 3' end (approx. 1/3 of the primer, targeting a
#' conserved
#' region) and a longer consensus part at the 5' end (approx.
#' 2/3 of the primer),
#' which instead of having ambiguous bases contains the most frequently occuring
#' nucleotide at each position.
#' This strategy resembles the widely-adopted Consensus-Degenerate Hybrid
#' Oligonucleotide Primer (CODEHOP) principle (Rose et al., 1998), and aims to
#' to allow amplification of highly variable targets using primers with
#' low degeneracy.
#' The idea is that the degenerate 3' end part will bind specifically to
#' the target sequence in the initial PCR cycles, and promote amplification
#' in spite of eventual mismatches at the 5' consensus part
#' (since 5' end mismatches are generally less detrimental than
#' 3' end mismatches). In this way, the generated products
#' will match the 5' ends of all primers perfectly, which allows them
#' to be efficiently amplified in later PCR cycles.
#' To provide a sufficiently high tm in spite of mismatches, it is
#' recommended to design relatively long primers (at least 25 bases) when using
#' this strategy.}
#' }
#'
#' Probes are always designed using the ambiguous strategy.
#'
#' \strong{Scoring system for oligos}
#'
#' All valid oligos are scored based on their identity, coverage and
#' average GC content. The scoring system is presented below.
#'
#' \strong{Identity and coverage}
#'
#' \tabular{lr}{
#' Value range \tab Score \cr
#' \eqn{(0.99, 1]} \tab 0 \cr
#' \eqn{(0.95, 0.99]} \tab 1 \cr
#' \eqn{(0.90, 0.95]} \tab 2 \cr
#' \eqn{\leq 0.90} \tab 3
#' }
#'
#' \strong{Average GC-content}
#'
#' This score is based on how much
#' the average GC-content deviates from 0.5 (in absolute value).
#'
#' \tabular{lr}{
#' Value range \tab Score \cr
#' \eqn{[0, 0.05)}
#' \tab 0 \cr
#' \eqn{[0.05, 0.1)}
#' \tab 1 \cr
#' \eqn{[0.1,  0.2)}
#' \tab 2 \cr
#' \eqn{\geq 0.2} \tab 3
#' }
#'
#' These scores are summarized. The weight of each individual score is 1, and
#' thus, the lowest and best
#' possible score for an oligo is 0, and the worst possible score is 9.
#'
#' @references
#' Rose, TM., Schultz ER., Henikoff JG., Pietrokovski S.,
#' McCallum CM., and Henikoff S. 1998. Consensus-Degenerate
#' Hybrid Oligonucleotide Primers for
#' Amplification of Distantly Related Sequences. Nucleic Acids Research 26 (7):
#' 1628-35.
#'
#' SantaLucia Jr, J., & Hicks, D. (2004).
#' The thermodynamics of DNA structural motifs.
#' Annu. Rev. Biophys. Biomol. Struct., 33, 415-440.
#'
#' @export
#'
#' @import mathjaxr
#'
#' @examples
#' data("exampleRprimerProfile")
#' x <- exampleRprimerProfile
#'
#' ## Design primers and probes with default values
#' designOligos(x)
designOligos <- function(x,
                         maxGapFrequency = 0.01,
                         lengthPrimer = c(18, 22),
                         maxDegeneracyPrimer = 4,
                         gcClampPrimer = TRUE,
                         avoidThreeEndRunsPrimer = TRUE,
                         gcPrimer = c(0.40, 0.65),
                         tmPrimer = c(50, 65),
                         concPrimer = 500,
                         designStrategyPrimer = "ambiguous",
                         probe = TRUE,
                         lengthProbe = c(18, 22),
                         maxDegeneracyProbe = 4,
                         avoidFiveEndGProbe = TRUE,
                         gcProbe = c(0.40, 0.65),
                         tmProbe = c(50, 70),
                         concProbe = 250,
                         concNa = 0.05) {
    if (!methods::is(x, "RprimerProfile")) {
        stop("'x' must be an RprimerProfile object.", call. = FALSE)
    }
    if (!(maxGapFrequency >= 0 && maxGapFrequency <= 1)) {
        stop("'maxGapFrequency' must be from 0 to 1.", call. = FALSE)
    }
    if (!(min(lengthPrimer) >= 15 && max(lengthPrimer) <= 40)) {
        stop("'lengthPrimer' must be from 15 to 40.", call. = FALSE)
    }
    if (!(maxDegeneracyPrimer >= 1 && maxDegeneracyPrimer <= 64)) {
        stop("'maxDegeneracyPrimer' must be from 1 to 64.", call. = FALSE)
    }
    if (!is.logical(gcClampPrimer)) {
        stop("'gcClampPrimer' must be TRUE or FALSE", call. = FALSE)
    }
    if (!is.logical(avoidThreeEndRunsPrimer)) {
        stop("'avoidThreeEndRunsPrimer' must be TRUE or FALSE", call. = FALSE)
    }
    if (!(min(gcPrimer) >= 0 && max(gcPrimer) <= 1)) {
        stop(
            "'gcPrimer' must be from 0 to 1, e.g. c(0.45, 0.65).",
            call. = FALSE
        )
    }
    if (!(min(tmPrimer) >= 20 && max(tmPrimer) <= 90)) {
        stop(
            "'tmPrimer' must be from 20 to 90, e.g. c(55, 60).",
            call. = FALSE
        )
    }
    if (!(concPrimer >= 20 && concPrimer <= 2000)) {
        stop("'concPrimer' must be from 20 to 2000.", call. = FALSE)
    }
    if (!(
        designStrategyPrimer == "ambiguous" || designStrategyPrimer == "mixed")
    ) {
        stop(
            "'designStrategyPrimer' must be either 'ambiguous' or 'mixed'.",
            call. = FALSE
        )
    }
    if (!is.logical(probe)) {
        stop("'probe' must be TRUE or FALSE", call. = FALSE)
    }
    if (!(min(lengthProbe) >= 15 && max(lengthProbe) <= 40)) {
        stop("'lengthProbe' must be from 15 to 40.", call. = FALSE)
    }
    if (!(maxDegeneracyProbe >= 1 && maxDegeneracyProbe <= 64)) {
        stop("'maxDegeneracyProbe' must be from 1 to 64.", call. = FALSE)
    }
    if (!is.logical(avoidFiveEndGProbe)) {
        stop("'avoidFiveEndGProbe' must be TRUE or FALSE", call. = FALSE)
    }
    if (!(min(gcProbe) >= 0 && max(gcProbe) <= 1)) {
        stop(
            "'gcProbe' must be from 0 to 1, e.g. c(0.45, 0.65).",
            call. = FALSE
        )
    }
    if (!(min(tmProbe) >= 20 && max(tmProbe) <= 90)) {
        stop(
            "'tmProbe' must be from 20 to 90, e.g. c(55, 60).",
            call. = FALSE
        )
    }
    if (!(concProbe >= 20 && concProbe <= 2000)) {
        stop("'concProbe' must be from 20 to 2000.", call. = FALSE)
    }
    if (!(concNa >= 0.01 && concNa <= 1)) {
        stop("'concNa' must be from 0.01 to 1.", call. = FALSE)
    }
    lengthPrimer <- seq(min(lengthPrimer), max(lengthPrimer))
    lengthProbe <- seq(min(lengthProbe), max(lengthProbe))
    if (designStrategyPrimer == "mixed") {
        primers <- .designMixedPrimers(
            x,
            maxGapFrequency = maxGapFrequency,
            lengthPrimer = lengthPrimer,
            maxDegeneracyPrimer = maxDegeneracyPrimer,
            gcClampPrimer = gcClampPrimer,
            avoidThreeEndRunsPrimer = avoidThreeEndRunsPrimer,
            gcPrimer = gcPrimer,
            tmPrimer = tmPrimer,
            concPrimer = concPrimer,
            concNa = concNa
        )
        if (nrow(primers) == 0L) {
            stop("No primers were found.", call. = FALSE)
        }
        oligos <- primers
        if (probe) {
            probes <- .designAmbiguousOligos(
                x,
                primer = FALSE,
                lengthProbe = lengthProbe,
                maxDegeneracyProbe = maxDegeneracyProbe,
                avoidFiveEndGProbe = avoidFiveEndGProbe,
                gcProbe = gcProbe,
                tmProbe = tmProbe,
                concProbe = concProbe,
                concNa = concNa
            )
            if (nrow(probes) == 0L) {
                stop("No probes were found.", call. = FALSE)
            }
            oligos <- rbind(oligos, probes)
        }
    } else {
        oligos <- .designAmbiguousOligos(
            x,
            maxGapFrequency = maxGapFrequency,
            primer = TRUE,
            lengthPrimer = lengthPrimer,
            maxDegeneracyPrimer = maxDegeneracyPrimer,
            gcClampPrimer = gcClampPrimer,
            avoidThreeEndRunsPrimer = avoidThreeEndRunsPrimer,
            gcPrimer = gcPrimer,
            tmPrimer = tmPrimer,
            concPrimer = concPrimer,
            probe = probe,
            lengthProbe = lengthProbe,
            maxDegeneracyProbe = maxDegeneracyProbe,
            avoidFiveEndGProbe = avoidFiveEndGProbe,
            gcProbe = gcProbe,
            tmProbe = tmProbe,
            concProbe = concProbe,
            concNa = concNa
        )
        if (nrow(oligos[oligos$type == "primer", ]) == 0L) {
            stop("No primers were found.", call. = FALSE)
        }
        if (probe) {
            if (nrow(oligos[oligos$type == "probe", ]) == 0L) {
                stop("No probes were found.", call. = FALSE)
            }
        }
    }
    oligos <- .scoreOligos(oligos)
    oligos <- .beautifyOligos(oligos)
    RprimerOligo(oligos)
}

# Helpers ======================================================================

#' @noRd
#'
#' @examples
#' .nmers(c("A", "G", "T", "T", "C", "G"), n = 4)
.nmers <- function(x, n) {
    start <- seq_len(length(x) - n + 1)
    end <- start + n - 1
    nmers <- lapply(start, \(i) x[start[[i]]:end[[i]]])
    do.call("rbind", nmers)
}

#' @noRd
#'
#' @examples
#' .countDegeneracy(c("A", "R", "T", "T", "N", "G"))
.countDegeneracy <- function(x) prod(lookup$degeneracy[x])

#' @noRd
#'
#' @examples
#' data("exampleRprimerProfile")
#' .generateAmbiguousOligos(exampleRprimerProfile, lengthOligo = 18)
.generateAmbiguousOligos <- function(x, lengthOligo = 20) {
    oligos <- list()
    oligos$iupacSequence <- .nmers(x$iupac, lengthOligo)
    oligos$start <- seq_len(nrow(oligos$iupacSequence)) + min(x$position) - 1
    oligos$end <- seq_len(
        nrow(oligos$iupacSequence)
    ) + lengthOligo - 1 + min(x$position) - 1
    oligos$length <- rep(lengthOligo, nrow(oligos$iupacSequence))
    oligos$degeneracy <- apply(oligos$iupacSequence, 1, .countDegeneracy)
    oligos$gapFrequency <- apply(.nmers(x$gaps, lengthOligo), 1, max)
    oligos$coverage <- .nmers(x$coverage, lengthOligo) |> rowMeans()
    oligos$identity <- .nmers(x$identity, lengthOligo) |> rowMeans()
    oligos$method <- rep("ambiguous", nrow(oligos$iupacSequence))
    oligos$roiStart <- rep(
        min(x$position, na.rm = TRUE), nrow(oligos$iupacSequence)
    )
    oligos$roiEnd <- rep(
        max(x$position, na.rm = TRUE), nrow(oligos$iupacSequence)
    )
    oligos
}

#' @noRd
#'
#' @examples
#' .splitAndPaste(t(matrix(rep(1, 10))), t(matrix(rep(2, 10))))
#' .splitAndPaste(t(matrix(rep(1, 10))), t(matrix(rep(2, 10))), combine = FALSE)
#' .splitAndPaste(t(matrix(rep(1, 10))), t(matrix(rep(2, 10))), rev = TRUE)
.splitAndPaste <- function(first, second, rev = FALSE, combine = TRUE) {
    n <- ncol(first)
    small <- seq_len(as.integer(n / 3))
    large <- seq(small[length(small)] + 1, n)
    if (rev) {
        first <- first[, small, drop = FALSE]
        second <- second[, large, drop = FALSE]
    } else {
        first <- first[, seq_along(large), drop = FALSE]
        second <- second[, small + n - length(small), drop = FALSE]
    }
    if (combine) {
        cbind(first, second)
    } else {
        list(first, second)
    }
}

#' @noRd
#'
#' @examples
#' data("exampleRprimerProfile")
#' .mixFwd(exampleRprimerProfile, 20)
.mixFwd <- function(x, lengthOligo = 20) {
    oligos <- list()
    majority <- .nmers(x$majority, lengthOligo)
    iupac <- .nmers(x$iupac, lengthOligo)
    oligos$iupacSequence <- .splitAndPaste(majority, iupac)
    oligos$start <- seq_len(nrow(oligos$iupacSequence)) + min(x$position) - 1
    oligos$end <- seq_len(
        nrow(oligos$iupacSequence)
    ) + lengthOligo - 1 + min(x$position) - 1
    oligos$length <- rep(lengthOligo, nrow(oligos$iupacSequence))
    oligos$degeneracy <- apply(oligos$iupacSequence, 1, .countDegeneracy)
    oligos$gapFrequency <- apply(.nmers(x$gaps, lengthOligo), 1, max)
    identityCoverage <- .splitAndPaste(
        .nmers(x$identity, lengthOligo), .nmers(x$coverage, lengthOligo),
        combine = FALSE
    )
    oligos$identity <- identityCoverage[[1]] |> rowMeans()
    oligos$coverage <- identityCoverage[[2]] |> rowMeans()
    oligos$method <- rep("mixedFwd", nrow(oligos$iupacSequence))
    oligos$roiStart <- rep(
        min(x$position, na.rm = TRUE), nrow(oligos$iupacSequence)
    )
    oligos$roiEnd <- rep(
        max(x$position, na.rm = TRUE), nrow(oligos$iupacSequence)
    )
    oligos
}

#' @noRd
#'
#' @examples
#' data("exampleRprimerProfile")
#' .mixRev(exampleRprimerProfile, 20)
.mixRev <- function(x, lengthOligo = 20) {
    oligos <- list()
    majority <- .nmers(x$majority, lengthOligo)
    iupac <- .nmers(x$iupac, lengthOligo)
    oligos$iupacSequence <- .splitAndPaste(iupac, majority, rev = TRUE)
    oligos$start <- seq_len(nrow(oligos$iupacSequence)) + min(x$position) - 1
    oligos$end <- seq_len(
        nrow(oligos$iupacSequence)
    ) + lengthOligo - 1 + min(x$position) - 1
    oligos$length <- rep(lengthOligo, nrow(oligos$iupacSequence))
    oligos$degeneracy <- apply(oligos$iupacSequence, 1, .countDegeneracy)
    oligos$gapFrequency <- apply(.nmers(x$gaps, lengthOligo), 1, max)
    coverageIdentity <- .splitAndPaste(
        .nmers(x$coverage, lengthOligo), .nmers(x$identity, lengthOligo),
        combine = FALSE, rev = TRUE
    )
    oligos$identity <- coverageIdentity[[2]] |> rowMeans()
    oligos$coverage <- coverageIdentity[[1]] |> rowMeans()
    oligos$method <- rep("mixedRev", nrow(oligos$iupacSequence))
    oligos$roiStart <- rep(
        min(x$position, na.rm = TRUE), nrow(oligos$iupacSequence)
    )
    oligos$roiEnd <- rep(
        max(x$position, na.rm = TRUE), nrow(oligos$iupacSequence)
    )
    oligos
}


#' @noRd
#'
#' @examples
#' data("exampleRprimerProfile")
#' fwd <- .mixFwd(exampleRprimerProfile)
#' rev <- .mixRev(exampleRprimerProfile)
#' .mergeLists(fwd, rev)
.mergeLists <- function(first, second) {
    x <- lapply(names(first), \(i) {
        if (is.matrix(first[[i]])) {
            rbind(first[[i]], second[[i]])
        } else {
            c(first[[i]], second[[i]])
        }
    })
    names(x) <- names(first)
    x
}

#' @noRd
#'
#' @examples
#' data("exampleRprimerProfile")
#' .generateMixedOligos(exampleRprimerProfile, lengthOligo = 20)
.generateMixedOligos <- function(x, lengthOligo = 20) {
    fwd <- .mixFwd(x, lengthOligo)
    rev <- .mixRev(x, lengthOligo)
    .mergeLists(fwd, rev)
}

#' @noRd
#'
#' @examples
#' data("exampleRprimerProfile")
#' x <- .generateMixedOligos(exampleRprimerProfile)
#' .filterOligos(x)
.filterOligos <- function(x, maxGapFrequency = 0.1, maxDegeneracy = 4) {
    invalidCharacters <- apply(x$iupacSequence, 1, \(x) {
        any(x == "-") | any(is.na(x))
    })
    invalid <- unique(c(
        which(x$degeneracy > maxDegeneracy),
        which(x$gapFrequency > maxGapFrequency),
        which(invalidCharacters)
    ))
    if (length(invalid) > 0L) {
        x <- lapply(x, \(x) {
            if (is.matrix(x)) x[-invalid, , drop = FALSE] else x[-invalid]
        })
    }
    x
}

#' @noRd
#'
#' @examples
#' .expandDegenerates(c("A", "R", "T", "T", "N", "G"))
.expandDegenerates <- function(x) {
    bases <- lapply(x, \(i) {
        degen <- unname(lookup$degenerates[[i]])
        unlist(strsplit(degen, split = ","))
    })
    all <- expand.grid(bases[seq_along(bases)], stringsAsFactors = FALSE)
    all <- as.matrix(all)
    colnames(all) <- NULL
    all
}

#' @noRd
#'
#' @examples
#' data("exampleRprimerProfile")
#' x <- .filterOligos(.generateMixedOligos(exampleRprimerProfile))
#' x$sequence <- apply(x$iupacSequence, 1, .expandDegenerates)
#' .oligoMatrix(x$sequence)
.oligoMatrix <- function(x) {
    degeneracy <- vapply(x, nrow, integer(1L))
    id <- lapply(seq_along(degeneracy), \(x) rep(x, degeneracy[[x]]))
    id <- unlist(id)
    x <- do.call("rbind", x)
    rownames(x) <- id
    x
}

#' @noRd
#'
#' @examples
#' .reverseComplement(matrix(c("A", "R", "T", "T", "N", "G")))
.reverseComplement <- function(x) {
    rc <- x[, rev(seq_len(ncol(x))), drop = FALSE]
    rc[] <- lookup$complement[rc]
    rc
}

#' @noRd
#'
#' @examples
#' seq <- matrix(c(1, 1, 0, 0, 1, 0, 1, 0, 0, 0, 0))
#' .gcClamp(gc, rev = FALSE)
#' .gcClamp(gc, rev = TRUE)
.gcClamp <- function(x, rev = FALSE) {
    if (rev) {
        end <- x[, seq_len(5), drop = FALSE]
        5 - rowSums(end) >= 2 & 5 - rowSums(end) <= 3 ## Because of complement..
    } else {
        end <- x[, seq((ncol(x) - 4), ncol(x)), drop = FALSE]
        rowSums(end) >= 2 & rowSums(end) <= 3
    }
}

#' @noRd
#'
#' @examples
#' seq <- matrix(c("A", "C", "G", "G", "T", "T", "A", "A"))
#' .endRuns(seq, rev = FALSE)
.endRuns <- function(x, rev = FALSE) {
    if (rev) {
        end <- x[, seq_len(3), drop = FALSE]
    } else {
        end <- x[, seq((ncol(x) - 2), ncol(x)), drop = FALSE]
    }
    apply(end, 1, \(x) {
        all(x == "A") | all(x == "C") | all(x == "T") | all(x == "G")
    })
}

#' @noRd
#'
#' @examples
#' .repeats(c("ACTTTTCT", "ACTTTTTCT", "ATCTCTCTCA"))
.repeats <- function(x) {
    di <- "(AT){4,}|(TA){4,}|(AC){4,}|(CA){4,}|(AG){4,}|(GA){4,}|(GT){4,}|(TG){4,}|(CG){4,}|(GC){4,}|(CT){4,}|(TC){4,}|)"
    mono <- "([A-Z])\\1\\1\\1\\1"
    vapply(x, \(y) {
        grepl(di, y) | grepl(mono, y)
    }, logical(1L))
}

#' @noRd
#'
#' @examples
#' data("exampleRprimerProfile")
#' x <- .filterOligos(.generateMixedOligos(exampleRprimerProfile))
#' .allVariants(x)
.allVariants <- function(x,
                         concPrimer = 500,
                         concProbe = 250,
                         concNa = 0.05) {
    all <- list()
    all$sequence <- apply(x$iupacSequence, 1, .expandDegenerates, simplify = FALSE)
    ## If there is only one variant of each oligo,
    ## (and apply returns a matrix instead of a list):
    if (is.matrix(all$sequence)) {
        all$sequence <- t(all$sequence)
        all$sequence <- lapply(seq_len(nrow(all$sequence)), \(i) {
            all$sequence[i, , drop = FALSE]
        })
    }
    all$sequence <- .oligoMatrix(all$sequence)
    all$sequenceRc <- .reverseComplement(all$sequence)
    gc <- all$sequence == "C" | all$sequence == "G"
    n <- rowSums(
        all$sequence == "A" | all$sequence == "C" |
            all$sequence == "G" | all$sequence == "T"
    )
    all$gcContent <- rowSums(gc) / n
    all$gcClampFwd <- .gcClamp(gc)
    all$gcClampRev <- .gcClamp(gc, rev = TRUE)
    all$threeEndRunsFwd <- .endRuns(all$sequence)
    all$threeEndRunsRev <- .endRuns(all$sequence, rev = TRUE)
    all$fiveEndGPlus <- all$sequence[, 1] == "G"
    all$fiveEndGMinus <- all$sequence[, ncol(all$sequence)] == "C"
    tmParam <- .tmParameters(all$sequence, concNa)
    all$tmPrimer <- .tm(tmParam, concPrimer)
    all$tmProbe <- .tm(tmParam, concProbe)
    all$deltaG <- .deltaG(tmParam)
    all$sequence <- apply(all$sequence, 1, paste, collapse = "")
    all$sequenceRc <- apply(all$sequenceRc, 1, paste, collapse = "")
    all$repeats <- .repeats(all$sequence)
    lapply(all, \(x) unname(split(unname(x), f = as.integer(names(x)))))
}

#' @noRd
#'
#' @examples
#' data("exampleRprimerProfile")
#' x <- .allVariants(.filterOligos(.generateMixedOligos(exampleRprimerProfile)))
#' .meanRange(x)
.meanRange <- function(x) {
    x <- x[c("gcContent", "tmPrimer", "tmProbe", "deltaG")]
    means <- lapply(x, \(y) {
        vapply(y, \(z) {
            sum(z) / length(z)
        }, double(1L))
    })
    means <- do.call("cbind.data.frame", means)
    names(means) <- paste0(names(means), "Mean")
    ranges <- lapply(x, \(y) {
        vapply(y, \(z) {
            max(z) - min(z)
        }, double(1L))
    })
    ranges <- do.call("cbind.data.frame", ranges)
    names(ranges) <- paste0(names(ranges), "Range")
    cbind(means, ranges)
}

#' @noRd
#'
#' @examples
#' data("exampleRprimerProfile")
#' x <- .filterOligos(.generateMixedOligos(exampleRprimerProfile))
#' .makeOligoDf(x)
.makeOligoDf <- function(x) {
    x <- within(x, rm("gapFrequency"))
    x$iupacSequenceRc <- .reverseComplement(x$iupacSequence)
    x$iupacSequence <- apply(x$iupacSequence, 1, paste, collapse = "")
    x$iupacSequenceRc <- apply(x$iupacSequenceRc, 1, paste, collapse = "")
    do.call("cbind.data.frame", x)
}

#' @noRd
#'
#' @examples
#' data("exampleRprimerProfile")
#' x <- .allVariants(.filterOligos(.generateMixedOligos(exampleRprimerProfile)))
#' .isWithinRange(x$gcContent, c(0.4, 0.6))
.isWithinRange <- function(x, range) {
    lapply(x, \(y) y >= min(range) & y <= max(range))
}

#' @noRd
#'
#' @examples
#' data("exampleRprimerProfile")
#' x <- .allVariants(.filterOligos(.generateMixedOligos(exampleRprimerProfile)))
#' gcInRange <- .isWithinRange(x$gcContent, c(0.4, 0.6))
#' x <- data.frame(cbind(gcInRange))
#' .convertToMatrices(x["gcInRange"])
.convertToMatrices <- function(x) {
    lapply(seq_len(nrow(x)), \(i) {
        y <- lapply(x[i, , drop = FALSE], unlist)
        do.call("cbind", y)
    })
}

#' Column means represent the proportion
#' of sequence variants that fulfill a specific criteria (e.g. GC clamp),
#' and row means represent the proportion of the desired design criteria that
#' are fulfilled by specific sequence variants.
#'
#' @noRd
#'
#' @examples
#' data("exampleRprimerProfile")
#' x <- .allVariants(.filterOligos(.generateMixedOligos(exampleRprimerProfile)))
#' gcInRange <- .isWithinRange(x$gcContent, c(0.4, 0.6))
#' x <- data.frame(cbind(gcInRange))
#' check <- .convertToMatrices(x["gcInRange"])
#' .isValid(check, rowThreshold = 0.5, colThreshold = 0.5)
.isValid <- function(x, rowThreshold = 1, colThreshold = 1) {
    valid <- vapply(x, \(y) {
        toInvert <- c(
            "repeats", "threeEndRunsFwd", "threeEndRunsRev",
            "fiveEndGPlus", "fiveEndGMinus"
        )
        select <- colnames(y) %in% toInvert
        y[, select] <- !y[, select]
        col <- colMeans(y)
        row <- rowMeans(y)
        all(col >= colThreshold) & all(row >= rowThreshold)
    }, logical(1L))
    valid
}

#' @noRd
.checkPrimers <- function(x,
                          gcClampPrimer = TRUE,
                          avoidThreeEndRunsPrimer = TRUE) {
    selectFwd <- c("repeats", "tmInRange", "gcInRange")
    selectRev <- c("repeats", "tmInRange", "gcInRange")
    if (gcClampPrimer) {
        selectFwd <- c(selectFwd, "gcClampFwd")
        selectRev <- c(selectRev, "gcClampRev")
    }
    if (avoidThreeEndRunsPrimer) {
        selectFwd <- c(selectFwd, "threeEndRunsFwd")
        selectRev <- c(selectRev, "threeEndRunsRev")
    }
    fwdPrimers <- x[selectFwd]
    fwdPrimers <- .convertToMatrices(fwdPrimers)
    fwd <- .isValid(fwdPrimers)
    revPrimers <- x[selectRev]
    revPrimers <- .convertToMatrices(revPrimers)
    rev <- .isValid(revPrimers)
    x <- cbind(x, fwd, rev)
    x[x$fwd | x$rev, , drop = FALSE]
}

#' @noRd
.filterPrimers <- function(x,
                           lengthPrimer = 18:22,
                           maxDegeneracyPrimer = 4,
                           gcClampPrimer = TRUE,
                           avoidThreeEndRunsPrimer = TRUE,
                           gcPrimer = c(0.45, 0.55),
                           tmPrimer = c(55, 65)) {
    x <- x[x$length %in% lengthPrimer, , drop = FALSE]
    x <- x[x$degeneracy <= maxDegeneracyPrimer, , drop = FALSE]
    gcInRange <- .isWithinRange(x$gcContent, gcPrimer)
    tmInRange <- .isWithinRange(x$tmPrimer, tmPrimer)
    x <- cbind(x, data.frame(cbind(tmInRange, gcInRange)))
    x <- .checkPrimers(
        x,
        gcClampPrimer,
        avoidThreeEndRunsPrimer
    )
    x$rev[x$method == "mixedFwd" & x$rev] <- FALSE
    x$fwd[x$method == "mixedRev" & x$fwd] <- FALSE
    x <- x[x$fwd | x$rev, , drop = FALSE]
    remove <- c(
        "gcInRange", "tmInRange",
        "okFwd", "okRev", "tmProbeMean", "tmProbeRange", "tmProbe"
    )
    x <- x[!names(x) %in% remove]
    oldnames <- c("tmPrimerMean", "tmPrimerRange", "tmPrimer")
    newnames <- c("tmMean", "tmRange", "tm")
    names(x)[names(x) %in% oldnames] <- newnames
    type <- rep("primer", nrow(x))
    cbind(type, x)
}

#' @noRd
#'
#' @examples
#' data("exampleRprimerProfile")
#' .designMixedPrimers(exampleRprimerProfile)
.designMixedPrimers <- function(x,
                                maxGapFrequency = 0.01,
                                lengthPrimer = 18:22,
                                maxDegeneracyPrimer = 4,
                                gcClampPrimer = TRUE,
                                avoidThreeEndRunsPrimer = TRUE,
                                gcPrimer = c(0.45, 0.55),
                                tmPrimer = c(55, 65),
                                concPrimer = 500,
                                concNa = 0.05) {
    lengthPrimer <- lengthPrimer[order(lengthPrimer)]
    all <- lapply(lengthPrimer, \(i) {
        mixed <- .generateMixedOligos(x, lengthOligo = i)
        mixed <- .filterOligos(
            mixed, maxGapFrequency, maxDegeneracyPrimer
        )
        if (length(mixed[[1]] > 0L)) {
            allVariants <- .allVariants(
                mixed, concPrimer,
                concProbe = NA, concNa
            )
            meansAndRanges <- .meanRange(allVariants)
            allVariants <- data.frame(do.call("cbind", allVariants))
            mixed <- .makeOligoDf(mixed)
            cbind(mixed, meansAndRanges, allVariants)
        } else {
            NULL
        }
    })
    all <- do.call("rbind", all)
    if (is.null(all)) {
        stop("No primers were found.", call. = FALSE)
    }
    all <- .filterPrimers(
        all,
        lengthPrimer,
        maxDegeneracyPrimer,
        gcClampPrimer,
        avoidThreeEndRunsPrimer,
        gcPrimer,
        tmPrimer
    )
    all
}

#' @noRd
.checkProbes <- function(x,
                         avoidFiveEndGProbe) {
    selectFwd <- c("repeats", "tmInRange", "gcInRange")
    selectRev <- c("repeats", "tmInRange", "gcInRange")
    if (avoidFiveEndGProbe) {
        selectFwd <- c(selectFwd, "fiveEndGPlus")
        selectRev <- c(selectRev, "fiveEndGMinus")
    }
    fwdProbes <- x[selectFwd]
    fwdProbes <- .convertToMatrices(fwdProbes)
    fwd <- .isValid(fwdProbes)
    revProbes <- x[selectRev]
    revProbes <- .convertToMatrices(revProbes)
    rev <- .isValid(revProbes)
    x <- cbind(x, fwd, rev)
    x[x$fwd | x$rev, , drop = FALSE]
}

#' @noRd
.filterProbes <- function(x,
                          lengthProbe = 18:22,
                          maxDegeneracyProbe = 4,
                          avoidFiveEndGProbe = TRUE,
                          gcProbe = c(0.45, 0.55),
                          tmProbe = c(55, 65)) {
    x <- x[x$length %in% lengthProbe, , drop = FALSE]
    x <- x[x$degeneracy <= maxDegeneracyProbe, , drop = FALSE]
    gcInRange <- .isWithinRange(x$gcContent, gcProbe)
    tmInRange <- .isWithinRange(x$tmProbe, tmProbe)
    x <- cbind(x, data.frame(cbind(tmInRange, gcInRange)))
    x <- .checkProbes(
        x,
        avoidFiveEndGProbe
    )
    remove <- c(
        "gcInRange", "tmInRange", "tmPrimerMean",
        "tmPrimerRange", "tmPrimer"
    )
    x <- x[!names(x) %in% remove]
    oldnames <- c("tmProbeMean", "tmProbeRange", "tmProbe")
    newnames <- c("tmMean", "tmRange", "tm")
    names(x)[names(x) %in% oldnames] <- newnames
    type <- rep("probe", nrow(x))
    cbind(type, x)
}

#' @noRd
#'
#' @examples
#' data("exampleRprimerProfile")
#' .designAmbiguousOligos(exampleRprimerProfile)
.designAmbiguousOligos <- function(x,
                                   maxGapFrequency = 0.01,
                                   primer = TRUE,
                                   lengthPrimer = 18:22,
                                   maxDegeneracyPrimer = 4,
                                   gcClampPrimer = TRUE,
                                   avoidThreeEndRunsPrimer = TRUE,
                                   gcPrimer = c(0.40, 0.65),
                                   tmPrimer = c(50, 65),
                                   concPrimer = 500,
                                   probe = TRUE,
                                   lengthProbe = 18:22,
                                   maxDegeneracyProbe = 4,
                                   avoidFiveEndGProbe = TRUE,
                                   gcProbe = c(0.40, 0.65),
                                   tmProbe = c(50, 70),
                                   concProbe = 250,
                                   concNa = 0.05) {
    if (probe && primer) {
        lengthOligo <- unique(c(lengthPrimer, lengthProbe))
        maxDegeneracy <- max(c(maxDegeneracyPrimer, maxDegeneracyProbe))
    } else if (!probe && primer) {
        lengthOligo <- lengthPrimer
        maxDegeneracy <- maxDegeneracyPrimer
    } else if (probe && !primer) {
        lengthOligo <- lengthProbe
        maxDegeneracy <- maxDegeneracyProbe
    }
    lengthOligo <- lengthOligo[order(lengthOligo)]
    ambiguous <- lapply(lengthOligo, \(i) {
        amb <- .generateAmbiguousOligos(x, lengthOligo = i)
        amb <- .filterOligos(
            amb,
            maxGapFrequency = maxGapFrequency,
            maxDegeneracy = maxDegeneracy
        )
        if (length(amb[[1]] > 0L)) {
            allVariants <- .allVariants(
                amb,
                concPrimer = concPrimer,
                concProbe = concProbe,
                concNa = concNa
            )
            meanAndRange <- .meanRange(allVariants)
            allVariants <- data.frame(do.call("cbind", allVariants))
            amb <- .makeOligoDf(amb)
            cbind(amb, meanAndRange, allVariants)
        } else {
            NULL
        }
    })
    ambiguous <- do.call("rbind", ambiguous)
    if (is.null(ambiguous)) {
        stop("No primers were found.", call. = FALSE)
    }
    if (primer) {
        primers <- .filterPrimers(
            ambiguous,
            lengthPrimer,
            maxDegeneracyPrimer,
            gcClampPrimer,
            avoidThreeEndRunsPrimer,
            gcPrimer,
            tmPrimer
        )
    } else {
        primers <- NULL
    }
    if (probe) {
        probes <- .filterProbes(
            ambiguous,
            lengthProbe,
            maxDegeneracyProbe,
            avoidFiveEndGProbe,
            gcProbe,
            tmProbe
        )
    } else {
        probes <- NULL
    }
    rbind(primers, probes)
}

#' @noRd
#'
#' @examples
#' data("exampleRprimerOligo")
#' x <- head(exampleRprimerOligo$identity)
#' .scoreIdentityCoverage(x)
.scoreIdentityCoverage <- function(x) {
    score <- vector(mode = "double", length = length(x))
    score[x <= 1 & x > 0.99] <- 0
    score[x <= 0.99 & x > 0.95] <- 1
    score[x <= 0.95 & x > 0.90] <- 2
    score[x <= 0.90] <- 3
    score
}

#' @noRd
#'
#' @examples
#' data("exampleRprimerOligo")
#' x <- head(exampleRprimerOligo$gcContentMean)
#' .scoreGcContent(x)
.scoreGcContent <- function(x) {
    deviation <- abs(x - 0.5)
    score <- vector(mode = "double", length = length(deviation))
    score[deviation >= 0 & deviation < 0.05] <- 0
    score[deviation >= 0.05 & deviation < 0.1] <- 1
    score[deviation >= 0.1 & deviation < 0.2] <- 2
    score[deviation >= 0.2] <- 3
    score
}

#' @noRd
#'
#' @examples
#' data("exampleRprimerOligo")
#' x <- head(exampleRprimerOligo)
#' .scoreOligos(x)
.scoreOligos <- function(x) {
    score <- list()
    score$identity <- .scoreIdentityCoverage(x$identity)
    score$coverage <- .scoreIdentityCoverage(x$coverage)
    score$gcContent <- .scoreGcContent(x$gcContentMean)
    score <- do.call("cbind", score)
    score <- rowSums(score)
    cbind(x, score)
}

#' @noRd
.beautifyOligos <- function(x) {
    keep <- c(
        "type", "fwd", "rev", "start", "end", "length",
        "iupacSequence", "iupacSequenceRc", "identity",
        "coverage", "degeneracy", "gcContentMean", "gcContentRange",
        "tmMean", "tmRange", "deltaGMean", "deltaGRange", "sequence",
        "sequenceRc", "gcContent", "tm", "deltaG", "method", "score",
        "roiStart", "roiEnd"
    )
    x <- x[keep]
    x <- x[order(x$start), ]
    rownames(x) <- NULL
    x
}

## Import data to test on
data("exampleRprimerProfile")
x <- exampleRprimerProfile

# oligos =======================================================================

test_that("designOligos returns an error when it should", {
    expect_error(designOligos(unclass(x)))
    expect_error(designOligos(x, maxGapFrequency = -1))
    expect_error(designOligos(x, maxGapFrequency = 1.1))
    expect_error(designOligos(x, maxGapFrequency = -1))
    expect_error(designOligos(x, lengthPrimer = 14))
    expect_error(designOligos(x, lengthPrimer = 41))
    expect_error(designOligos(x, maxDegeneracyPrimer = 65))
    expect_error(designOligos(x, maxDegeneracyPrimer = 0))
    expect_error(designOligos(x, gcClampPrimer = 1))
    expect_error(designOligos(x, avoidThreeEndRunsPrimer = "true"))
    expect_error(designOligos(x, gcPrimer = c(-0.1, 1)))
    expect_error(designOligos(x, gcPrimer = c(0, 1.1)))
    expect_error(designOligos(x, tmPrimer = c(19, 90)))
    expect_error(designOligos(x, tmPrimer = c(20, 91)))
    expect_error(designOligos(x, concPrimer = 19))
    expect_error(designOligos(x, concPrimer = 2001))
    expect_error(designOligos(x, designStrategyPrimer = ""))
    expect_error(designOligos(x, probe = 1))
    expect_error(designOligos(x, lengthProbe = 14))
    expect_error(designOligos(x, lengthProbe = 41))
    expect_error(designOligos(x, maxDegeneracyProbe = 0))
    expect_error(designOligos(x, maxDegeneracyProbe = 65))
    expect_error(designOligos(x, avoidFiveEndGProbe = 1))
    expect_error(designOligos(x, gcProbe = c(-0.1, 1)))
    expect_error(designOligos(x, gcProbe = c(0, 1.1)))
    expect_error(designOligos(x, tmProbe = c(19, 90)))
    expect_error(designOligos(x, tmProbe = c(20, 91)))
    expect_error(designOligos(x, concProbe = 19))
    expect_error(designOligos(x, concProbe = 2001))
    expect_error(designOligos(x, concNa = 0.009))
    expect_error(designOligos(x, concNa = 1.1))
})

test_that("oligos works", {
    test <- designOligos(x)
    expect_s4_class(test, "RprimerOligo")
    test <- designOligos(x, probe = FALSE)
    expect_true(all(test$type == "primer"))
    test <- designOligos(
        x,
        designStrategyPrimer = "mixed",
        maxDegeneracyPrimer = 1,
        maxDegeneracyProbe = 2,
        lengthPrimer = 25:30,
        lengthProbe = 18
    )
    primers <- test[test$type == "primer", ]
    expect_true(all(primers$method[primers$fwd] == "mixedFwd"))
    expect_true(all(primers$method[primers$rev] == "mixedRev"))
    expect_true(max(primers$degeneracy) == 1)
    expect_true(all(primers$length >= 25 & primers$length <= 30))
    probes <- test[test$type == "probe", ]
    expect_true(all(probes$method == "ambiguous"))
    expect_true(all(probes$length == 18))
    expect_true(max(probes$degeneracy) == 2)
    expect_error(designOligos(x[1:100, ]))
    expect_error(designOligos(x[1:1000, ], maxDegeneracyProbe = 1))

    ## test with only one sequence
    infile <- system.file("extdata", "example_alignment.txt", package = "rprimer")
    testdata <- MultipleAlignment::readDNAMultipleAlignment(infile)
    MultipleAlignment::rowmask(testdata, invert = TRUE) <- 3
    prof <- consensusProfile(testdata)
    test <- designOligos(prof[1:100, ])
    expect_s4_class(test, "RprimerOligo")
})

# .nmers =======================================================================

test_that(".nmers works", {
    seq <- sample(c("A", "C", "G", "T"), 100, replace = TRUE)
    nmer <- .nmers(seq, n = 4)
    expect_equal(ncol(nmer), 4)
    expect_equal(seq[1:4], nmer[1, ])
    expect_equal(seq[2:5], nmer[2, ])
    expect_equal(seq[(length(seq) - 3):length(seq)], nmer[nrow(nmer), ])

    ## Confirm that it returns a matrix even if it is only one row
    nmer <- .nmers(c("A", "C", "G"), 3)
    expect_true(is.matrix(nmer))
    expect_equal(dim(nmer), c(1, 3))
})

# .countDegeneracy =============================================================

test_that(".countDegeneracy works", {
    seq <- c("A", "C", "R", "N", "Y")
    degen <- .countDegeneracy(seq)
    table <- rprimer:::lookup$degeneracy
    expect_equal(degen, unname(table["R"] * table["N"] * table["Y"]))
    expect_equal(.countDegeneracy(c("A", "C", "G", "T")), 1)
})

# .splitAndPaste ===============================================================

test_that(".splitAndPaste works", {
    first <- t(matrix(rep(0, 12)))
    second <- t(matrix(rep(1, 12)))
    test <- .splitAndPaste(first, second)
    testRev <- .splitAndPaste(first, second, rev = TRUE)

    expect_true(is.matrix(test))
    expect_equal(dim(test), c(1, 12))
    expect_equal(dim(testRev), c(1, 12))

    ## If fwd, first (0) should be two thirds (= 8)
    ## and second (1) one third (= 4)
    expect_equal(sum(test[, 1:8]), 0)
    expect_equal(sum(test[9:12]), 4)

    ## If rev, first (0) should be one third (= 4)
    ## and second (1) two thirds (= 8)
    expect_equal(sum(testRev[, 1:4]), 0)
    expect_equal(sum(testRev[5:12]), 8)

    # Confirm that it works for numbers not dividable with three
    first <- matrix(rep(0, 17 * 4), ncol = 17, nrow = 4)
    second <- matrix(rep(1, 17 * 4), ncol = 17, nrow = 4)
    test <- .splitAndPaste(first, second)
    testRev <- .splitAndPaste(first, second, rev = TRUE)
    expect_equal(
        ncol(test[, colSums(test) == 0]),
        ncol(testRev[, colSums(testRev) == nrow(testRev)])
    )

    testSeparate <- .splitAndPaste(first, second, combine = FALSE)
    expect_true(is.list(testSeparate))
    expect_true(is.matrix(testSeparate[[1]]))
    expect_true(is.matrix(testSeparate[[2]]))
})

# .generateAmbiguousOligos =====================================================

test_that(".generateAmbiguousOligos works", {
    test <- .generateAmbiguousOligos(x, lengthOligo = 18)
    expect_equal(ncol(test$iupacSequence), 18)
    expect_equal(unique(test$length), 18)
    expect_equal(unique(test$roiStart), min(x$position))
    expect_equal(unique(test$roiEnd), max(x$position))
    expect_equal(
        x$iupac[test$start[4]:test$end[4]], test$iupacSequence[4, ]
    )
    expect_equal(
        test$degeneracy[4],
        .countDegeneracy(x$iupac[test$start[4]:test$end[4]])
    )
    expect_equal(
        test$gapFrequency[4], max(x$gaps[test$start[4]:test$end[4]])
    )
    expect_equal(
        test$coverage[4], mean(x$coverage[test$start[4]:test$end[4]])
    )
})

# .mergeLists ==================================================================

test_that(".mergeLists works", {
    l1 <- list("M" = matrix(rep(1, 10)), "V" = rep("A", 10))
    l2 <- list("M" = matrix(rep(2, 10)), "V" = rep("B", 10))
    l <- .mergeLists(l1, l2)
    expect_equal(names(l), names(l1))
    expect_true(is.matrix(l[[1]]))
    expect_true(is.vector(l[[2]]))
})

# .filterOligos ================================================================

test_that(".filterOligos works", {
    x <- .generateAmbiguousOligos(x)
    test <- .filterOligos(x, maxDegeneracy = 32, maxGapFrequency = 0)
    expect_true(all(test$degeneracy <= 32))
    expect_true(all(test$gapFrequency == 0))
    expect_true(all(!is.na(test$majoritySequence)))
    expect_true(all(test$majoritySequence != "-"))
})

# .expandDegenerates ===========================================================

test_that(".expandDegenerates works", {
    seq <- c("A", "R", "T", "T", "N", "G")
    degen <- .expandDegenerates(seq)
    nDegen <- .countDegeneracy(seq)
    expect_equal(nrow(degen), nDegen)
    seq2 <- c("A", "C", "G", "T")
    degen2 <- .expandDegenerates(seq2)
    expect_true(is.matrix(degen2))
    expect_equal(nrow(degen2), 1)
})

# .oligoMatrix =================================================================

test_that(".oligoMatrix works", {
    y <- .generateAmbiguousOligos(x[5000:6000, ])
    y <- .filterOligos(y, maxDegeneracy = 8)
    oligoList <- apply(y$iupacSequence, 1, .expandDegenerates)
    expect_true(all(vapply(oligoList, is.matrix, logical(1))))
    expect_equal(
        ncol(y$iupacSequence), unique(vapply(oligoList, ncol, integer(1)))
    )
    oligoMatrix <- .oligoMatrix(oligoList)
    expect_equal(length(unique(rownames(oligoMatrix))), length(oligoList))

    ## Test how it works when the maximum degeneracy is one
    y <- .generateAmbiguousOligos(x[5000:6000, ])
    y <- .filterOligos(y, maxDegeneracy = 1)
    oligoList <- apply(y$iupacSequence, 1, .expandDegenerates)
    if (!is.list(oligoList)) {
        oligoList <- t(oligoList)
        oligoList <- lapply(seq_len(nrow(oligoList)), function(i) {
            oligoList[i, , drop = FALSE]
        })
    }
    expect_equal(
        ncol(y$iupacSequence), unique(vapply(oligoList, ncol, integer(1)))
    )
    oligoMatrix <- .oligoMatrix(oligoList)
    expect_equal(length(unique(rownames(oligoMatrix))), length(oligoList))

    ## Test if there is only one oligo as input
    y <- lapply(y, function(x) {
        if (is.matrix(x)) x[1, , drop = FALSE] else x[1]
    })
    oligoList <- apply(y$iupacSequence, 1, .expandDegenerates)
    if (!is.list(oligoList)) oligoList <- list(t(oligoList))
    oligoMatrix <- .oligoMatrix(oligoList)
    expect_equal(length(unique(rownames(oligoMatrix))), length(oligoList))

    ## Test if degeneracy is 64 (max)
    y <- .generateAmbiguousOligos(x[5000:6000, ])
    y <- .filterOligos(y, maxDegeneracy = 64)
    oligoList <- apply(y$iupacSequence, 1, .expandDegenerates)
    expect_true(all(vapply(oligoList, is.matrix, logical(1))))
    expect_equal(
        ncol(y$iupacSequence), unique(vapply(oligoList, ncol, integer(1)))
    )
    oligoMatrix <- .oligoMatrix(oligoList)
    expect_equal(length(unique(rownames(oligoMatrix))), length(oligoList))
})

# .reverseComplement ===========================================================

test_that(".reverseComplement works", {
    seq <- t(matrix(c("A", "C", "G")))
    rc <- .reverseComplement(seq)
    expect_true(is.matrix(rc))
    expect_equal(dim(rc), c(1, 3))
    expect_equal(rc, c("C", "G", "T"), ignore_attr = TRUE)
})

# .gcClamp =====================================================================

test_that(".gcClamp works", {
    seq <- c("A", "C", "G", "T", "G", "C", "T", "A")
    gc <- as.integer(seq == "C" | seq == "G")
    gc <- t(matrix(gc))
    expect_true(.gcClamp(gc))
    seq <- c("A", "C", "G", "T", "T", "C", "T", "A")
    gc <- ifelse(seq == "C" | seq == "G", 1, 0)
    gc <- t(matrix(gc))
    expect_false(.gcClamp(gc))
    seq <- c("A", "C", "G", "T", "C", "C", "G", "C")
    gc <- ifelse(seq == "C" | seq == "G", 1, 0)
    gc <- t(matrix(gc))
    expect_false(.gcClamp(gc))
    seq <- t(matrix(c("A", "C", "G", "T", "C", "C", "G", "C")))
    rc <- .reverseComplement(seq)
    gcRc <- ifelse(rc == "C" | rc == "G", 1, 0)
    gcRc <- t(matrix(gcRc))
    expect_true(.gcClamp(gc, rev = TRUE))
    expect_true(.gcClamp(gcRc))
    gc <- matrix(rep(c(1, 0, 1, 1, 0, 0, 0, 0, 0), 8), ncol = 9, byrow = TRUE)
    expect_equal(sum(.gcClamp(gc)), 0)
    expect_equal(sum(.gcClamp(gc, rev = TRUE)), nrow(gc))
})

# .endRuns =====================================================================

test_that(".endRuns works", {
    seq <- t(matrix(c("A", "T", "C", "C", "C")))
    expect_true(.endRuns(seq))
    seq <- t(matrix(c("A", "T", "T", "C", "C")))
    expect_false(.endRuns(seq))
    expect_false(.endRuns(seq, rev = TRUE))
    seq <- t(matrix(c("G", "G", "G", "T", "A")))
    expect_true(.endRuns(seq, rev = TRUE))
    seq <- matrix(rep(c("G", "G", "G", "T", "A"), 10), ncol = 5, byrow = TRUE)
    expect_equal(sum(.endRuns(seq)), 0)
    expect_equal(sum(.endRuns(seq, rev = TRUE)), 10)
})

# .repeats =====================================================================

test_that(".repeats works", {
    mono <- "CTTTTTA"
    di <- "CTCTCTCTCTA"
    monodi <- "CTCTCTCTCTACTTTTTA"
    expect_true(.repeats(mono))
    expect_true(.repeats(di))
    expect_true(.repeats(monodi))
    expect_equal(sum(.repeats(c(di, mono))), 2)
})

# .allVariants ==============================================================

test_that(".allVariants works", {
    y <- .generateAmbiguousOligos(x[5000:6000, ])

    ## Make sure it works if max degeneracy is one
    y <- .filterOligos(y, maxDegeneracy = 1)
    all <- .allVariants(y)
    expect_true(is.list(all))
    expect_equal(
        all$sequence[[1]], paste(y$iupacSequence[1, ], collapse = "")
    )
    expect_equal(
        all$sequence[[length(all$sequence)]],
        paste(y$iupacSequence[nrow(y$iupacSequence), ], collapse = "")
    )
    expect_equal(unique(vapply(all$sequence, length, integer(1))), 1)

    ## If degeneracy is high
    y <- .generateAmbiguousOligos(x[100:150, ])
    y <- .filterOligos(y, maxDegeneracy = 16)
    all <- .allVariants(y)
    nVariants <- lapply(all, function(x) vapply(x, length, integer(1)))
    nVariants <- do.call("rbind", nVariants)
    expect_true(all(apply(nVariants, 2, function(x) length(unique(x))) == 1))
    expect_true(all(nVariants <= 16))
})

# .meanRange ===================================================================

test_that(".meanRange works", {
    x <- .allVariants(.filterOligos(.generateAmbiguousOligos(x)))
    test <- .meanRange(x)
    range <- function(x) max(x) - min(x)
    expect_equal(test$gcContentMean, vapply(x$gcContent, mean, double(1L)))
    expect_equal(test$tmPrimerMean, vapply(x$tmPrimer, mean, double(1L)))
    expect_equal(test$tmProbeMean, vapply(x$tmProbe, mean, double(1L)))
    expect_equal(test$gcContentRange, vapply(x$gcContent, range, double(1L)))
    expect_equal(test$tmPrimerRange, vapply(x$tmPrimer, range, double(1L)))
    expect_equal(test$tmProbeRange, vapply(x$tmProbe, range, double(1L)))
})

# .makeOligoDf =================================================================

test_that(".makeOligoDf works", {
    x <- .filterOligos(.generateMixedOligos(x))
    test <- .makeOligoDf(x)
    expect_true(is.data.frame(test))
})

# .designAmbiguousOligos =======================================================

test_that(".designAmbiguousOligos works", {
    test <- .designAmbiguousOligos(
        x,
        maxDegeneracyPrimer = 4,
        maxGapFrequency = 0,
        lengthPrimer = 15:18,
        probe = FALSE
    )
    expect_true(is.data.frame(test))
    expect_true(all(test$degeneracy <= 4))
    expect_true(all(test$method == "ambiguous"))
    expect_true(
        all(test$length >= 15 & test$length <= 18)
    )
    expect_error(.designAmbiguousOligos(x[1:50, ], maxDegeneracy = 1))
})

# .isWithinRange ===============================================================

test_that(".isWithinRange works", {
    x <- .allVariants(.filterOligos(.generateAmbiguousOligos(x)))
    x$gcContent <- x$gcContent[1]
    test <- unlist(.isWithinRange(x$gcContent, c(0.5, 0.6)))
    expect_equal(test, c(TRUE, TRUE, TRUE, FALSE))
})

# .isValid =====================================================================

test_that(".isValid works", {
    x <- .designAmbiguousOligos(x)
    gcInRange <- .isWithinRange(x$gcContent, c(0.4, 0.6))
    x <- cbind(x, data.frame(cbind(gcInRange)))
    check <- .convertToMatrices(x[c("gcInRange", "repeats", "threeEndRunsFwd")])
    test <- .isValid(check, rowThreshold = 1, colThreshold = 1)
    expect_equal(nrow(x), length(test))
})

# .filterPrimers ===============================================================

# .checkAllProbeVariants =======================================================

# .filterProbes ================================================================

# .beautifyOligos ==============================================================

#' Example datasets
#'
#' @name example-datasets
#'
#' @description
#' The purpose of these datasets is to illustrate the functionality of
#' rprimer. The following datasets are provided:
#'
#' \itemize{
#' \item \code{exampleRprimerAlignment} - a \code{MultipleAlignment::DNAMultipleAlignment}
#' object (Carlson & Aboyoun, 2020)
#' containing an alignment of 50 hepatitis E virus
#' sequences collected from NCBI GenBank. See
#' "documentation_example_alignment.txt" within the inst/script folder of this
#' package for more details.
#' \item \code{exampleRprimerProfile} - an \code{RprimerProfile} object, generated
#' from the alignment above.
#' \item \code{exampleRprimerOligo} - an \code{RprimerOligo} object, generated from
#' the consensus profile above.
#' \item \code{exampleRprimerAssay} - an \code{RprimerAssay} object, generated from
#' the oligos above.
#' \item \code{exampleRprimerMatchOligo} - an \code{RprimerMatchOligo} object,
#' describing how well some oligos match with the sequences in
#' exampleRprimerAlignment.
#' \item \code{exampleRprimerMatchAssay} - an \code{RprimerMatchAssay} object,
#' describing how well some assays match with the seuqences in
#' exampleRprimerAlignment.
#' }
#'
#' @references
#' M. Carlson and P. Aboyoun (2026). MultipleAlignment:
#' Representation of multiple sequence alignments in Bioconductor.
#' R package version 0.99.5.
NULL

#' @rdname example-datasets
#'
#' @usage data("exampleRprimerAlignment")
"exampleRprimerAlignment"

#' @rdname example-datasets
#'
#' @usage data("exampleRprimerProfile")
"exampleRprimerProfile"

#' @rdname example-datasets
#'
#' @usage data("exampleRprimerOligo")
"exampleRprimerOligo"

#' @rdname example-datasets
#'
#' @usage data("exampleRprimerAssay")
"exampleRprimerAssay"

#' @rdname example-datasets
#'
#' @usage data("exampleRprimerMatchOligo")
"exampleRprimerMatchOligo"

#' @rdname example-datasets
#'
#' @usage data("exampleRprimerMatchAssay")
"exampleRprimerMatchAssay"

rprimerApp <- function() {
    ui <- shiny::fluidPage(

        shiny::tags$head(
            shiny::tags$style(
                shiny::HTML(".shiny-notification {
             position:fixed;
             top: calc(50%);
             left: calc(50%);
             }
             ")
            )
        ),
        shinyFeedback::useShinyFeedback(),
        theme = bslib::bs_theme(bootswatch = "yeti"),
        shiny::titlePanel(
            title = "Design degenerate oligos from a multiple DNA sequence alignment"
        ),
        shiny::tabsetPanel(
            id = "wizard",
            type = "hidden",
            shiny::tabPanelBody(
                "page1", "",
                welcomeUI("welcome"),
                shiny::br(),
                shiny::fluidRow(
                    shiny::column(width = 6),
                    shiny::column(
                        width = 6, align = "right",
                        nextButton(2)
                    )
                ),
            ),
            shiny::tabPanelBody(
                "page2", "",
                dataUI("data"),
                shiny::br(),
                shiny::fluidRow(
                    shiny::column(
                        width = 6,
                        previousButton(1)
                    ),
                    shiny::column(
                        width = 6, align = "right",
                        shiny::uiOutput("page23")
                    )
                )
            ),
            shiny::tabPanelBody(
                "page3", "",
                consensusUI("consensus"),
                shiny::br(),
                shiny::fluidRow(
                    shiny::column(
                        width = 6,
                        previousButton(2)
                    ),
                    shiny::column(
                        width = 6, align = "right",
                        shiny::uiOutput("page34")
                    )
                )
            ),
            shiny::tabPanelBody(
                "page4", "",
                oligoUI("oligo"),
                shiny::br(),
                shiny::fluidRow(
                    shiny::column(
                        width = 6,
                        previousButton(3)
                    ),
                    shiny::column(
                        width = 6, align = "right",
                        shiny::uiOutput("page45")
                    )
                )
            ),
            shiny::tabPanelBody(
                "page5", "",
                oligoFilterUI("oligoFilter"),
                shiny::br(),
                shiny::fluidRow(
                    shiny::column(
                        width = 6,
                        previousButton(4)
                    ),
                    shiny::column(
                        width = 6, align = "right",
                        nextButton(6)
                    )
                )
            ),
            shiny::tabPanelBody(
                "page6", "",
                assayUI("assay"),
                shiny::br(),
                shiny::fluidRow(
                    shiny::column(
                        width = 6,
                        previousButton(5)
                    ),
                    shiny::column(
                        width = 6, align = "right",
                        shiny::uiOutput("page67")
                    )
                )
            ),
            shiny::tabPanelBody(
                "page7", "",
                assayFilterUI("assayFilter"),
                shiny::br(),
                shiny::fluidRow(
                    shiny::column(
                        width = 6,
                        previousButton(6)
                    )
                )
            )
        )
    )

    server <- function(input, output, session) {
        alignment <- dataServer("data")

        consensus <- consensusServer("consensus", inputAlignment = alignment$data)

        oligo <- oligoServer(
            "oligo",
            alignment = alignment$data, consensus = consensus$data
        )

        oligoFilter <- oligoFilterServer(
            "oligoFilter",
            alignment = alignment$data, consensus = consensus$data,
            allOligos = oligo$data
        )

        assay <- assayServer(
            "assay",
            alignment = alignment$data, consensus = consensus$data,
            oligo = oligoFilter$data
        )

        assayFilterServer(
            "assayFilter",
            alignment = alignment$data, consensus = consensus$data,
            allAssays = assay$data
        )

        output$page23 <- shiny::renderUI({
            shiny::req(alignment$data())
            nextButton(3)
        })

        output$page34 <- shiny::renderUI({
            shiny::req(consensus$data())
            nextButton(4)
        })

        output$page45 <- shiny::renderUI({
            shiny::req(oligo$data())
            nextButton(5)
        })

        output$page56 <- shiny::renderUI({
            shiny::req(oligoFilter$data())
            nextButton(6)
        })

        output$page67 <- shiny::renderUI({
            shiny::req(assay$data())
            nextButton(7)
        })

        shiny::observeEvent(input$page12, switchPage(2))
        shiny::observeEvent(input$page21, switchPage(1))
        shiny::observeEvent(input$page23, switchPage(3))
        shiny::observeEvent(input$page32, switchPage(2))
        shiny::observeEvent(input$page34, switchPage(4))
        shiny::observeEvent(input$page43, switchPage(3))
        shiny::observeEvent(input$page45, switchPage(5))
        shiny::observeEvent(input$page54, switchPage(4))
        shiny::observeEvent(input$page56, switchPage(6))
        shiny::observeEvent(input$page65, switchPage(5))
        shiny::observeEvent(input$page67, switchPage(7))
        shiny::observeEvent(input$page76, switchPage(6))
    }

    shiny::shinyApp(ui, server)
}

welcomeUI <- function(id) {
    shiny::tagList(
        shiny::h5("Welcome!"),
        shiny::hr(),
        shiny::h6(
            "This application provides tools for visualizing sequence
                                conservation and designing degenerate primers, probes and (RT)-(q/d)PCR
                                assays from a multiple DNA sequence alignment."
        ),
        shiny::h6(
            "The workflow consists of five steps:"
        ),
        shiny::h6(shiny::tags$ul(
            shiny::tags$li("Generation of a consensus profile"),
            shiny::tags$li("Oligo design"),
            shiny::tags$li("Optional filtering of oligos"),
            shiny::tags$li("Assay design"),
            shiny::tags$li("Optional filtering of assays")
        )),
        shiny::h6(
            "Please see the package vingette or the package manual for
            further information on the different steps of the design procedure."
        ),
        shiny::hr()
    )
}

# Rprimer classes =============================================================

#' S4 classes for representation of different Rprimer objects
#'
#' @name Rprimer-classes
#'
#' @description
#' The rprimer package contains five different S4 classes. Each class is used
#' as input or output for the different functions within the oligo and assay
#' design workflow:
#'
#'  \itemize{
#'  \item \code{RprimerProfile}: output from \code{consensusProfile()},
#'  input for \code{oligos()}.
#'  \item \code{RprimerOligo}: output from \code{oligos()}, input
#'  for \code{assays()} and \code{checkMatch()}.
#'  \item \code{RprimerAssay}: output from \code{assays()}, input
#'  for \code{checkMatch()}.
#'  \item \code{RprimerMatchOligo}: output from \code{checkMatch()}.
#'  \item \code{RprimerMatchAssay}: output from \code{checkMatch()}.
#' }
#'
#' These classes extends the DFrame class from S4vectors
#' (Pages et al., 2020), without any additional slots, but with some
#' additional checks for validity.
#'
#' @section Coercion:
#' Each class can be converted to a traditional data frame, by using either
#' \code{as()} or \code{as.data.frame()}.
#'
#' Moreover, \code{as()} can also be used for converting oligo sequences
#' within an
#' \code{RprimerOligo} or \code{RprimerAssay} object into a
#' \code{Biostrings::DNAStringSet}
#' object (Pages et al., 2020).
#' Note that all oligo sequences will be written in the
#' same direction as the input alignment that was used to generate
#' the oligos.
#'
#' @param ...
#' A data frame or list to be converted into an Rprimer-object.
#'
#' @return
#' An Rprimer-object if validation succeeds, an error
#' message otherwise.
#'
#' @import methods
#'
#' @importClassesFrom S4Vectors DataFrame
#'
#' @importClassesFrom Biostrings DNAStringSet
#'
#' @seealso consensusProfile, oligos, assays, checkMatch
#'
#' @references
#' Pages, H., Lawrence, M., and Aboyoun, R. (2020). S4Vectors:
#' Foundation of vector-like and list-like containers in
#' Bioconductor. R package version 0.28.0.
#'
#' Pages, H., Aboyoun, P., Gentleman R., and DebRoy S. (2020). Biostrings:
#' Efficient manipulation of biological strings. R package version
#' 2.57.2.
#'
#' @examples
#' ## Constructors
#'
#' data("exampleRprimerProfile")
#' x <- as.data.frame(exampleRprimerProfile)
#' RprimerProfile(x)
#'
#' data("exampleRprimerOligo")
#' x <- as.data.frame(exampleRprimerOligo)
#' RprimerOligo(x)
#'
#' data("exampleRprimerAssay")
#' x <- as.data.frame(exampleRprimerAssay)
#' RprimerAssay(x)
#'
#' data("exampleRprimerMatchOligo")
#' x <- as.data.frame(exampleRprimerMatchOligo)
#' RprimerMatchOligo(x)
#'
#' data("exampleRprimerMatchAssay")
#' x <- as.data.frame(exampleRprimerMatchAssay)
#' RprimerMatchAssay(x)
#'
#' ## Coercion methods for RprimerOligo and RprimerAssay objects
#'
#' ## Convert an RprimerOligo object to a DNAStringSet
#' data("exampleRprimerOligo")
#'
#' ## Pick rows to convert
#' x <- exampleRprimerOligo[1:2, ]
#' as(x, "DNAStringSet")
#'
#' ## Convert an RprimerAssay object to a DNAStringSet
#' data("exampleRprimerAssay")
#'
#' ## Pick rows to convert
#' x <- exampleRprimerAssay[1:2, ]
#' as(x, "DNAStringSet")
NULL

# RprimerProfile ===============================================================

#' @rdname Rprimer-classes
.RprimerProfile <- setClass("RprimerProfile", contains = "DFrame")

#' @export
#'
#' @rdname Rprimer-classes
RprimerProfile <- function(...) {
    S4Vectors::DataFrame(..., row.names = NULL, check.names = TRUE) |>
        .RprimerProfile()
}

S4Vectors::setValidity2("RprimerProfile", \(object) {
    msg <- NULL
    colnames <- c(
        "position", "a", "c", "g", "t", "other", "gaps", "majority", "identity",
        "iupac", "coverage"
    )
    if (!all(colnames %in% names(object))) {
        msg <- c(
            msg, "This type of subsetting is not allowed for an
            RprimerProfile object."
        )
    }
    if (is.null(msg)) {
        TRUE
    } else {
        msg
    }
})

# RprimerOligo ================================================================

#' @rdname Rprimer-classes
.RprimerOligo <- setClass("RprimerOligo", contains = "DFrame")

#' @export
#'
#' @rdname Rprimer-classes
RprimerOligo <- function(...) {
    S4Vectors::DataFrame(..., row.names = NULL, check.names = TRUE) |>
        .RprimerOligo()
}

S4Vectors::setValidity2("RprimerOligo", \(object) {
    msg <- NULL
    colnames <- c(
        "type", "fwd", "rev", "start", "end", "length", "iupacSequence",
        "iupacSequenceRc", "identity",
        "coverage", "degeneracy", "gcContentMean", "gcContentRange", "tmMean",
        "tmRange", "deltaGMean", "deltaGRange",
        "sequence", "sequenceRc", "gcContent", "tm", "deltaG", "method",
        "score", "roiStart", "roiEnd"
    )
    if (!all(colnames %in% names(object))) {
        msg <- c(
            msg, "This type of subsetting is not allowed for an
            RprimerOligo object."
        )
    }
    if (is.null(msg)) {
        TRUE
    } else {
        msg
    }
})

# RprimerAssay ================================================================

#' @rdname Rprimer-classes
.RprimerAssay <- setClass("RprimerAssay", contains = "DFrame")

#' @export
#'
#' @rdname Rprimer-classes
RprimerAssay <- function(...) {
    S4Vectors::DataFrame(..., row.names = NULL, check.names = TRUE) |>
        .RprimerAssay()
}

S4Vectors::setValidity2("RprimerAssay", \(object) {
    msg <- NULL
    oligoColnames <- c(
        "start", "end",
        "length", "iupacSequence", "identity",
        "coverage", "degeneracy", "gcContentMean",
        "gcContentRange", "tmMean", "tmRange",
        "sequence",
        "gcContent", "tm", "deltaG",
        "method"
    )
    colnames <- c(
        "start", "end", "length",
        "totalDegeneracy", "score",
        paste0(oligoColnames, "Fwd"),
        paste0(oligoColnames, "Rev"),
        "roiStart", "roiEnd"
    )
    if (!all(colnames %in% names(object))) {
        msg <- c(
            msg, "This type of subsetting is not allowed for an
            RprimerAssay object."
        )
    }
    if (is.null(msg)) {
        TRUE
    } else {
        msg
    }
})

# RprimerMatchOligo ============================================================

#' @rdname Rprimer-classes
.RprimerMatchOligo <- setClass("RprimerMatchOligo", contains = "DFrame")

#' @export
#'
#' @rdname Rprimer-classes
RprimerMatchOligo <- function(...) {
    S4Vectors::DataFrame(..., row.names = NULL, check.names = TRUE) |>
        .RprimerMatchOligo()
}

S4Vectors::setValidity2("RprimerMatchOligo", \(object) {
    msg <- NULL
    colnames <- c(
        "perfectMatch", "oneMismatch", "twoMismatches",
        "threeMismatches", "fourOrMoreMismatches", "offTargetMatch"
    )
    if (!all(colnames %in% names(object))) {
        msg <- c(
            msg, "This type of subsetting is not allowed for an
            RprimerMatchOligo object."
        )
    }
    if (is.null(msg)) {
        TRUE
    } else {
        msg
    }
})

# RprimerMatchAssay ============================================================

#' @rdname Rprimer-classes
.RprimerMatchAssay <- setClass("RprimerMatchAssay", contains = "DFrame")

#' @export
#'
#' @rdname Rprimer-classes
RprimerMatchAssay <- function(...) {
    S4Vectors::DataFrame(..., row.names = NULL, check.names = TRUE) |>
        .RprimerMatchAssay()
}

S4Vectors::setValidity2("RprimerMatchAssay", \(object) {
    msg <- NULL
    oligoColnames <- c(
        "perfectMatch", "oneMismatch", "twoMismatches",
        "threeMismatches", "fourOrMoreMismatches", "offTargetMatch"
    )
    colnames <- c(
        paste0(oligoColnames, "Fwd"),
        paste0(oligoColnames, "Rev")
    )
    if (!all(colnames %in% names(object))) {
        msg <- c(
            msg, "This type of subsetting is not allowed for an
            RprimerMatchAssay object."
        )
    }
    if (is.null(msg)) {
        TRUE
    } else {
        msg
    }
})

# Coerce =======================================================================

#' @name coerce
#'
#' @rdname Rprimer-classes
setAs("RprimerOligo", "DNAStringSet", \(from) .toDNAStringSetOligo(from))

#' @name coerce
#'
#' @rdname Rprimer-classes
setAs("RprimerAssay", "DNAStringSet", \(from) .toDNAStringSetAssay(from))

# Helpers ======================================================================

.addNames <- function(x, type, additionalInfo = "") {
    unlist(lapply(seq_along(x), \(i) {
        names(x[[i]]) <- paste0(
            type, "_", i, additionalInfo, "_variant_", seq_along(x[[i]])
        )
        x[[i]]
    }))
}

.toDNAStringSetOligo <- function(x) {
    oligo <- x$sequence
    oligo <- .addNames(oligo, "oligo") |>
        Biostrings::DNAStringSet()
    oligo
}

.toDNAStringSetAssay <- function(x) {
    fwd <- x$sequenceFwd
    fwd <- .addNames(fwd, "assay", "_fwd") |>
        Biostrings::DNAStringSet()
    rev <- x$sequenceRev
    rev <- .addNames(rev, "assay", "_rev") |>
        Biostrings::DNAStringSet() |>
        Biostrings::reverseComplement()
    if ("sequencePr" %in% names(x)) {
        pr <- x$sequencePr
        pr <- .addNames(pr, "assay", "_pr") |>
            Biostrings::DNAStringSet()
    } else {
        pr <- NULL
    }
    c(fwd, rev, pr)
}

## Import data to test on
data("exampleRprimerProfile")
data("exampleRprimerOligo")
data("exampleRprimerAssay")

# RprimerProfile ===============================================================

test_that("RprimerProfile validation works", {
    expect_error(RprimerProfile(10))

    ## Should be subsettable by row, but not by column
#    expect_error(exampleRprimerProfile[, 1:2])
    expect_s4_class(exampleRprimerProfile[1:10, ], "RprimerProfile")
})

# RprimerOligo =================================================================

test_that("RprimerOligo validation works", {
    expect_error(RprimerOligo(10))

    ## Should be subsettable by row, but not by column
 #   expect_error(exampleRprimerOligo[, 1:2])
    expect_s4_class(exampleRprimerOligo[1:10, ], "RprimerOligo")
})

# RprimerAssay =================================================================

test_that("RprimerAssay validation works", {
    expect_error(RprimerAssay(10))

    ## Should be subsettable by row, but not by column
 #   expect_error(exampleRprimerAssay[, 1:2])
    expect_s4_class(exampleRprimerAssay[1:10, ], "RprimerAssay")
})

assayFilterUI <- function(id) {
    ns <- shiny::NS(id)

    shiny::tagList(
        shiny::h5("Filter assays (step 5/5)"),
        shiny::hr(),
        shiny::sidebarLayout(
            shiny::sidebarPanel(
                width = 3,
                shiny::uiOutput(ns("assayFilter"))
            ),
            shiny::mainPanel(
                width = 9,
                shiny::tabsetPanel(
                    id = ns("wizard"),
                    shiny::tabPanel(
                        title = "Plot",
                        shiny::br(),
                        spinnerPlot(
                            ns("assayPlot"),
                            width = "100%", height = 600
                        )
                    ),
                    shiny::tabPanel(
                        title = "Table",
                        shiny::br(),
                        shiny::uiOutput(ns("getDownloadLinkTxt")),
                        shiny::uiOutput(ns("getDownloadLinkFasta")),
                        shiny::br(),
                        shiny::h6("Select a row for more details"),
                        shiny::br(),
                        DT::dataTableOutput(ns("assayTable"))
                    ),
                    shiny::tabPanel(
                        title = "Selection",
                        shiny::br(),
                        shiny::uiOutput(ns("getDownloadLinkTxtSel")),
                        shiny::uiOutput(ns("getDownloadLinkFastaSel")),
                        shiny::br(),
                        DT::dataTableOutput(ns("overviewTable")),
                        shiny::br(),
                        shiny::br(),
                        spinnerPlot(ns("assayPositionPlot")),
                        shiny::br(),
                        shiny::h6(shiny::tags$b("Amplicon sequence")),
                        shiny::hr(),
                        shiny::verbatimTextOutput(ns("ampliconSequence")),
                        shiny::h6(shiny::tags$b("Oligo details")),
                        shiny::hr(),
                        shiny::uiOutput(ns("detailsTab"))
                    )
                )
            )
        ),
        shiny::hr()
    )
}


assayFilterServer <- function(id, alignment, consensus, oligo, allAssays) {
    shiny::moduleServer(id, function(input, output, session) {
        assay <- shiny::reactive({
            shiny::req(nrow(allAssays()) > 0)
            x <- allAssays()
            x <- as.data.frame(x)
            emptyRow <- makeEmptyRow(x)
            x <- x[
                x$start >= input$assayRegionFrom & x$end <= input$assayRegionTo,
            ]
            x <- x[x$score <= input$maxAssayScore, ]
            if (nrow(x) == 0) {
                x <- emptyRow
            }
             RprimerAssay(x)
        })

        output$assayFilter <- shiny::renderUI({
            ns <- session$ns
            shiny::req(nrow(allAssays()) > 0)

            list(
                numericInputFrom(allAssays(), ns("assayRegionFrom")),
                numericInputTo(allAssays(), ns("assayRegionTo")),
                sliderInput(ns("maxAssayScore"),
                            round = -4, step = 0.0001,
                            "Maximum score (lower is better)",
                            min = round(min(allAssays()$score, na.rm = TRUE), 4),
                            max = round(max(allAssays()$score, na.rm = TRUE), 4),
                            value = round(max(allAssays()$score, na.rm = TRUE), 4)
                )
            )
        })

        output$assayPlot <- shiny::renderPlot({
            shiny::req(is(assay(), "RprimerAssay"))
            plotData(assay())
        })


        output$getDownloadLinkTxt <- shiny::renderUI({
            shiny::req(is(assay(), "RprimerAssay"))
            ns <- session$ns
            list(
                shiny::downloadLink(
                    ns("downloadTxt"), "Download table as .txt"
                ),
                shiny::br()
            )
        })

        output$getDownloadLinkFasta <- shiny::renderUI({
            shiny::req(is(assay(), "RprimerAssay"))
            ns <- session$ns
            list(
                shiny::downloadLink(
                    ns("downloadFasta"), "Download sequence(s) in fasta-format"
                ),
                shiny::br()
            )
        })

        output$downloadTxt <- shiny::downloadHandler(
            filename <- function() {
                paste0("assay", "-", Sys.Date(), ".txt")
            },
            content <- function(file) {
                utils::write.table(
                    as.data.frame(assay()), file,
                    quote = FALSE, sep = "\t",
                    row.names = FALSE
                )
            }
        )

        output$downloadFasta <- shiny::downloadHandler(
            filename <- function() {
                paste0("assay", "-fasta-", Sys.Date(), ".txt")
            },
            content <- function(file) {
                data <- as(assay(), "DNAStringSet")
                Biostrings::writeXStringSet(data, file)
            }
        )

        output$assayTable <- DT::renderDataTable(
            {
                shiny::req(!is.na(assay()$length[[1]]))
                x <- roundDbls(removeListColumns(as.data.frame(assay())))
                if (any(grepl("Pr", names(x)))) {
                    x$iupacSequencePr <- ifelse(x$plusPr, x$iupacSequencePr, NA)
                    x$iupacSequenceRcPr <- ifelse(x$minusPr, x$iupacSequenceRcPr, NA)
                }
                names(x) <- if (any(grepl("Pr", names(x)))) {
                    c(
                        "Start", "End", "Length", "Total degeneracy", "Score",
                        "Start, forward", "End, forward", "Length, forward",
                        "IUPAC sequence, forward", "Identity, forward",
                        "Coverage, forward", "Degeneracy, forward",
                        "GC content, mean, forward", "GC content, range, forward",
                        "Tm, mean, forward", "Tm, range, forward",
                        "Delta G, mean, forward", "Delta G, range, forward",
                        "Design method, forward",
                        "Start, reverse", "End, reverse", "Length, reverse",
                        "IUPAC sequence, reverse", "Identity, reverse",
                        "Coverage, reverse", "Degeneracy, reverse",
                        "GC content, mean, reverse", "GC content, range, reverse",
                        "Tm, mean, reverse", "Tm, range, reverse",
                        "Delta G, mean, reverse", "Delta G, range, reverse",
                        "Design method, reverse",
                        "Plus sense, probe", "Minus sense, probe",
                        "Start, probe", "End, probe", "Length, probe",
                        "IUPAC sequence, probe", "IUPAC sequence RC, probe",
                        "Identity, probe",
                        "Coverage, probe", "Degeneracy, probe",
                        "GC content, mean, probe", "GC content, range, probe",
                        "Tm, mean, probe", "Tm, range, probe",
                        "Delta G, mean, probe", "Delta G, range, probe",
                        "Design method, probe", "ROI, start", "ROI, end"
                    )
                } else {
                    c(
                        "Start", "End", "Length", "Total degeneracy", "Score",
                        "Start, forward", "End, forward", "Length, forward",
                        "IUPAC sequence, forward", "Identity, forward",
                        "Coverage, forward", "Degeneracy, forward",
                        "GC content, mean, forward", "GC content, range, forward",
                        "Tm, mean, forward", "Tm, range, forward",
                        "Delta G, mean, forward", "Delta G, range, forward",
                        "Design method, forward",
                        "Start, reverse", "End, reverse", "Length, reverse",
                        "IUPAC sequence, reverse", "Identity, reverse",
                        "Coverage, reverse", "Degeneracy, reverse",
                        "GC content, mean, reverse", "GC content, range, reverse",
                        "Tm, mean, reverse", "Tm, range, reverse",
                        "Delta G, mean, reverse", "Delta G, range, reverse",
                        "Design method, reverse",
                        "ROI, start", "ROI, end"
                    )
                }
                x
            },
            options = list(
                info = FALSE,
                searching = FALSE, paging = TRUE,
                pageLength = 100,
                scrollX = TRUE, autoWidth = TRUE,
                ordering = TRUE, scrollY = "1000"
            ),
            rownames = FALSE,
            selection = list(mode = "single")
        )

        shiny::observeEvent(input$assayTable_rows_selected, {
            Sys.sleep(1)
            shiny::updateTabsetPanel(session, "wizard", selected = "Selection")
        })

        selectedAssay <- shiny::reactive({
            shiny::req(assay())
            if (!is.null(input$assayTable_rows_selected)) {
                assay()[input$assayTable_rows_selected, ]
            } else {
                NULL
            }
        })

        assayList <- shiny::reactive({
            shiny::req(selectedAssay())
            splitAssayToList(selectedAssay())
        })

        assayMatch <- shiny::reactive({
            shiny::req(selectedAssay())
            shiny::req(alignment())
            if (is.na(selectedAssay()$length[[1]])) {
                NULL
            } else {
                checkMatch(selectedAssay(), alignment())
            }
        })

        assayMatchList <- shiny::reactive({
            shiny::req(assayMatch())
            splitAssayToList(assayMatch())
        })

        output$ampliconSequence <- shiny::renderPrint({
            shiny::req(selectedAssay())
            from <- selectedAssay()$start
            to <- selectedAssay()$end
            sequence <- consensus()$iupac[
                consensus()$position >= from & consensus()$position <= to
            ]
            paste(sequence, collapse = "")
        })

        output$getDownloadLinkTxtSel <- shiny::renderUI({
            shiny::req(is(selectedAssay(), "RprimerAssay"))
            ns <- session$ns
            list(
                shiny::downloadLink(
                    ns("downloadTxtSel"), "Download table as .txt"
                ),
                shiny::br()
            )
        })

        output$getDownloadLinkFastaSel <- shiny::renderUI({
            shiny::req(is(selectedAssay(), "RprimerAssay"))
            ns <- session$ns
            list(
                shiny::downloadLink(
                    ns("downloadFastaSel"),
                    "Download sequence(s) in fasta-format"
                ),
                shiny::br()
            )
        })

        output$downloadTxtSel <- shiny::downloadHandler(
            filename <- function() {
                paste0("assay-selection", "-", Sys.Date(), ".txt")
            },
            content <- function(file) {
                utils::write.table(
                    as.data.frame(selectedAssay()), file,
                    quote = FALSE, sep = "\t",
                    row.names = FALSE
                )
            }
        )

        output$downloadFastaSel <- shiny::downloadHandler(
            filename <- function() {
                paste0("assay-selection", "-fasta-", Sys.Date(), ".txt")
            },
            content <- function(file) {
                data <- as(selectedAssay(), "DNAStringSet")
                Biostrings::writeXStringSet(data, file)
            }
        )

        output$overviewTable <- DT::renderDataTable(
            {
                shiny::req(is(selectedAssay(), "RprimerAssay"))
                if (is.na(selectedAssay()$length[[1]])) {
                    NULL
                } else {
                    x <- as.data.frame(selectedAssay())
                    if (any(grepl("Pr", names(x)))) {
                        x$iupacSequencePr <- ifelse(x$plusPr, x$iupacSequencePr, NA)
                        x$iupacSequenceRcPr <- ifelse(x$minusPr, x$iupacSequenceRcPr, NA)
                        x <- x[c(
                            "start", "end", "length", "iupacSequenceFwd",
                            "iupacSequenceRev",
                            "iupacSequencePr", "iupacSequenceRcPr"
                        )]
                        names(x) <- c(
                            "Start", "End", "Length", "Forward", "Reverse",
                            "Probe, plus", "Probe, minus"
                        )
                    } else {
                        x <- x[c(
                            "start", "end", "length",
                            "iupacSequenceFwd", "iupacSequenceRev"
                        )]
                        names(x) <- c(
                            "Start", "End", "Length", "Forward", "Reverse"
                        )
                    }
                    x
                }
            },
            options = list(
                info = FALSE,
                searching = FALSE, paging = FALSE,
                scrollX = TRUE, autoWidth = FALSE,
                ordering = FALSE
            ),
            rownames = FALSE,
            selection = "none"
        )

        output$assayPositionPlot <- shiny::renderPlot({
            req(selectedAssay())
            req(consensus())
            plotData(
                consensus(),
                highlight = c(selectedAssay()$start, selectedAssay()$end)
            )
        })

        output$detailsTab <- shiny::renderUI({
            ns <- session$ns
            tabs <- list(
                shiny::tabPanel(
                    title = "Forward",
                    br(),
                    shiny::tabsetPanel(
                        type = "pills",
                        shiny::tabPanel(
                            title = "Oligo information",
                            shiny::br(),
                            shiny::h6(shiny::tags$b("Overview")),
                            DT::dataTableOutput(ns("overviewTableFwd")),
                            shiny::br(),
                            shiny::h6(shiny::tags$b("All sequence variants")),
                            DT::dataTableOutput(ns("allVariantTableFwd")),
                            shiny::br(),
                            shiny::h6(shiny::tags$b("Nucleotide distribution in target alignment")),
                            shiny::hr(),
                            shiny::br(),
                            shiny::column(
                                width = 12, align = "center",
                                spinnerPlot(
                                    ns("bindingRegionPlotFwd"),
                                    width = "75%"
                                )
                            )
                        ),
                        shiny::tabPanel(
                            title = "Match details",
                            shiny::br(),
                            shiny::h6(shiny::tags$b("Proportion of matching sequences")),
                            shiny::br(),
                            DT::dataTableOutput(ns("matchTableFwd")),
                            shiny::br(),
                            shiny::column(
                                width = 12, align = "center",
                                spinnerPlot(ns("matchPlotFwd"), width = "75%")
                            ),
                            shiny::br(),
                            shiny::h6(shiny::tags$b("Sequence names")),
                            shiny::hr(),
                            shiny::htmlOutput(ns("matchIdFwd"))
                        )
                    )
                ),
                shiny::tabPanel(
                    title = "Reverse",
                    br(),
                    shiny::tabsetPanel(
                        type = "pills",
                        shiny::tabPanel(
                            title = "Oligo information",
                            shiny::br(),
                            shiny::h6(shiny::tags$b("Overview")),
                            DT::dataTableOutput(ns("overviewTableRev")),
                            shiny::br(),
                            shiny::h6(shiny::tags$b("All sequence variants")),
                            DT::dataTableOutput(ns("allVariantTableRev")),
                            shiny::br(),
                            shiny::h6(shiny::tags$b("Nucleotide distribution in target alignment")),
                            shiny::hr(),
                            shiny::br(),
                            shiny::column(
                                width = 12, align = "center",
                                spinnerPlot(
                                    ns("bindingRegionPlotRev"),
                                    width = "75%"
                                )
                            )
                        ),
                        shiny::tabPanel(
                            title = "Match details",
                            shiny::br(),
                            shiny::h6(shiny::tags$b("Proportion of matching sequences")),
                            shiny::br(),
                            DT::dataTableOutput(ns("matchTableRev")),
                            shiny::br(),
                            shiny::column(
                                width = 12, align = "center",
                                spinnerPlot(ns("matchPlotRev"), width = "75%")
                            ),
                            shiny::br(),
                            shiny::h6(shiny::tags$b("Sequence names")),
                            shiny::hr(),
                            shiny::htmlOutput(ns("matchIdRev"))
                        )
                    )
                ),
                shiny::tabPanel(
                    title = "Probe",
                    br(),
                    shiny::tabsetPanel(
                        type = "pills",
                        shiny::tabPanel(
                            title = "Oligo information",
                            shiny::br(),
                            shiny::h6(shiny::tags$b("Overview")),
                            DT::dataTableOutput(ns("overviewTablePr")),
                            shiny::br(),
                            shiny::h6(shiny::tags$b("All sequence variants")),
                            DT::dataTableOutput(ns("allVariantTablePr")),
                            shiny::br(),
                            shiny::h6(shiny::tags$b("Nucleotide distribution in target alignment")),
                            shiny::hr(),
                            shiny::br(),
                            shiny::column(
                                width = 12, align = "center",
                                spinnerPlot(ns("bindingRegionPlotPr"), width = "75%")
                            )
                        ),
                        shiny::tabPanel(
                            title = "Match details",
                            shiny::br(),
                            shiny::h6(shiny::tags$b("Proportion of matching sequences")),
                            shiny::br(),
                            DT::dataTableOutput(ns("matchTablePr")),
                            shiny::br(),
                            shiny::column(
                                width = 12, align = "center",
                                spinnerPlot(ns("matchPlotPr"), width = "75%")
                            ),
                            shiny::br(),
                            shiny::h6(shiny::tags$b("Sequence names")),
                            shiny::hr(),
                            shiny::htmlOutput(ns("matchIdPr"))
                        )
                    )
                )
            )

            if (length(assayList()) == 3) {
                do.call(shiny::tabsetPanel, tabs)
            } else {
                do.call(shiny::tabsetPanel, tabs[seq_len(2)])
            }
        })

        oligoFwd <- reactive(convertToOligo(assayList()[[1]]))

        oligoRev <- reactive(convertToOligo(assayList()[[2]], rev = TRUE))

        oligoPr <- reactive(
            if (length(assayList()) == 3L) {
                x <- assayList()[[3]]
                convertToOligo(x, type = "probe")
            } else {
                NULL
            }
        )

        matchFwd <- shiny::reactive({
            shiny::req(oligoFwd())
            shiny::req(alignment())
            checkMatch(oligoFwd(), alignment())
        })

        output$overviewTableFwd <- DT::renderDataTable(
            {
                shiny::req(is(oligoFwd(), "RprimerOligo"))
                x <- roundDbls(removeListColumns(as.data.frame(oligoFwd())))
                names(x) <- c(
                    "Type", "Forward", "Reverse", "Start", "End", "Length",
                    "IUPAC sequence", "IUPAC sequence, RC", "Identity",
                    "Coverage", "Degeneracy", "GC content, mean",
                    "GC content, range", "Tm, mean", "Tm, range", "Delta G, mean",
                    "Delta G, range", "Design method", "Score", "ROI, start",
                    "ROI, end"
                )
                x <- x[!names(x) %in% c(
                    "Type", "Forward", "Reverse", "Score", "IUPAC sequence, RC",
                    "ROI, start", "ROI, end"
                )]
                x
            },
            options = list(
                info = FALSE,
                searching = FALSE, paging = FALSE,
                scrollX = TRUE, autoWidth = TRUE,
                ordering = FALSE
            ),
            rownames = FALSE,
            selection = "none"
        )

        output$allVariantTableFwd <- DT::renderDataTable(
            {
                shiny::req(is(oligoFwd(), "RprimerOligo"))
                x <- roundDbls(makeListTable(as.data.frame(oligoFwd())))
                names(x) <- c(
                    "Sequence", "Sequence, RC", "GC content", "Tm", "Delta G"
                )
                x <- x[!names(x) %in% c("Sequence, RC")]
                x
            },
            options = list(
                info = FALSE,
                searching = FALSE, paging = FALSE,
                scrollX = TRUE, autoWidth = FALSE,
                ordering = FALSE
            ),
            rownames = FALSE,
            selection = "none"
        )

        output$bindingRegionPlotFwd <- shiny::renderPlot({
            shiny::req(is(oligoFwd(), "RprimerOligo"))
            shiny::req(consensus())
            from <- oligoFwd()$start
            to <- oligoFwd()$end
            bindingRegion <- consensus()[
                consensus()$position >= from & consensus()$position <= to,
            ]
            plotData(bindingRegion, type = "nucleotide")
        })

        output$matchTableFwd <- DT::renderDataTable(
            {
                shiny::req(matchFwd())
                x <- roundDbls(removeListColumns(as.data.frame(matchFwd())))
                names(x) <- c(
                    "IUPAC sequence", "Perfect match", "1 mismatch", "2 mismatches",
                    "3 mismatches", "4 or more mismatches",
                    "Off target match (<3 mismatches)"
                )
                x
            },
            options = list(
                info = FALSE,
                searching = FALSE, paging = FALSE,
                scrollX = TRUE, autoWidth = FALSE,
                ordering = FALSE
            ),
            rownames = FALSE,
            selection = "none"
        )

        output$matchPlotFwd <- shiny::renderPlot({
            shiny::req(matchFwd())
            plotData(matchFwd())
        })

        output$matchIdFwd <- shiny::renderText({
            shiny::req(matchFwd())
            x <- matchFwd()
            c(
                "<b>Perfect match</b><br>",
                x$idPerfectMatch[[1]],
                "<br><br><b>One mismatch</b><br>",
                x$idOneMismatch[[1]],
                "<br><br><b>Two mismatches</b><br>",
                x$idTwoMismatches[[1]],
                "<br><br><b>Three mismatches</b><br>",
                x$idThreeMismatches[[1]],
                "<br><br><b>Four or more mismatches</b><br>",
                x$idFourOrMoreMismatches[[1]],
                "<br><br><b>Off target match (< 3 mismatches)</b><br>",
                x$idOffTargetMatch[[1]]
            )
        })

        matchRev <- shiny::reactive({
            shiny::req(oligoRev())
            shiny::req(alignment())
            checkMatch(oligoRev(), alignment())
        })

        output$overviewTableRev <- DT::renderDataTable(
            {
                shiny::req(is(oligoRev(), "RprimerOligo"))
                x <- roundDbls(removeListColumns(as.data.frame(oligoRev())))
                names(x) <- c(
                    "Type", "Forward", "Reverse", "Start", "End", "Length",
                    "IUPAC sequence, RC", "IUPAC sequence", "Identity",
                    "Coverage", "Degeneracy", "GC content, mean",
                    "GC content, range", "Tm, mean", "Tm, range", "Delta G, mean",
                    "Delta G, range", "Design method", "Score", "ROI, start",
                    "ROI, end"
                )
                x <- x[!names(x) %in% c(
                    "Type", "Forward", "Reverse", "Score", "IUPAC sequence, RC",
                    "ROI, start", "ROI, end"
                )]
                x
            },
            options = list(
                info = FALSE,
                searching = FALSE, paging = FALSE,
                scrollX = TRUE, autoWidth = TRUE,
                ordering = FALSE
            ),
            rownames = FALSE,
            selection = "none"
        )

        output$allVariantTableRev <- DT::renderDataTable(
            {
                shiny::req(is(oligoRev(), "RprimerOligo"))
                x <- roundDbls(makeListTable(as.data.frame(oligoRev())))
                names(x) <- c(
                    "Sequence, RC", "Sequence", "GC content", "Tm", "Delta G"
                )
                x <- x[!names(x) %in% c("Sequence, RC")]
                x
            },
            options = list(
                info = FALSE,
                searching = FALSE, paging = FALSE,
                scrollX = TRUE, autoWidth = FALSE,
                ordering = FALSE
            ),
            rownames = FALSE,
            selection = "none"
        )

        output$bindingRegionPlotRev <- shiny::renderPlot({
            shiny::req(is(oligoRev(), "RprimerOligo"))
            shiny::req(consensus())
            from <- oligoRev()$start
            to <- oligoRev()$end
            bindingRegion <- consensus()[
                consensus()$position >= from & consensus()$position <= to,
            ]
            plotData(bindingRegion, type = "nucleotide", rc = TRUE)
        })

        output$matchTableRev <- DT::renderDataTable(
            {
                shiny::req(matchRev())
                x <- roundDbls(removeListColumns(as.data.frame(matchRev())))
                names(x) <- c(
                    "IUPAC sequence", "Perfect match", "1 mismatch", "2 mismatches",
                    "3 mismatches", "4 or more mismatches",
                    "Off target match (<3 mismatches)"
                )
                x
            },
            options = list(
                info = FALSE,
                searching = FALSE, paging = FALSE,
                scrollX = TRUE, autoWidth = FALSE,
                ordering = FALSE
            ),
            rownames = FALSE,
            selection = "none"
        )

        output$matchPlotRev <- shiny::renderPlot({
            shiny::req(matchRev())
            plotData(matchRev())
        })

        output$matchIdRev <- shiny::renderText({
            shiny::req(matchRev())
            x <- matchRev()
            c(
                "<b>Perfect match</b><br>",
                x$idPerfectMatch[[1]],
                "<br><br><b>One mismatch</b><br>",
                x$idOneMismatch[[1]],
                "<br><br><b>Two mismatches</b><br>",
                x$idTwoMismatches[[1]],
                "<br><br><b>Three mismatches</b><br>",
                x$idThreeMismatches[[1]],
                "<br><br><b>Four or more mismatches</b><br>",
                x$idFourOrMoreMismatches[[1]],
                "<br><br><b>Off target match (< 3 mismatches)</b><br>",
                x$idOffTargetMatch[[1]]
            )
        })

        matchPr <- shiny::reactive({
            shiny::req(oligoPr())
            shiny::req(alignment())
            checkMatch(oligoPr(), alignment())
        })

        output$overviewTablePr <- DT::renderDataTable(
            {
                shiny::req(is(oligoPr(), "RprimerOligo"))
                x <- roundDbls(removeListColumns(as.data.frame(oligoPr())))
                names(x) <- c(
                    "Type", "Plus", "Minus", "Start", "End", "Length",
                    "IUPAC sequence, plus", "IUPAC sequence, minus", "Identity",
                    "Coverage", "Degeneracy", "GC content, mean",
                    "GC content, range", "Tm, mean", "Tm, range", "Delta G, mean",
                    "Delta G, range", "Design method", "Score", "ROI, start",
                    "ROI, end"
                )
                x <- x[!names(x) %in% c(
                    "Type", "Score", "ROI, start", "ROI, end"
                )]
                x
            },
            options = list(
                info = FALSE,
                searching = FALSE, paging = FALSE,
                scrollX = TRUE, autoWidth = TRUE,
                ordering = FALSE
            ),
            rownames = FALSE,
            selection = "none"
        )

        output$allVariantTablePr <- DT::renderDataTable(
            {
                shiny::req(is(oligoPr(), "RprimerOligo"))
                x <- roundDbls(makeListTable(as.data.frame(oligoPr())))
                names(x) <- c(
                    "Sequence, plus", "Sequence, minus", "GC content", "Tm", "Delta G"
                )
                x
            },
            options = list(
                info = FALSE,
                searching = FALSE, paging = FALSE,
                scrollX = TRUE, autoWidth = FALSE,
                ordering = FALSE
            ),
            rownames = FALSE,
            selection = "none"
        )

        output$bindingRegionPlotPr <- shiny::renderPlot({
            shiny::req(is(oligoPr(), "RprimerOligo"))
            shiny::req(consensus())
            from <- oligoPr()$start
            to <- oligoPr()$end
            bindingRegion <- consensus()[
                consensus()$position >= from & consensus()$position <= to,
            ]
            plotData(bindingRegion, type = "nucleotide")
        })

        output$matchTablePr <- DT::renderDataTable(
            {
                shiny::req(matchPr())
                x <- roundDbls(removeListColumns(as.data.frame(matchPr())))
                names(x) <- c(
                    "IUPAC sequence", "Perfect match", "1 mismatch", "2 mismatches",
                    "3 mismatches", "4 or more mismatches",
                    "Off target match (<3 mismatches)"
                )
                x
            },
            options = list(
                info = FALSE,
                searching = FALSE, paging = FALSE,
                scrollX = TRUE, autoWidth = FALSE,
                ordering = FALSE
            ),
            rownames = FALSE,
            selection = "none"
        )

        output$matchPlotPr <- shiny::renderPlot({
            shiny::req(matchPr())
            plotData(matchPr())
        })

        output$matchIdPr <- shiny::renderText({
            shiny::req(matchPr())
            x <- matchPr()
            c(
                "<b>Perfect match</b><br>",
                x$idPerfectMatch[[1]],
                "<br><br><b>One mismatch</b><br>",
                x$idOneMismatch[[1]],
                "<br><br><b>Two mismatches</b><br>",
                x$idTwoMismatches[[1]],
                "<br><br><b>Three mismatches</b><br>",
                x$idThreeMismatches[[1]],
                "<br><br><b>Four or more mismatches</b><br>",
                x$idFourOrMoreMismatches[[1]],
                "<br><br><b>Off target match (< 3 mismatches)</b><br>",
                x$idOffTargetMatch[[1]]
            )
        })

        list(data = shiny::reactive(assay()))
    })
}

dataUI <- function(id) {
    ns <- shiny::NS(id)

    shiny::tagList(
        shiny::h5("File import"),
        shiny::hr(),
        shiny::sidebarLayout(
            shiny::sidebarPanel(
                shiny::radioButtons(
                    ns("dataSelection"),
                    shiny::h6(shiny::tags$b("Option")),
                    choices = c(
                        "Upload alignment", "Use example data"
                    ), selected = "Upload alignment"
                ),
                shiny::conditionalPanel(
                    ns = shiny::NS(id),
                    condition = "input.dataSelection == 'Upload alignment'",
                    shiny::radioButtons(
                        ns("filetype"), shiny::h6(shiny::tags$b("File format")),
                        choices = c("Fasta" = "fasta", "Clustal" = "clustal"),
                        selected = "fasta"
                    ),
                    shiny::fileInput(
                        ns("file"), "",
                        multiple = FALSE,
                        accept = c("text", "fasta")
                    ),
                    shiny::htmlOutput(ns("fileName"))
                ),
                shiny::conditionalPanel(
                    ns = shiny::NS(id),
                    condition = "input.dataSelection == 'Use example data'",
                    shiny::h6("Hepatitis E virus alignment")
                )
            ),
            shiny::mainPanel(
                shiny::htmlOutput(ns("nSequences")),
                shiny::htmlOutput(ns("alnLength"))
            )
        ),
        shiny::hr()
    )
}

dataServer <- function(id) {
    shiny::moduleServer(id, function(input, output, session) {
        options(shiny.maxRequestSize=30*1024^2)

        aln <- shiny::reactive({
            if (input$dataSelection == "Upload alignment") {
                shiny::req(input$file)
                tryCatch(
                    {
                        readDNAMultipleAlignment(
                            input$file$datapath,
                            format = input$filetype
                        )
                    },
                    error = function(cond) {
                        shiny::showNotification(
                            "Failed to import sequence data from the file.\n
                        Please check that the file format is correct.",
                            type = "error",
                            duration = NULL
                        )
                    },
                    silent = TRUE
                )
            } else if (input$dataSelection == "Use example data") {
                utils::data("exampleRprimerAlignment", package = "rprimer")
                exampleRprimerAlignment
            }
        })

        output$fileName <- shiny::renderText({
            shiny::req(is(aln(), "DNAMultipleAlignment"))
            if (input$dataSelection == "Upload alignment") {
                paste0(
                    "'", input$file[[1]], "' was successfully uploaded! "
                )
            } else {
                NULL
            }
        })

        output$nSequences <- shiny::renderText({
            shiny::req(is(aln(), "DNAMultipleAlignment"))
            paste("Number of sequences:", nrow(aln()))
        })

        output$alnLength <- shiny::renderText({
            shiny::req(is(aln(), "DNAMultipleAlignment"))
            paste("Alignment length:", ncol(aln()))
        })

        list(data = shiny::reactive(aln()))
    })
}

consensusUI <- function(id) {
    ns <- shiny::NS(id)

    shiny::tagList(
        shiny::h5("Generate consensus profile (step 1/5)"),
        shiny::hr(),
        shiny::sidebarLayout(
            shiny::sidebarPanel(width = 3,
                shiny::h6(shiny::tags$b("Region of interest")),
                shiny::h6("Specify start and end position of the consensus profile"),
                shiny::uiOutput(ns("roiFrom")),
                shiny::uiOutput(ns("roiTo")),
                shiny::sliderInput(
                    ns("ambiguityThreshold"),
                    shiny::h6(shiny::tags$b("Threshold for ambiguous (degenerate) bases")),
                    value = 0.05, min = 0, max = 0.2,
                ),
                shiny::h6(
                    "The threshold specifies the 'detection level' for
                    ambiguous bases. All DNA bases that occur with a
                    proportion higher than the specified value will be
                    included in the IUPAC consensus character"
                ),
                shiny::h6(shiny::tags$b("Apply mask(s)")),
                shiny::h6(
                    "Masked positions will not appear as oligo binding sites.
                No mask will be applied if 'From' and 'To' are set to 0"
                ),
                shiny::h6("Mask 1"),
                shiny::numericInput(ns("mask1From"), label = "From", value = 0),
                shiny::numericInput(ns("mask1To"), label = "To",  value = 0),
                shiny::h6("Mask 2"),
                shiny::numericInput(ns("mask2From"), label = "From", value = 0),
                shiny::numericInput(ns("mask2To"), label = "To", value = 0),
                shiny::h6("Mask 3"),
                shiny::numericInput(ns("mask3From"), label = "From",  value = 0),
                shiny::numericInput(ns("mask3To"), label = "To", value = 0),
                shiny::br(),
                shiny::actionButton(
                    ns("getConsensusProfile"), "Get consensus profile",
                    class = "btn btn-primary"
                ),
            ),
            shiny::mainPanel(width = 9,
                shiny::tabsetPanel(
                    shiny::tabPanel(
                        title = "Plot",
                        shiny::br(),
                        spinnerPlot(ns("consensusPlot"))
                    ),
                    shiny::tabPanel(
                        title = "Table",
                        shiny::br(),
                        shiny::uiOutput(ns("getDownloadLinkTxt")),
                        shiny::br(),
                        DT::dataTableOutput(ns("consensusTable"))
                    )
                )
            )
        ),
        shiny::hr()
    )
}

consensusServer <- function(id, inputAlignment) {
    shiny::moduleServer(id, function(input, output, session) {
        output$roiFrom <- shiny::renderUI({
            shiny::req(inputAlignment())
            ns <- session$ns
            shiny::numericInput(
                ns("roiFrom"), label = "From",
                min = 1, max = ncol(inputAlignment()) - 1, value = 1,
            )
        })

        output$roiTo <- shiny::renderUI({
            req(inputAlignment())
            ns <- session$ns
            shiny::numericInput(
                ns("roiTo"), label = "To",
                min = 2, max = ncol(inputAlignment()), value = ncol(inputAlignment())
            )
        })

        alignment <- shiny::reactive({
            shiny::req(inputAlignment())
            maskedAln <- inputAlignment()
            mask1 <- arrangeMask(input$mask1From, input$mask1To, inputAlignment())
            mask2 <- arrangeMask(input$mask2From, input$mask2To, inputAlignment())
            mask3 <- arrangeMask(input$mask3From, input$mask3To, inputAlignment())
            mask <- unique(c(mask1, mask2, mask3))
            if (!is.null(mask)) {
                MultipleAlignment::colmask(maskedAln) <- mask
            }
            maskedAln
        })

        consensus <- shiny::eventReactive(input$getConsensusProfile, {
            shiny::req(alignment())
            consensus <- consensusProfile(alignment(), input$ambiguityThreshold)
            consensus[
                consensus$position >= as.numeric(input$roiFrom) &
                    consensus$position <= as.numeric(input$roiTo),
            ]
        })

        output$consensusPlot <- shiny::renderPlot({
            shiny::req(is(consensus(), "RprimerProfile"))
            plotData(consensus())
        })

        output$getDownloadLinkTxt <- shiny::renderUI({
            shiny::req(is(consensus(), "RprimerProfile"))
            ns <- session$ns
            list(
                shiny::downloadLink(
                    ns("downloadTxt"), "Download table as .txt"
                ),
                shiny::br()
            )
        })

        output$downloadTxt <- shiny::downloadHandler(
            filename <- function() {
                paste0("consensus", "-", Sys.Date(), ".txt")
            },
            content <- function(file) {
                utils::write.table(
                    as.data.frame(consensus()), file,
                    quote = FALSE, sep = "\t",
                    row.names = FALSE
                )
            }
        )

        output$consensusTable <- DT::renderDataTable(
            {
                shiny::req(is(consensus(), "RprimerProfile"))
                x <- roundDbls(as.data.frame(consensus()))
                names(x) <- c(
                    "Position", "A", "C", "G", "T", "Other", "Gaps", "Majority",
                    "Identity", "IUPAC", "Coverage"
                )
                x
            },
            options = list(
                info = FALSE,
                searching = FALSE, paging = TRUE,
                pageLength = 100,
                scrollX = TRUE, autoWidth = FALSE,
                ordering = FALSE, scrollY = "1000"
            ),
            rownames = FALSE,
            selection = "none"
        )

        list(data = shiny::reactive(consensus()))
    })
}

#' @details
#' The package contains five functions: \code{consensusProfile()},
#' \code{designOligos()}, \code{designAssays()}, \code{checkMatch()} 
#' and \code{plotData()}
#'
#' @keywords internal
"_PACKAGE"

# The following block is used by usethis to automatically manage
# roxygen namespace tags. Modify with care!
## usethis namespace: start
## usethis namespace: end
NULL

#' Design (RT)-PCR assays
#'
#' \code{designAssays()} combines primers to (RT)-PCR assays from an
#' \code{RprimerOligo} object.
#' If probes are present in the input dataset,
#' only assays with a probe present between the primer pair will be kept.
#'
#' @param x
#' An \code{RprimerOligo} object, which can be with or without probes.
#'
#' @param length
#' Amplicon length range, a numeric vector [40, 5000], defaults to
#' \code{c(65, 120)}.
#'
#' @param tmDifferencePrimers
#' Maximum allowed difference between the mean tm of the forward and reverse
#' primer (in Celcius degrees, as an absolute value). Defaults to \code{NULL},
#' which means that primers will be paired regardless of their tm.
#'
#' @return
#' An \code{RprimerAssay} object, containing the following information:
#'
#' \describe{
#'   \item{start}{Position where the assay starts.}
#'   \item{end}{Position where the assay ends.}
#'   \item{length}{Length of the amplicon.}
#'   \item{totalDegeneracy}{Total number of oligos in the assay.}
#'   \item{score}{Average oligo score. The best
#'   possible score is 0 and the worst possible score is 9.
#'   See \code{?oligos} for more information about the scoring system.}
#'   \item{startFwd}{Start position of the forward primer.}
#'   \item{endFwd}{End positon of the forward primer.}
#'   \item{lengthFwd}{Length of the forward primer.}
#'   \item{iupacSequenceFwd}{Forward primer sequence in IUPAC format
#'   (i.e. with ambiguous bases).}
#'   \item{identityFwd}{For ambiguous primers: average identity of the
#'   forward primer. For mixed primers: average identity of the 5' (consensus)
#'   part of the forward primer. The value can range from 0 to 1.}
#'   \item{coverageFwd}{For ambiguous primers: average coverage of the
#'   forward primer. For mixed primers: average coverage of the 3' (degenerate)
#'   part of the forward primer. The value can range from 0 to 1.}
#'   \item{degeneracyFwd}{Number of sequence variants of the forward primer.}
#'   \item{gcContentMeanFwd}{Mean GC-content of all sequence variants of the
#'   forward primer.}
#'   \item{gcContentRangeFwd}{Range in GC-content of all sequence variants of
#'     the forward primer.}
#'   \item{tmMeanFwd}{Mean tm of all sequence variants of the forward primer
#'   (in Celcius degrees).}
#'   \item{tmRangeFwd}{Range in tm of all sequence variants of the forward
#'   primer (in Celcius degrees).}
#'   \item{deltaGMeanFwd}{Mean delta G of all sequence variants of the
#'   forward primer (in kcal/mol).}
#'   \item{deltaGRangeFwd}{Range in delta G of all sequence variants of the
#'   forward primer (in kcal/mol).}
#'   \item{sequenceFwd}{All sequence variants of the forward primer.}
#'   \item{gcContentFwd}{GC-content of all sequence variants of the forward
#'   primer.}
#'   \item{tmFwd}{Tm of all sequence variants of the forward primer
#'   (in Celcius degrees).}
#'   \item{deltaGFwd}{Delta G of all sequence variants
#'   of the forward primer (in kcal/mol).}
#'   \item{methodFwd}{Design method used to generate the forward
#'   primer: "ambiguous" or "mixedFwd".}
#'   \item{startRev}{Start position of the reverse primer.}
#'   \item{endRev}{End positon of the reverse primer.}
#'   \item{lengthRev}{Length of the reverse primer.}
#'   \item{iupacSequenceRev}{Reverse primer sequence in IUPAC format
#'   (i.e. with ambiguous bases).}
#'   \item{identityRev}{For ambiguous primers: average identity of the
#'   reverse primer. For mixed primers: average identity of the 5' (consensus)
#'   part of the reverse primer. The value can range from 0 to 1.}
#'   \item{coverageRev}{For ambiguous primers: average coverage of the
#'   reverse primer. For mixed primers: average coverage of the 3' (degenerate)
#'   part of the reverse primer. The value can range from 0 to 1.}
#'   \item{degeneracyRev}{Number of sequence variants of the reverse primer.}
#'   \item{gcContentMeanRev}{Mean GC-content of all sequence variants of the
#'   reverse primer.}
#'   \item{gcContentRangeRev}{Range in GC-content of all sequence variants of
#'     the reverse primer.}
#'   \item{tmMeanRev}{Mean tm of all sequence variants of the reverse primer
#'   (in Celcius degrees).}
#'   \item{tmRangeRev}{Range in tm of all sequence variants of the reverse
#'   primer (in Celcius degrees).}
#'   \item{deltaGMeanRev}{Mean delta G of all sequence variants of the
#'   reverse primer (in kcal/mol).}
#'   \item{deltaGRangeRev}{Range in delta G of all sequence variants of the
#'   reverse primer (in kcal/mol).}
#'   \item{sequenceRev}{All sequence variants of the reverse primer.}
#'   \item{gcContentRev}{GC-content of all sequence variants of the reverse
#'   primer.}
#'   \item{tmRev}{Tm of all sequence variants of the reverse primer
#'   (in Celcius degrees).}
#'   \item{deltaGRev}{Delta G of all sequence variants
#'   of the reverse primer (in kcal/mol).}
#'   \item{methodRev}{Design method used to generate the forward
#'   primer: "ambiguous" or "mixedRev.}
#'   \item{roiStart}{Start position of the input consensus profile
#'   used for oligo design.}
#'   \item{roiEnd}{End position of the input consensus profile used
#'   for oligo design.}
#' }
#'
#' If a probe is included in the input \code{RprimerOligo} object,
#' the following columns are also included:
#'
#' \describe{
#'   \item{startPr}{Start position of the probe.}
#'   \item{endPr}{End positon of the probe.}
#'   \item{lengthPr}{Length of the probe.}
#'   \item{iupacSequencePr}{Probe sequence in plus sense, in IUPAC format.}
#'   \item{iupacSequenceRcPr}{Probe sequence in minus sense,
#'   in IUPAC format.}
#'   \item{identityPr}{For ambiguous primers: average identity of the
#'   probe. For mixed primers: average identity of the 5' (consensus)
#'   part of the probe. The value can range from 0 to 1.}
#'   \item{coveragePr}{For ambiguous primers: average coverage of the
#'   probe. For mixed primers: average coverage of the 3' (degenerate)
#'   part of the probe. The value can range from 0 to 1.}
#'   \item{degeneracyPr}{Number of sequence variants of the probe.}
#'   \item{gcContentMeanPr}{Mean GC-content of all sequence variants of the
#'   probe.}
#'   \item{gcContentRangePr}{Range in GC-content of all sequence variants of
#'     the probe.}
#'   \item{tmMeanPr}{Mean tm of all sequence variants of the probe
#'   (in Celcius degrees).}
#'   \item{tmRangePr}{Range in tm of all sequence variants of the forward
#'   primer (in Celcius degrees).}
#'   \item{deltaGMeanPr}{Mean delta G of all sequence variants of the
#'   probe (in kcal/mol).}
#'   \item{deltaGRangePr}{Range in delta G of all sequence variants of the
#'   probe (in kcal/mol).}
#'   \item{sequencePr}{All sequence variants of the probe, in plus sense.}
#'   \item{sequenceRcPr}{All sequence variants of the probe, in minus sense.}
#'   \item{gcContentPr}{GC-content of all sequence variants of the probe.}
#'   \item{tmPr}{Tm of all sequence variants of the probe
#'   (in Celcius degrees).}
#'   \item{deltaGPr}{Delta G of all sequence variants
#'   of the probe (in kcal/mol).}
#'   \item{methodPr}{Design method used to generate the probe.}
#'   \item{plusPr}{If the probe is valid in plus sense.}
#'   \item{minusPr}{If the probe is valid in minus sense.}
#' }
#'
#' An error message will return if no assays are found.
#'
#' @export
#'
#' @examples
#' data("exampleRprimerOligo")
#'
#' ## Design assays using default settings
#' designAssays(exampleRprimerOligo)
#'
#' ## Modify the length range
#' designAssays(exampleRprimerOligo, length = c(1000, 2000))
designAssays <- function(x, length = c(65, 120), tmDifferencePrimers = NULL) {
    if (!methods::is(x, "RprimerOligo")) {
        stop("'x' must be an RprimerOligo object.", call. = FALSE)
    }
    if (!(min(length) >= 40 && max(length) <= 5000)) {
        stop("'length' must be from 40 to 5000.", call. = FALSE)
    }
    if (!is.null(tmDifferencePrimers) && !is.numeric(tmDifferencePrimers)) {
        stop(
            "'tmDifferencePrimers must be either 'NULL' or a number.",
            .call = FALSE
        )
    }
    x <- as.data.frame(x)
    assays <- .combinePrimers(
        x[x$type == "primer", , drop = FALSE],
        length,
        tmDifferencePrimers
    )
    if (any(x$type == "probe")) {
        assays <- .addProbes(assays, x[x$type == "probe", , drop = FALSE])
    }
    assays <- .beautifyPrimers(assays)
    if (any(x$type == "probe")) {
        assays <- .beautifyProbes(assays)
    }
    if (any(x$type == "probe")) {
        assays$score <- assays$score / 3
    } else {
        assays$score <- assays$score / 2
    }
    RprimerAssay(assays)
}

# Helpers ======================================================================

#' @noRd
#'
#' @examples
#' data("exampleRprimerOligo")
#' x <- exampleRprimerOligo
#' .pairPrimers(x)
.pairPrimers <- function(x) {
    fwd <- x[x$fwd, , drop = FALSE]
    rev <- x[x$rev, , drop = FALSE]
    pairs <- expand.grid(
        fwd$iupacSequence, rev$iupacSequenceRc,
        stringsAsFactors = FALSE
    )
    names(pairs) <- c("fwd", "rev")
    fwdMatch <- x[x$method == "ambiguous" | x$method == "mixedFwd", , drop = FALSE]
    fwd <- fwdMatch[match(pairs$fwd, fwdMatch$iupacSequence), , drop = FALSE]
    revMatch <- x[x$method == "ambiguous" | x$method == "mixedRev", , drop = FALSE]
    rev <- revMatch[match(pairs$rev, revMatch$iupacSequenceRc), , drop = FALSE]
    names(fwd) <- paste0(names(fwd), "Fwd")
    names(rev) <- paste0(names(rev), "Rev")
    cbind(fwd, rev)
}

#' @noRd
#'
#' @examples
#' data("exampleRprimerOligo")
#' x <- exampleRprimerOligo
#' .combinePrimers(x)
.combinePrimers <- function(x,
                            length = c(65, 120),
                            tmDifferencePrimers = NULL) {
    assays <- .pairPrimers(x)
    ampLength <- assays$endRev - assays$startFwd + 1
    start <- assays$startFwd
    end <- assays$endRev
    totalDegeneracy <- assays$degeneracyFwd + assays$degeneracyRev
    score <- assays$scoreFwd + assays$scoreRev
    assays <- cbind(
        start, end,
        "length" = ampLength, totalDegeneracy, score,
        assays
    )
    assays <- assays[assays$length >= min(length), , drop = FALSE]
    assays <- assays[assays$length <= max(length), , drop = FALSE]
    if (!is.null(tmDifferencePrimers)) {
        tmDifferencePrimers <- abs(tmDifferencePrimers)
        tmDifference <- abs(assays$tmMeanFwd - assays$tmMeanRev)
        assays <- assays[tmDifference <= tmDifferencePrimers, , drop = FALSE]
    }
    if (nrow(assays) == 0L) {
        stop("No assays were found.", call. = FALSE)
    }
    assays
}

#' @noRd
#'
#' @examples
#' data("exampleRprimerOligo")
#' x <- exampleRprimerOligo
#' assays <- .combinePrimers(x)
#' .identifyProbes(assays, x[x$type == "probe", , drop = FALSE])
.identifyProbes <- function(x, probes) {
    lapply(seq_len(nrow(x)), \(i) {
        from <- x$endFwd[[i]] + 1
        to <- x$startRev[[i]] - 1
        probes[probes$start >= from & probes$end <= to, , drop = FALSE]
    })
}

#' @noRd
#'
#' @examples
#' data("exampleRprimerOligo")
#' x <- exampleRprimerOligo
#' probes <- .identifyProbes(assays, x[x$type == "probe", , drop = FALSE])
#' .extractProbes(assays, probes)
.extractProbes <- function(x, probes) {
    nProbes <- vapply(probes, nrow, integer(1), USE.NAMES = FALSE)
    select <- lapply(seq_along(nProbes), \(x) {
        rep(x, nProbes[[x]])
    })
    select <- unlist(select)
    x <- x[select, , drop = FALSE]
    if (nrow(x) == 0L) {
        stop("No assays with probes could be generated.", call. = FALSE)
    }
    probes <- do.call("rbind", probes)
    names(probes) <- paste0(names(probes), "Pr")
    x <- cbind(x, probes)
    x$totalDegeneracy <- x$totalDegeneracy + probes$degeneracyPr
    x$score <- x$score + probes$scorePr
    if (nrow(x) == 0L) {
        stop("No assays with probes could be generated.", call. = FALSE)
    }
    x
}

#' @noRd
#'
#' @examples
#' data("exampleRprimerOligo")
#' x <- exampleRprimerOligo
#' assays <- .combinePrimers(x)
#' .addProbes(assays, x[x$type == "probe", , drop = FALSE])
.addProbes <- function(x, probes) {
    probeCandidates <- .identifyProbes(x, probes)
    .extractProbes(x, probeCandidates)
}

#' @noRd
.beautifyPrimers <- function(x) {
    drop <- c(
        "typeFwd", "fwdFwd", "revFwd", "typeRev", "fwdRev", "revRev",
        "iupacSequenceRcFwd", "sequenceRcFwd", "iupacSequenceRev",
        "sequenceRev", "scoreFwd", "scoreRev", "roiStartFwd", "roiEndFwd"
    )
    x <- x[!(names(x) %in% drop)]
    names(x)[
        names(x) %in% c("iupacSequenceRcRev", "sequenceRcRev")
    ] <- c("iupacSequenceRev", "sequenceRev")
    names(x)[
        names(x) %in% c("roiStartRev", "roiEndRev")
    ] <- c("roiStart", "roiEnd")
    x <- x[order(x$start), ]
    rownames(x) <- NULL
    x
}

#' @noRd
.beautifyProbes <- function(x) {
    drop <- c("typePr", "scorePr", "roiStartPr", "roiEndPr")
    x <- x[!(names(x) %in% drop)]
    rename <- c("fwdPr", "revPr")
    names(x)[names(x) %in% rename] <- c("plusPr", "minusPr")
    moveToLast <- c("roiStart", "roiEnd")
    x <- x[c(setdiff(names(x), moveToLast), moveToLast)]
    moveToFirst <- c("start", "end", "length")
    x <- x[c(moveToFirst, setdiff(names(x), moveToFirst))]
    x
}

## Import data to test on
data("exampleRprimerOligo")
x <- exampleRprimerOligo

# assays =======================================================================

test_that("designAssays returns an error when it sould", {
    expect_error(designAssays(unclass(x)))
    expect_error(designAssays(x, length = c(39, 120)))
    expect_error(designAssays(x, length = c(40, 50001)))
    expect_error(designAssays(x, tmDifferencePrimers = FALSE))
})

test_that("designAssays works", {
    test <- designAssays(x)
    expect_s4_class(test, "RprimerAssay")
})

# .pairPrimers =================================================================

# .combinePrimers ==============================================================

test_that(".combinePrimers works", {
    test <- .combinePrimers(x,
        length = c(100, 2000), tmDifferencePrimers = 2
    )
    expect_true(all(test$length >= 100))
    expect_true(all(test$length <= 2000))
    expect_equal(test$end - test$start, test$length - 1)
    expect_error(.combinePrimers(x[1, ]))
    expect_true(all(abs(test$tmMeanFwd - test$tmMeanRev) <= 2))
})

# .identifyProbes ==============================================================

test_that(".identifyProbes works", {
    assays <- .combinePrimers(x)
    test <- .identifyProbes(assays, x[x$type == "probe", ])
    expect_equal(nrow(assays), length(test))
})

# .extractProbes ===============================================================

test_that(".extractProbes works", {
    assays <- .combinePrimers(x)
    probes <- .identifyProbes(assays, x[x$type == "probe", ])
    test <- .extractProbes(assays, probes)
    expect_true(all(test$startPr - test$endFwd >= 1))
    expect_true(all(test$startRev - test$endPr >= 1))
})

# .addProbes ===================================================================

# .beautifyPrimers =============================================================

# .beautifyProbes ==============================================================

oligoUI <- function(id) {
    ns <- shiny::NS(id)

    shiny::tagList(
        shiny::h5("Design oligos (step 2/5)"),
        shiny::hr(),
        shiny::sidebarLayout(
            shiny::sidebarPanel(
                width = 3,
                shiny::h5(shiny::tags$b("Primer settings")),
                shiny::hr(),
                shiny::sliderInput(
                    ns("lengthPrimer"),
                    "Length",
                    value = c(18, 22), min = 15, max = 40
                ),
                shiny::numericInput(
                    ns("maxDegeneracyPrimer"),
                    "Maximum degeneracy (number of sequence variants, 1-64)",
                    value = 4, min = 1, max = 64
                ),
                shiny::checkboxInput(
                    ns("avoidThreeEndRunsPrimer"),
                    "Avoid 3' end runs",
                    value = TRUE
                ),
                shiny::h6(
                    "If primers with more than two runs of the same
                          nucleotide at the terminal 3' end should be avoided
                          (to reduce the risk of mispriming)"
                ),
                shiny::br(),
                shiny::checkboxInput(
                    ns("gcClampPrimer"),
                    "Use GC-clamp",
                    value = TRUE
                ),
                shiny::h6(
                    "A GC-clamp is identified as two to three G or C:s
                    within the last five bases (3' end) of the primer"
                ),
                shiny::br(),
                shiny::sliderInput(
                    ns("gcPrimer"),
                    "GC-content range",
                    value = c(0.4, 0.65), min = 0, max = 1
                ),
                shiny::sliderInput(
                    ns("tmPrimer"),
                    "Melting temperature (Tm) range (Celcius degrees)",
                    value = c(50, 65), min = 20, max = 90
                ),
                shiny::numericInput(
                    ns("concPrimer"),
                    "Primer concentration (20-2000 nM) (for Tm calculation)",
                    value = 500, min = 20, max = 2000
                ),
                shiny::radioButtons(
                    ns("designStrategyPrimer"),
                    "Primer design strategy",
                    choices = c(
                        "Ambiguous" = "ambiguous",
                        "Mixed" = "mixed"
                    ),
                    selected = "ambiguous"
                ),
                shiny::h6(
                    "Ambigous primers can have degenerate bases at any position"
                ),
                shiny::h6(
                    "Mixed primers consist of a shorter degenerate part at the 3' end
                    (1/3 of the primer) and a longer consensus part at the 5'
                    end (2/3 of the primer), which instead of having ambiguous
                    bases contains the most frequently occuring nucleotides"
                ),
                shiny::br(),
                shiny::h5(shiny::tags$b("Probe settings")),
                shiny::hr(),
                shiny::checkboxInput(
                    ns("probe"),
                    "Design probes",
                    value = FALSE
                ),
                shiny::conditionalPanel(
                    ns = shiny::NS(id),
                    condition = "input.probe == true",
                    shiny::sliderInput(
                        ns("lengthProbe"),
                        "Length",
                        value = c(18, 22), min = 15, max = 40
                    ),
                    shiny::numericInput(
                        ns("maxDegeneracyProbe"),
                        "Maximum degeneracy (number of sequence variants, 1-64)",
                        value = 4, min = 1, max = 64
                    ),
                    shiny::checkboxInput(
                        ns("avoidFiveEndGProbe"),
                        "Avoid probes with a 5' end G",
                        value = TRUE
                    ),
                    shiny::sliderInput(
                        ns("gcProbe"),
                        "GC-content range",
                        value = c(0.4, 0.65), min = 0, max = 1,
                    ),
                    shiny::sliderInput(
                        ns("tmProbe"),
                        "Melting temperature range (Celcius degrees)",
                        value = c(50, 70), min = 20, max = 90
                    ),
                    shiny::numericInput(
                        ns("concProbe"),
                        "Probe concentration (20-2000 nM) (for Tm calculation)",
                        value = 250, min = 20, max = 2000
                    )
                ),
                shiny::br(),
                shiny::h5(shiny::tags$b("General settings")),
                shiny::hr(),
                shiny::numericInput(
                    ns("maxGapFrequency"),
                    "Maximum allowed gap proportion at binding sites in
                        target alignment (0-1)",
                    value = 0.01, min = 0, max = 0.2
                ),
                shiny::numericInput(
                    ns("concNa"),
                    "Sodium ion concentration (0.01-1 M) (for Tm calculation)",
                    value = 0.05, min = 0, max = 1
                ),
                shiny::hr(),
                shiny::actionButton(
                    ns("getOligos"), "Get oligos",
                    class = "btn btn-primary"
                )
            ),
            shiny::mainPanel(width = 9,
                             shiny::tabsetPanel(
                                 id = ns("wizard"),
                                 shiny::tabPanel(
                                     title = "Plot",
                                     shiny::br(),
                                     spinnerPlot(
                                         ns("oligoPlot"),
                                         width = "100%", height = 600
                                     )
                                 ),
                                 shiny::tabPanel(
                                     title = "Table",
                                     shiny::br(),
                                     shiny::uiOutput(ns("getDownloadLinkTxt")),
                                     shiny::uiOutput(ns("getDownloadLinkFasta")),
                                     shiny::br(),
                                     shiny::h6("Select a row for more details"),
                                     shiny::br(),
                                     DT::dataTableOutput(ns("oligoTable"))
                                 ),
                                 shiny::tabPanel(
                                     title = "Selection",
                                     shiny::br(),
                                     shiny::tabsetPanel(
                                         type = "pills",
                                         shiny::tabPanel(
                                             title = "Oligo information",
                                             shiny::br(),
                                             shiny::uiOutput(ns("getDownloadLinkTxtSel")),
                                             shiny::uiOutput(ns("getDownloadLinkFastaSel")),
                                             shiny::br(),
                                             shiny::h6(shiny::tags$b("Overview")),
                                             DT::dataTableOutput(ns("overviewTable")),
                                             shiny::br(),
                                             shiny::h6(shiny::tags$b("All sequence variants")),
                                             DT::dataTableOutput(ns("allVariantTable")),
                                             shiny::br(),
                                             shiny::h6(shiny::tags$b("Nucleotide distribution in target alignment")),
                                             shiny::hr(),
                                             shiny::br(),
                                             shiny::column(
                                                 width = 12, align = "center",
                                                 spinnerPlot(
                                                     ns("bindingRegionPlot"),
                                                     width = "75%"
                                                 )
                                             )
                                         ),
                                         shiny::tabPanel(
                                             title = "Match details",
                                             shiny::br(),
                                             shiny::h6(shiny::tags$b("Proportion of matching sequences")),
                                             shiny::br(),
                                             DT::dataTableOutput(ns("matchTable")),
                                             shiny::br(),
                                             shiny::column(
                                                 width = 12, align = "center",
                                                 spinnerPlot(ns("matchPlot"), width = "75%")
                                             ),
                                             shiny::br(),
                                             shiny::h6(shiny::tags$b("Sequence names")),
                                             shiny::hr(),
                                             shiny::htmlOutput(ns("matchId"))
                                         )
                                     )
                                 )
                             )
            )
        ),
        shiny::hr()
    )
}

oligoServer <- function(id, alignment, consensus) {
    shiny::moduleServer(id, function(input, output, session) {
        oligo <- shiny::eventReactive(input$getOligos, {
            shiny::req(consensus())

            validDegenPrimer <- input$maxDegeneracyPrimer >= 1 &&
                input$maxDegeneracyPrimer <= 64
            shinyFeedback::feedbackDanger(
                "maxDegeneracyPrimer", !validDegenPrimer,
                "Enter a value from 1 to 64"
            )

            validDegenProbe <- input$maxDegeneracyProbe >= 1 &&
                input$maxDegeneracyProbe <= 64
            shinyFeedback::feedbackDanger(
                "maxDegeneracyProbe", !validDegenProbe,
                "Enter a value from 1 to 64"
            )

            validConcPrimer <- input$concPrimer >= 20 &&
                input$concPrimer <= 2000
            shinyFeedback::feedbackDanger(
                "concPrimer", !validConcPrimer,
                "Enter a value from 20 to 2000"
            )

            validConcProbe <- input$concProbe >= 20 &&
                input$concProbe <= 2000
            shinyFeedback::feedbackDanger(
                "concProbe", !validConcProbe,
                "Enter a value from 20 to 2000"
            )

            validConcNa <- input$concNa >= 0.01 && input$concNa <= 1
            shinyFeedback::feedbackDanger(
                "concNa", !validConcNa,
                "Enter a value from 0.01 to 1"
            )

            validGap <- input$maxGapFrequency >= 0 &&
                input$maxGapFrequency <= 1
            shinyFeedback::feedbackDanger(
                "maxGapFrequency", !validGap,
                "Enter a value from 0 to 1"
            )

            tryCatch(
                {
                    designOligos(consensus(),
                                 maxGapFrequency = input$maxGapFrequency,
                                 lengthPrimer = input$lengthPrimer,
                                 maxDegeneracyPrimer = input$maxDegeneracyPrimer,
                                 avoidThreeEndRunsPrimer = input$avoidThreeEndRunsPrimer,
                                 gcClampPrimer = input$gcClampPrimer,
                                 gcPrimer = input$gcPrimer,
                                 tmPrimer = input$tmPrimer,
                                 concPrimer = input$concPrimer,
                                 designStrategyPrimer = input$designStrategyPrimer,
                                 probe = input$probe,
                                 lengthProbe = input$lengthProbe,
                                 maxDegeneracyProbe = input$maxDegeneracyProbe,
                                 avoidFiveEndGProbe = input$avoidFiveEndGProbe,
                                 gcProbe = input$gcProbe,
                                 tmProbe = input$tmProbe,
                                 concNa = input$concNa
                    )
                },
                error = function(cond) {
                    if (grepl("must", cond)) {
                        shiny::showNotification(
                            "At least one design setting was invalid. \n
Please review.",
                            type = "error",
                            duration = NULL
                        )
                    } else {
                        shiny::showNotification(
                            "No oligos were found.\n
Try to adjust design settings.",
                            type = "error",
                            duration = NULL
                        )
                    }
                },
                silent = TRUE
            )
        })

        output$oligoPlot <- shiny::renderPlot({
            shiny::req(is(oligo(), "RprimerOligo"))
            plotData(oligo())
        })

        output$getDownloadLinkTxt <- shiny::renderUI({
            shiny::req(is(oligo(), "RprimerOligo"))
            ns <- session$ns
            list(
                shiny::downloadLink(
                    ns("downloadTxt"), "Download table as .txt"
                ),
                shiny::br()
            )
        })

        output$getDownloadLinkFasta <- shiny::renderUI({
            shiny::req(is(oligo(), "RprimerOligo"))
            ns <- session$ns
            list(
                shiny::downloadLink(
                    ns("downloadFasta"), "Download sequence(s) in fasta-format"
                ),
                shiny::br()
            )
        })

        output$downloadTxt <- shiny::downloadHandler(
            filename <- function() {
                paste0("oligo", "-", Sys.Date(), ".txt")
            },
            content <- function(file) {
                utils::write.table(
                    as.data.frame(oligo()), file,
                    quote = FALSE, sep = "\t",
                    row.names = FALSE
                )
            }
        )

        output$downloadFasta <- shiny::downloadHandler(
            filename <- function() {
                paste0("oligo", "-fasta-", Sys.Date(), ".txt")
            },
            content <- function(file) {
                data <- as(oligo(), "DNAStringSet")
                Biostrings::writeXStringSet(data, file)
            }
        )

        output$oligoTable <- DT::renderDataTable(
            {
                shiny::req(is(oligo(), "RprimerOligo"))
                x <- roundDbls(removeListColumns(as.data.frame(oligo())))
                names(x) <- c(
                    "Type", "Forward", "Reverse", "Start", "End", "Length",
                    "IUPAC sequence", "IUPAC sequence, RC", "Identity",
                    "Coverage", "Degeneracy", "GC content, mean",
                    "GC content, range", "Tm, mean", "Tm, range", "Delta G, mean",
                    "Delta G, range", "Design method", "Score", "ROI, start",
                    "ROI, end"
                )
                if (is.na(x$Score[[1]])) {
                    x <- x[!names(x) %in% "Score"]
                }
                x
            },
            options = list(
                info = FALSE,
                searching = FALSE, paging = TRUE,
                pageLength = 100,
                scrollX = TRUE, autoWidth = TRUE,
                ordering = TRUE, scrollY = "1000"
            ),
            rownames = FALSE,
            selection = list(mode = "single")
        )

        shiny::observeEvent(input$oligoTable_rows_selected, {
            Sys.sleep(1)
            shiny::updateTabsetPanel(session, "wizard", selected = "Selection")
        })

        selectedOligo <- shiny::reactive({
            shiny::req(is(oligo(), "RprimerOligo"))
            if (!is.null(input$oligoTable_rows_selected)) {
                oligo()[input$oligoTable_rows_selected, ]
            } else {
                NULL
            }
        })


        match <- shiny::reactive({
            shiny::req(selectedOligo())
            shiny::req(alignment())
            checkMatch(selectedOligo(), alignment())
        })

        output$getDownloadLinkTxtSel <- shiny::renderUI({
            shiny::req(is(selectedOligo(), "RprimerOligo"))
            ns <- session$ns
            list(
                shiny::downloadLink(
                    ns("downloadTxtSel"), "Download table as .txt"
                ),
                shiny::br()
            )
        })

        output$getDownloadLinkFastaSel <- shiny::renderUI({
            shiny::req(is(selectedOligo(), "RprimerOligo"))
            ns <- session$ns
            list(
                shiny::downloadLink(
                    ns("downloadFastaSel"), "Download sequence(s) in fasta-format"
                ),
                shiny::br()
            )
        })

        output$downloadTxtSel <- shiny::downloadHandler(
            filename <- function() {
                paste0("oligo-selection", "-", Sys.Date(), ".txt")
            },
            content <- function(file) {
                utils::write.table(
                    as.data.frame(selectedOligo()), file,
                    quote = FALSE, sep = "\t",
                    row.names = FALSE
                )
            }
        )

        output$downloadFastaSel <- shiny::downloadHandler(
            filename <- function() {
                paste0("oligo-selection", "-fasta-", Sys.Date(), ".txt")
            },
            content <- function(file) {
                data <- as(selectedOligo(), "DNAStringSet")
                Biostrings::writeXStringSet(data, file)
            }
        )

        output$overviewTable <- DT::renderDataTable(
            {
                shiny::req(is(selectedOligo(), "RprimerOligo"))
                x <- roundDbls(removeListColumns(as.data.frame(selectedOligo())))
                names(x) <- c(
                    "Type", "Forward", "Reverse", "Start", "End", "Length",
                    "IUPAC sequence", "IUPAC sequence, RC", "Identity",
                    "Coverage", "Degeneracy", "GC content, mean",
                    "GC content, range", "Tm, mean", "Tm, range", "Delta G, mean",
                    "Delta G, range", "Design method", "Score", "ROI, start",
                    "ROI, end"
                )
                x
            },
            options = list(
                info = FALSE,
                searching = FALSE, paging = FALSE,
                scrollX = TRUE, autoWidth = TRUE,
                ordering = FALSE
            ),
            rownames = FALSE,
            selection = "none"
        )

        output$allVariantTable <- DT::renderDataTable(
            {
                shiny::req(is(selectedOligo(), "RprimerOligo"))
                x <- roundDbls(makeListTable(as.data.frame(selectedOligo())))
                names(x) <- c(
                    "Sequence", "Sequence, RC", "GC content", "Tm", "Delta G"
                )
                x
            },
            options = list(
                info = FALSE,
                searching = FALSE, paging = FALSE,
                scrollX = TRUE, autoWidth = FALSE,
                ordering = FALSE
            ),
            rownames = FALSE,
            selection = "none"
        )

        output$bindingRegionPlot <- shiny::renderPlot({
            shiny::req(is(selectedOligo(), "RprimerOligo"))
            shiny::req(consensus())
            from <- selectedOligo()$start
            to <- selectedOligo()$end
            bindingRegion <- consensus()[
                consensus()$position >= from & consensus()$position <= to,
            ]
            plotData(bindingRegion, type = "nucleotide")
        })

        output$matchTable <- DT::renderDataTable(
            {
                shiny::req(match())
                x <- roundDbls(removeListColumns(as.data.frame(match())))
                names(x) <- c(
                    "IUPAC sequence", "Perfect match", "1 mismatch", "2 mismatches",
                    "3 mismatches", "4 or more mismatches",
                    "Off target match (<3 mismatches)"
                )
                x
            },
            options = list(
                info = FALSE,
                searching = FALSE, paging = FALSE,
                scrollX = TRUE, autoWidth = FALSE,
                ordering = FALSE
            ),
            rownames = FALSE,
            selection = "none"
        )

        output$matchPlot <- shiny::renderPlot({
            shiny::req(match())
            plotData(match())
        })

        output$matchId <- shiny::renderText({
            shiny::req(match())
            x <- match()
            c(
                "<b>Perfect match</b><br>",
                x$idPerfectMatch[[1]],
                "<br><br><b>One mismatch</b><br>",
                x$idOneMismatch[[1]],
                "<br><br><b>Two mismatches</b><br>",
                x$idTwoMismatches[[1]],
                "<br><br><b>Three mismatches</b><br>",
                x$idThreeMismatches[[1]],
                "<br><br><b>Four or more mismatches</b><br>",
                x$idFourOrMoreMismatches[[1]],
                "<br><br><b>Off target match (< 3 mismatches)</b><br>",
                x$idOffTargetMatch[[1]]
            )
        })

        list(data = shiny::reactive(oligo()))
    })
}

oligoFilterUI <- function(id) {
    ns <- shiny::NS(id)

    shiny::tagList(
        shiny::h5("Filter oligos (step 3/5)"),
        shiny::hr(),
        shiny::sidebarLayout(
            shiny::sidebarPanel(
                width = 3,
                shiny::uiOutput(ns("fwd")),
                shiny::uiOutput(ns("rev")),
                shiny::uiOutput(ns("pr"))
            ),
            shiny::mainPanel(
                width = 9,
                shiny::tabsetPanel(
                    id = ns("wizard"),
                    shiny::tabPanel(
                        title = "Plot",
                        shiny::br(),
                        spinnerPlot(
                            ns("oligoPlot"),
                            width = "100%", height = 600
                        )
                    ),
                    shiny::tabPanel(
                        title = "Table",
                        shiny::br(),
                        shiny::uiOutput(ns("getDownloadLinkTxt")),
                        shiny::uiOutput(ns("getDownloadLinkFasta")),
                        shiny::br(),
                        shiny::h6("Select a row for more details"),
                        shiny::br(),
                        DT::dataTableOutput(ns("oligoTable"))
                    ),
                    shiny::tabPanel(
                        title = "Selection",
                        shiny::br(),
                        shiny::tabsetPanel(
                            type = "pills",
                            shiny::tabPanel(
                                title = "Oligo information",
                                shiny::br(),
                                shiny::uiOutput(ns("getDownloadLinkTxtSel")),
                                shiny::uiOutput(ns("getDownloadLinkFastaSel")),
                                shiny::br(),
                                shiny::h6(shiny::tags$b("Overview")),
                                DT::dataTableOutput(ns("overviewTable")),
                                shiny::br(),
                                shiny::h6(shiny::tags$b("All sequence variants")),
                                DT::dataTableOutput(ns("allVariantTable")),
                                shiny::br(),
                                shiny::h6(shiny::tags$b("Nucleotide distribution in target alignment")),
                                shiny::hr(),
                                shiny::br(),
                                shiny::column(
                                    width = 12, align = "center",
                                    spinnerPlot(
                                        ns("bindingRegionPlot"),
                                        width = "75%"
                                    )
                                )
                            ),
                            shiny::tabPanel(
                                title = "Match details",
                                shiny::br(),
                                shiny::h6(shiny::tags$b("Proportion of matching sequences")),
                                shiny::br(),
                                DT::dataTableOutput(ns("matchTable")),
                                shiny::br(),
                                shiny::column(
                                    width = 12, align = "center",
                                    spinnerPlot(ns("matchPlot"), width = "75%")
                                ),
                                shiny::br(),
                                shiny::h6(shiny::tags$b("Sequence names")),
                                shiny::hr(),
                                shiny::htmlOutput(ns("matchId"))
                            )
                        )
                    )
                )
            )
        ),
        shiny::hr()
    )
}

oligoFilterServer <- function(id, alignment, consensus, allOligos) {
    shiny::moduleServer(id, function(input, output, session) {
        output$fwd <- shiny::renderUI({
            shiny::req(any(allOligos()$type == "primer" & allOligos()$fwd))

            ns <- session$ns

            list(
                shiny::h6(shiny::tags$b("Forward")),
                shiny::hr(),
                numericInputFrom(allOligos(), ns("fwdRegionFrom")),
                numericInputTo(allOligos(), ns("fwdRegionTo")),
                conservationInput(
                    allOligos(), ns("minOligoIdentityFwd"),
                    direction = "fwd", variable = "identity"
                ),
                conservationInput(
                    allOligos(), ns("minOligoCoverageFwd"),
                    direction = "fwd", variable = "coverage"
                )
            )
        })

        output$rev <- shiny::renderUI({
            shiny::req(any(allOligos()$type == "primer" & allOligos()$rev))

            ns <- session$ns

            list(
                shiny::h6(shiny::tags$b("Reverse")),
                shiny::hr(),
                numericInputFrom(allOligos(), ns("revRegionFrom")),
                numericInputTo(allOligos(), ns("revRegionTo")),
                conservationInput(
                    allOligos(), ns("minOligoIdentityRev"),
                    direction = "rev", variable = "identity"
                ),
                conservationInput(
                    allOligos(), ns("minOligoCoverageRev"),
                    direction = "rev", variable = "coverage"
                )
            )
        })

        output$pr <- shiny::renderUI({
            shiny::req(any(allOligos()$type == "probe"))

            ns <- session$ns

            list(
                shiny::h6(shiny::tags$b("Probe")),
                shiny::hr(),
                numericInputFrom(allOligos(), ns("prRegionFrom")),
                numericInputTo(allOligos(), ns("prRegionTo")),
                conservationInput(
                    allOligos(), ns("minOligoIdentityPr"),
                    type = "probe", variable = "identity"
                ),
                conservationInput(
                    allOligos(), ns("minOligoCoveragePr"),
                    type = "probe", variable = "coverage"
                )
            )
        })

        oligo <- shiny::reactive({
            shiny::req(allOligos())
            filterOligos(allOligos(),
                fwdFrom = input$fwdRegionFrom,
                fwdTo = input$fwdRegionTo,
                revFrom = input$revRegionFrom,
                revTo = input$revRegionTo,
                prFrom = input$prRegionFrom,
                prTo = input$prRegionTo,
                fwdIdentity = input$minOligoIdentityFwd,
                revIdentity = input$minOligoIdentityRev,
                prIdentity = input$minOligoIdentityPr,
                fwdCoverage = input$minOligoCoverageFwd,
                revCoverage = input$minOligoCoverageRev,
                prCoverage = input$minOligoCoveragePr
            )
        })

        output$oligoPlot <- shiny::renderPlot({
            shiny::req(is(oligo(), "RprimerOligo"))
            plotData(oligo())
        })

        output$getDownloadLinkTxt <- shiny::renderUI({
            shiny::req(is(oligo(), "RprimerOligo"))
            ns <- session$ns
            list(
                shiny::downloadLink(
                    ns("downloadTxt"), "Download table as .txt"
                ),
                shiny::br()
            )
        })

        output$getDownloadLinkFasta <- shiny::renderUI({
            shiny::req(is(oligo(), "RprimerOligo"))
            ns <- session$ns
            list(
                shiny::downloadLink(
                    ns("downloadFasta"), "Download sequence(s) in fasta-format"
                ),
                shiny::br()
            )
        })

        output$downloadTxt <- shiny::downloadHandler(
            filename <- function() {
                paste0("oligo", "-", Sys.Date(), ".txt")
            },
            content <- function(file) {
                utils::write.table(
                    as.data.frame(oligo()), file,
                    quote = FALSE, sep = "\t",
                    row.names = FALSE
                )
            }
        )

        output$downloadFasta <- shiny::downloadHandler(
            filename <- function() {
                paste0("oligo", "-fasta-", Sys.Date(), ".txt")
            },
            content <- function(file) {
                data <- as(oligo(), "DNAStringSet")
                Biostrings::writeXStringSet(data, file)
            }
        )

        output$oligoTable <- DT::renderDataTable(
            {
                shiny::req(is(oligo(), "RprimerOligo"))
                x <- roundDbls(removeListColumns(as.data.frame(oligo())))
                names(x) <- c(
                    "Type", "Forward", "Reverse", "Start", "End", "Length",
                    "IUPAC sequence", "IUPAC sequence, RC", "Identity",
                    "Coverage", "Degeneracy", "GC content, mean",
                    "GC content, range", "Tm, mean", "Tm, range", "Delta G, mean",
                    "Delta G, range", "Design method", "Score", "ROI, start",
                    "ROI, end"
                )
                if (is.na(x$Score[[1]])) {
                    x <- x[!names(x) %in% "Score"]
                }
                x
            },
            options = list(
                info = FALSE,
                searching = FALSE, paging = TRUE,
                pageLength = 100,
                scrollX = TRUE, autoWidth = TRUE,
                ordering = TRUE, scrollY = "1000"
            ),
            rownames = FALSE,
            selection = list(mode = "single")
        )

        shiny::observeEvent(input$oligoTable_rows_selected, {
            Sys.sleep(1)
            shiny::updateTabsetPanel(session, "wizard", selected = "Selection")
        })

        selectedOligo <- shiny::reactive({
            shiny::req(is(oligo(), "RprimerOligo"))
            if (!is.null(input$oligoTable_rows_selected)) {
                oligo()[input$oligoTable_rows_selected, ]
            } else {
                NULL
            }
        })


        match <- shiny::reactive({
            shiny::req(selectedOligo())
            shiny::req(alignment())
            checkMatch(selectedOligo(), alignment())
        })

        output$getDownloadLinkTxtSel <- shiny::renderUI({
            shiny::req(is(selectedOligo(), "RprimerOligo"))
            ns <- session$ns
            list(
                shiny::downloadLink(
                    ns("downloadTxtSel"), "Download table as .txt"
                ),
                shiny::br()
            )
        })

        output$getDownloadLinkFastaSel <- shiny::renderUI({
            shiny::req(is(selectedOligo(), "RprimerOligo"))
            ns <- session$ns
            list(
                shiny::downloadLink(
                    ns("downloadFastaSel"), "Download sequence(s) in fasta-format"
                ),
                shiny::br()
            )
        })

        output$downloadTxtSel <- shiny::downloadHandler(
            filename <- function() {
                paste0("oligo-selection", "-", Sys.Date(), ".txt")
            },
            content <- function(file) {
                utils::write.table(
                    as.data.frame(selectedOligo()), file,
                    quote = FALSE, sep = "\t",
                    row.names = FALSE
                )
            }
        )

        output$downloadFastaSel <- shiny::downloadHandler(
            filename <- function() {
                paste0("oligo-selection", "-fasta-", Sys.Date(), ".txt")
            },
            content <- function(file) {
                data <- as(selectedOligo(), "DNAStringSet")
                Biostrings::writeXStringSet(data, file)
            }
        )

        output$overviewTable <- DT::renderDataTable(
            {
                shiny::req(is(selectedOligo(), "RprimerOligo"))
                x <- roundDbls(removeListColumns(as.data.frame(selectedOligo())))
                names(x) <- c(
                    "Type", "Forward", "Reverse", "Start", "End", "Length",
                    "IUPAC sequence", "IUPAC sequence, RC", "Identity",
                    "Coverage", "Degeneracy", "GC content, mean",
                    "GC content, range", "Tm, mean", "Tm, range", "Delta G, mean",
                    "Delta G, range", "Design method", "Score", "ROI, start",
                    "ROI, end"
                )
                x
            },
            options = list(
                info = FALSE,
                searching = FALSE, paging = FALSE,
                scrollX = TRUE, autoWidth = TRUE,
                ordering = FALSE
            ),
            rownames = FALSE,
            selection = "none"
        )

        output$allVariantTable <- DT::renderDataTable(
            {
                shiny::req(is(selectedOligo(), "RprimerOligo"))
                x <- roundDbls(makeListTable(as.data.frame(selectedOligo())))
                names(x) <- c(
                    "Sequence", "Sequence, RC", "GC content", "Tm", "Delta G"
                )
                x
            },
            options = list(
                info = FALSE,
                searching = FALSE, paging = FALSE,
                scrollX = TRUE, autoWidth = FALSE,
                ordering = FALSE
            ),
            rownames = FALSE,
            selection = "none"
        )

        output$bindingRegionPlot <- shiny::renderPlot({
            shiny::req(is(selectedOligo(), "RprimerOligo"))
            shiny::req(consensus())
            from <- selectedOligo()$start
            to <- selectedOligo()$end
            bindingRegion <- consensus()[
                consensus()$position >= from & consensus()$position <= to,
            ]
            plotData(bindingRegion, type = "nucleotide")
        })

        output$matchTable <- DT::renderDataTable(
            {
                shiny::req(match())
                x <- roundDbls(removeListColumns(as.data.frame(match())))
                names(x) <- c(
                    "IUPAC sequence", "Perfect match", "1 mismatch", "2 mismatches",
                    "3 mismatches", "4 or more mismatches",
                    "Off target match (<3 mismatches)"
                )
                x
            },
            options = list(
                info = FALSE,
                searching = FALSE, paging = FALSE,
                scrollX = TRUE, autoWidth = FALSE,
                ordering = FALSE
            ),
            rownames = FALSE,
            selection = "none"
        )

        output$matchPlot <- shiny::renderPlot({
            shiny::req(match())
            plotData(match())
        })

        output$matchId <- shiny::renderText({
            shiny::req(match())
            x <- match()
            c(
                "<b>Perfect match</b><br>",
                x$idPerfectMatch[[1]],
                "<br><br><b>One mismatch</b><br>",
                x$idOneMismatch[[1]],
                "<br><br><b>Two mismatches</b><br>",
                x$idTwoMismatches[[1]],
                "<br><br><b>Three mismatches</b><br>",
                x$idThreeMismatches[[1]],
                "<br><br><b>Four or more mismatches</b><br>",
                x$idFourOrMoreMismatches[[1]],
                "<br><br><b>Off target match (< 3 mismatches)</b><br>",
                x$idOffTargetMatch[[1]]
            )
        })

        list(data = shiny::reactive(oligo()))
    })
}

# check only on unique seq varianrs


#' Check how oligos and assays match to their target sequences
#'
#' \code{checkMatch()} investigates how well oligos or assays match with their
#' intended target sequences
#' within a multiple DNA sequence alignment.
#'
#' The output provides information on the proportion and names of target
#' sequences
#' that match perfectly as well as with one, two, three, or four or more
#' mismatches to the oligo within the intended oligo binding region
#' in the input alignment (on-target match).
#' It also gives the proportion and names of target sequences that
#' match with a maximum of two mismatches to at least one sequence variant of
#' the oligo outside the
#' intended oligo binding region (off-target match).
#' The function is a wrapper to \code{Biostrings::vcountPDict()}
#' (Pages et al., 2020).
#'
#' @md
#'
#' @section Limitations:
#' There are a few limitations with this function, which is important to be
#' aware of:
#'
#' * False negatives or positives may occur due to
#' poorly aligned sequences
#' * The output does not tell which strand
#' (minus or plus) the oligo matches to.
#' This is important to consider when
#' assessing off-target matches to single-stranded
#' targets
#' * Ambiguous bases and gaps in the
#' target sequences are identified as mismatches
#' * The function checks strictly on- and off-target, and may therefore
#' miss off-target matches that partially overlap the intended target
#'
#' @param x
#' An \code{RprimerOligo} or \code{RprimerAssay} object.
#'
#' @param target
#' A \code{MultipleAlignment::DNAMultipleAlignment} alignment with
#' intended target sequences. Note that it must be
#' same alignment that was used for generating the oligos/assays in \code{x}.
#'
#' @return
#' An \code{RprimerMatchOligo} or \code{RprimerMatchAssay} object, depending
#' on whether an \code{RprimerOligo} or \code{RprimerAssay} object was used
#' as input.
#'
#' \code{RprimerMatchOligo} objects contain the following information:
#'
#' \describe{
#'   \item{iupacSequence}{The oligo sequence in IUPAC format.}
#'   \item{perfectMatch}{Proportion of target sequences that matches perfectly
#'   to the oligo within the intended binding region.}
#'   \item{idPerfectMatch}{Names of all sequences that matches perfectly.}
#'   \item{oneMismatch}{Proportion of target sequences with one mismatch to
#'   the oligo
#'   within the intended binding region.}
#'   \item{idOneMismatch}{Names of all sequences that matches with one
#'   mismatch.}
#'   \item{twoMismatches}{Proportion of target sequences with two mismatches
#'   to the oligo
#'   within the intended binding region.}
#'   \item{idTwoMismatches}{Names of all sequences that matches with two
#'   mismatches.}
#'   \item{threeMismatches}{Proportion of target sequences with
#'   three mismatches to the oligo
#'   within the intended binding region.}
#'   \item{idThreeMismatches}{Names of all sequences that matches with three
#'   mismatches.}
#'   \item{fourOrMoreMismatches}{Proportion of target sequences with
#'   four or more mismatches to the oligo
#'   within the intended binding region.}
#'   \item{idFourOrMoreMismatches}{Names of all sequences that matches with four
#'   or more mismatches.}
#'   \item{offTargetMatch}{Proportion of target sequences with maximum two
#'   mismatches to at least one site outside the intended
#'   oligo binding region in the input alignment.
#'   }
#'   \item{idOffTargetMatch}{Names of all off-target matching sequences.}
#'
#' \code{RprimerMatchAssay} objects contain the following information:
#'
#'   \item{iupacSequenceFwd}{The forward primer sequence in IUPAC format.}
#'   \item{perfectMatchFwd}{Proportion of target sequences that matches
#'   perfectly
#'   with the forward primer withing the intended binding region.}
#'   \item{idPerfectMatchFwd}{Names of all sequences that matches perfectly.}
#'   \item{oneMismatchFwd}{Proportion of target sequences with one mismatch to
#'   the forward primer
#'   within the intended binding region.}
#'   \item{idOneMismatchFwd}{Names of all sequences that matches with one
#'   mismatch.}
#'   \item{twoMismatchesFwd}{Proportion of target sequences with two mismatches
#'   to the forward primer
#'   within the intended binding region.}
#'   \item{idTwoMismatchesFwd}{Names of all sequences that matches with two
#'   mismatches.}
#'   \item{threeMismatchesFwd}{Proportion of target sequences with
#'   three mismatches to the forward primer
#'   within the intended binding region.}
#'   \item{idThreeMismatchesFwd}{Names of all sequences that matches with three
#'   mismatches.}
#'   \item{fourOrMoreMismatchesFwd}{Proportion of target sequences with
#'   four or more mismatches to the forward primer
#'   within the intended binding region.}
#'   \item{idFourOrMoreMismatchesFwd}{Names of all sequences that matches with
#'   four or more mismatches.}
#'   \item{offTargetMatchFwd}{Proportion of target sequences with maximum two
#'   mismatches to at least one site outside the intended
#'   forward primer binding region in the input alignment.}
#'   \item{idOffTargetMatchFwd}{Names of all off-target matching sequences.}
#'   \item{iupacSequenceRev}{The reverse primer sequence in IUPAC format.}
#'   \item{perfectMatchRev}{Proportion of target sequences that matches
#'   perfectly
#'   with the reverse primer withing the intended binding region.}
#'   \item{idPerfectMatchRev}{Names of all sequences that matches perfectly.}
#'   \item{oneMismatchRev}{Proportion of target sequences with one mismatch to
#'   the reverse primer
#'   within the intended binding region.}
#'   \item{idOneMismatchRev}{Names of all sequences that matches with one
#'   mismatch.}
#'   \item{twoMismatchesRev}{Proportion of target sequences with two mismatches
#'   to the reverse primer
#'   within the intended binding region.}
#'   \item{idTwoMismatchesRev}{Names of all sequences that matches with two
#'   mismatches.}
#'   \item{threeMismatchesRev}{Proportion of target sequences with
#'   three mismatches to the reverse primer
#'   within the intended binding region.}
#'   \item{idThreeMismatchesRev}{Names of all sequences that matches with three
#'   mismatches.}
#'   \item{fourOrMoreMismatchesRev}{Proportion of target sequences with
#'   four or more mismatches to the reverse primer
#'   within the intended binding region.}
#'   \item{idFourOrMoreMismatchesRev}{Names of all sequences that matches with
#'   four or more mismatches.}
#'   \item{offTargetMatchRev}{Proportion of target sequences with maximum two
#'   mismatches to at least one site outside the intended
#'   reverse primer binding region in the input alignment.}
#'   \item{idOffTargetMatchRev}{Names of all off-target matching sequences.}
#'
#' If the input assay contains probes, the following information
#' is also added:
#'
#'   \item{iupacSequencePr}{The probe sequence in IUPAC format.}
#'   \item{perfectMatchPr}{Proportion of target sequences that matches perfectly
#'   with the probe withing the intended binding region.}
#'   \item{idPerfectMatchPr}{Names of all sequences that matches perfectly.}
#'   \item{oneMismatchPr}{Proportion of target sequences with one mismatch to
#'   the probe
#'   within the intended binding region.}
#'   \item{idOneMismatchPr}{Names of all sequences that matches with one
#'   mismatch.}
#'   \item{twoMismatchesPr}{Proportion of target sequences with two mismatches
#'   to the probe
#'   within the intended binding region.}
#'   \item{idTwoMismatchesPr}{Names of all sequences that matches with two
#'   mismatches.}
#'   \item{threeMismatchesPr}{Proportion of target sequences with
#'   three mismatches to the probe
#'   within the intended binding region.}
#'   \item{idThreeMismatchesPr}{Names of all sequences that matches with three
#'   mismatches.}
#'   \item{fourOrMoreMismatchesPr}{Proportion of target sequences with
#'   four or more mismatches to the probe
#'   within the intended binding region.}
#'   \item{idFourOrMoreMismatchesPr}{Names of all sequences that matches with
#'   four or more mismatches.}
#'    \item{offTargetMatchPr}{Proportion of target sequences with maximum two
#'    mismatches to at least one site outside the intended
#'   probe binding region in the input alignment.}
#'   \item{idOffTargetMatchPr}{Names of all off-target matching sequences.}
#' }
#'
#' @export
#'
#' @references
#' Pages, H., Aboyoun, P., Gentleman R., and DebRoy S. (2020). Biostrings:
#' Efficient manipulation of biological strings. R package version
#' 2.57.2.
#'
#' @examples
#' #### RprimerOligo objects
#'
#' data("exampleRprimerOligo")
#' data("exampleRprimerAlignment")
#'
#' x <- exampleRprimerOligo[1:2, ]
#' target <- exampleRprimerAlignment
#'
#' checkMatch(x, target)
#'
#' #### RprimerAssay objects
#'
#' data("exampleRprimerAssay")
#' data("exampleRprimerAlignment")
#'
#' x <- exampleRprimerAssay[1:2, ]
#' target <- exampleRprimerAlignment
#'
#' checkMatch(x, target)
setGeneric("checkMatch", \(x, target) standardGeneric("checkMatch"))

# Methods ======================================================================

#' @describeIn checkMatch
#'
#' @export
setMethod("checkMatch", "RprimerOligo", \(x, target) {
    if (!methods::is(target, "DNAMultipleAlignment")) {
        stop("'target' must be a DNAMultipleAlignment object.", call. = FALSE)
    }
    match <- .checkMatchOligo(x, target)
    RprimerMatchOligo(match)
})

#' Check match of an RprimerAssay object (method)
#'
#' @describeIn checkMatch
#'
#' @export
setMethod("checkMatch", "RprimerAssay", \(x, target) {
    if (!methods::is(target, "DNAMultipleAlignment")) {
        stop("'target' must be a DNAMultipleAlignment object.", call. = FALSE)
    }
    match <- .checkMatchAssay(x, target)
    RprimerMatchAssay(match)
})

# Helpers ======================================================================

#' @noRd
#'
#' @examples
#' data("exampleRprimerAlignment")
#' .maskRange(from = 1, to = 10, exampleRprimerAlignment, invert = TRUE)
.maskRange <- function(from, to, target, ...) {
    selection <- target
    MultipleAlignment::colmask(selection, ...) <- IRanges::IRanges(
        start = from, end = to
    )
    Biostrings::DNAStringSet(selection)
}

#' @noRd
#'
#' @examples
#' data("exampleRprimerAlignment")
#' target <- Biostrings::DNAStringSet(exampleRprimerAlignment)
#' .getSequenceNames(1:2, target)
.getSequenceNames <- function(x, target) {
    names(target) <- sub(" .*", "", names(target))
    lapply(x, \(i) names(target)[i])
}

#' @noRd
#'
#' @examples
#' data("exampleRprimerAlignment")
#' target <- exampleRprimerAlignment
#' x <- .maskRange(from = 5200, to = 5220, target, invert = TRUE)
.uniqueSequences <- function(x) {
    sequence <- unique(x)
    index <- match(unname(as.vector(x)), as.vector(sequence))
    name <- as.vector(names(x))
    names(name) <- index
    list(sequence = sequence, name = name)
}

#' @noRd
#'
#' @examples
#' data("exampleRprimerAlignment")
#' data("exampleRprimerOligo")
#' x <- exampleRprimerOligo$sequence[[1]]
#' x <- Biostrings::DNAStringSet(x)
#' target <- Biostrings::DNAStringSet(exampleRprimerAlignment)
#' .getMatchIndex(x, target)
.getMatchIndex <- function(x, target, offTarget = FALSE) {
    if (!offTarget) {
        ## 0-3: number of mismatches we are looking for
        res <- lapply(seq(0, 3), \(i) {
            result <- Biostrings::vcountPDict(
                x, target,
                max.mismatch = i
            )
            which(colSums(result) == 0) ## Sequences who "fails"
        })
        res[-length(res)] <- lapply(seq(1, length(res) - 1), \(i) {
            setdiff(res[[i]], res[[i + 1]])
        })
        res[[length(res) + 1]] <- setdiff(seq_along(target), unlist(res))
        res[c(length(res), seq_len(length(res) - 1))]
    } else {
        result <- Biostrings::vcountPDict(
            x, target,
            max.mismatch = 2
        )
        list(which(colSums(result) > 0))
    }
}

#' @noRd
#'
#' @examples
#' data("exampleRprimerAlignment")
#' data("exampleRprimerOligo")
#' target <- Biostrings::DNAStringSet(exampleRprimerAlignment)
#' x <- exampleRprimerOligo$sequence[[1]]
#' x <- Biostrings::DNAStringSet(x)
#' .getMatchProportion(x, target)
.getMatchProportion <- function(x, target, offTarget = FALSE) {
    matching <- .getMatchIndex(x, target, offTarget)
    sequenceNames <- .getSequenceNames(matching, target)
    matching <- lapply(matching, \(x) length(x) / length(target))
    matching <- do.call("cbind.data.frame", matching)
    sequenceNames <- as.data.frame(t(do.call("cbind", list(sequenceNames))))
    matching <- cbind(matching, sequenceNames)
    if (!offTarget) {
        matching <- matching[c(1, 6, 2, 7, 3, 8, 4, 9, 5, 10)]
        names(matching) <- c(
            "perfectMatch", "idPerfectMatch",
            "oneMismatch", "idOneMismatch",
            "twoMismatches", "idTwoMismatches",
            "threeMismatches", "idThreeMismatches",
            "fourOrMoreMismatches", "idFourOrMoreMismatches"
        )
    } else {
        names(matching) <- c("offTargetMatch", "idOffTargetMatch")
    }
    matching
}

#' @noRd
#'
#' @examples
#' data("exampleRprimerAlignment")
#' data("exampleRprimerOligo")
#' x <- exampleRprimerOligo[1:2, ]
#' target <- exampleRprimerAlignment
#' .checkMatchOligo(x, target)
.checkMatchOligo <- function(x, target) {
    onTarget <- lapply(seq_len(nrow(x)), \(i) {
        target <- .maskRange(
            x$start[[i]], x$end[[i]], target,
            invert = TRUE
        )
        check <- Biostrings::DNAStringSet(x$sequence[[i]])
        .getMatchProportion(check, target)
    })
    offTarget <- lapply(seq_len(nrow(x)), \(i) {
        target <- .maskRange(
            x$start[[i]], x$end[[i]], target,
            invert = FALSE
        )
        check <- Biostrings::DNAStringSet(x$sequence[[i]])
        .getMatchProportion(check, target, offTarget = TRUE)
    })
    onTarget <- do.call("rbind", onTarget)
    offTarget <- do.call("rbind", offTarget)
    cbind(
        "iupacSequence" = x$iupacSequence,
        onTarget, offTarget
    )
}

#' @noRd
#'
#' @examples
#' data("exampleRprimerAssay")
#' x <- eampleRprimerAssay[1:4, ]
#' .convertAssay(x)
.convertAssay <- function(x) {
    x <- as.data.frame(x)
    fwd <- x[, grepl("Fwd", names(x))]
    names(fwd) <- gsub("Fwd", "", names(fwd))
    rev <- x[, grepl("Rev", names(x))]
    names(rev) <- gsub("Rev", "", names(rev))
    rev$sequence <- lapply(rev$sequence, \(x) {
        x <- Biostrings::DNAStringSet(x)
        x <- Biostrings::reverseComplement(x)
        as.character(x)
    })
    pr <- x[, grepl("Pr", names(x))]
    names(pr) <- gsub("Pr", "", names(pr))
    list("fwd" = fwd, "rev" = rev, "pr" = pr)
}

#' @noRd
#'
#' @examples
#' data("exampleRprimerAlignment")
#' data("exampleRprimerAssay")
#' x <- exampleRprimerAssay[1:2, ]
#' target <- exampleRprimerAlignment
#' .checkMatchAssay(x, target)
.checkMatchAssay <- function(x, target) {
    oligos <- .convertAssay(x)
    fwd <- .checkMatchOligo(oligos$fwd, target)
    names(fwd) <- paste0(names(fwd), "Fwd")
    rev <- .checkMatchOligo(oligos$rev, target)
    names(rev) <- paste0(names(rev), "Rev")
    all <- cbind(fwd, rev)
    if (any(grepl("Pr$", names(x)))) {
        pr <- .checkMatchOligo(oligos$pr, target)
        names(pr) <- paste0(names(pr), "Pr")
        all <- cbind(all, pr)
    }
    all
}

## Load example data
data("exampleRprimerAlignment")
data("exampleRprimerOligo")
data("exampleRprimerAssay")
data("exampleRprimerMatchOligo")
data("exampleRprimerMatchAssay")

MultipleAlignment::rowmask(exampleRprimerAlignment, invert = TRUE) <- 1

# checkMatch ===================================================================

test_that("checkMatch returns an error when it should", {
    expect_error(
        checkMatch(exampleRprimerOligo[1:2, ], unclass(exampleRprimerAlignment))
    )
    expect_error(
        checkMatch(exampleRprimerAssay[1:2, ], unclass(exampleRprimerAlignment))
    )
})

test_that("checkMatch works", {
    oligoMatch <- checkMatch(exampleRprimerOligo[1, ], exampleRprimerAlignment)
    assayMatch <- checkMatch(exampleRprimerAssay[1, ], exampleRprimerAlignment)
    expect_s4_class(oligoMatch, "RprimerMatchOligo")
    expect_s4_class(assayMatch, "RprimerMatchAssay")
})

# .maskRange ===================================================================

# .getMatchIndex ===============================================================

test_that(".getMatchIndex works", {
    target <- c(
        "PM" = "AAAAAA",
        "MM1" = "AAAAAT",
        "MM2" = "AAAATT",
        "MM3" = "AAATTT",
        "MM4" = "AATTTT"
    )
    target <- Biostrings::DNAStringSet(target)
    x <- Biostrings::DNAStringSet("AAAAAA") ## Oligo
    result <- .getMatchIndex(x, target)
    expectedResult <- list(1, 2, 3, 4, 5)
    expect_equal(result, expectedResult)
    xDegenerate <- Biostrings::DNAStringSet(c("AAAAAA", "AAAAAT"))
    result <- .getMatchIndex(xDegenerate, target)
    expectedResult <- list(c(1, 2), 3, 4, 5, integer(0L))
    expect_equal(result, expectedResult)
})

# .getSequenceNames ============================================================

test_that(".getSequenceNames works", {
    target <- list("A BLABLA" = "CCC", "B BLABLA" = "TTT")
    result <- .getSequenceNames(1, target)
    expect_equal(result, list("A"))
})

# .getMatchProportion ==========================================================

test_that(".getMatchProportion works", {
    x <- Biostrings::DNAStringSet("AAAAA")
    target <- c(
        "AAAAA",
        "AAAAT",
        "AAATT",
        "AATTT",
        "ATTTT",
        "TTTTT"
    )
    target <- Biostrings::DNAStringSet(target)
    result <- .getMatchProportion(x, target)
    expect_true(is.data.frame(result))
})

# .checkMatchOligo =============================================================

test_that(".checkMatchOligo works", {
    x <- as.data.frame(exampleRprimerOligo[1, ])

    ## Check that target and off target identification works
    onTarget <- exampleRprimerAlignment
    MultipleAlignment::colmask(onTarget, invert = TRUE) <- x$start:x$end
    on <- .checkMatchOligo(x, onTarget)
    offTarget <- exampleRprimerAlignment
    MultipleAlignment::colmask(offTarget) <- x$start:x$end
    off <- .checkMatchOligo(x, offTarget)
    both <- .checkMatchOligo(x, exampleRprimerAlignment)
    expect_equal(on[!grepl("Off", names(on))], both[!grepl("Off", names(both))])
    expect_equal(off[grepl("Off", names(off))], both[grepl("Off", names(both))])
})

# .convertAssay and .checkMatchAssay ===========================================

test_that(".convertAssay and checkMatchAssay work", {
    x <- as.data.frame(exampleRprimerAssay[1:2, ])
    x <- x[!grepl("Pr$", names(x))]
    result <- .convertAssay(x)
    expect_true(is.list(result))
    result <- .checkMatchAssay(x, exampleRprimerAlignment)
    expect_true(is.data.frame(result))
})

## Helper functions to shiny app ===============================================

## UI functions ================================================================

switchPage <- function(i) {
    shiny::updateTabsetPanel(inputId = "wizard", selected = paste0("page", i))
}

previousButton <- function(i) {
    id <- paste0("page", i + 1, i)
    shiny::actionButton(
        id, "Previous",
        icon = shiny::icon("chevron-left")
    )
}

nextButton <- function(i) {
    id <- paste0("page", i - 1, i)
    shiny::actionButton(
        id,
        label = shiny::div(
            "Next", shiny::icon("chevron-right"),
            class = "btn btn-primary"
        )
    )
}

spinnerPlot <- function(id, ...) {
    shinycssloaders::withSpinner(
        shiny::plotOutput(id, ...),
        color = "grey"
    )
}

numericInputFrom <- function(x, id) {
    shiny::numericInput(id, "From",
        min = x$roiStart[[1]],
        max = x$roiEnd[[1]],
        value = x$roiStart[[1]]
    )
}

numericInputTo <- function(x, id) {
    shiny::numericInput(id, "To",
        min = x$roiStart[[1]],
        max = x$roiEnd[[1]],
        value = x$roiEnd[[1]]
    )
}

conservationInput <- function(x,
                              id,
                              type = "primer",
                              direction = "both",
                              variable = "identity") {
    x <- x[x$type == type, ]
    if (direction == "fwd") {
        x <- x[x$fwd, ]
    } else if (direction == "rev") {
        x <- x[x$rev, ]
    } else {
        x <- x[x$rev & x$fwd, ]
    }
    if (variable == "identity") {
        var <- x$identity
    } else if (variable == "coverage") {
        var <- x$coverage
    }
    minValue <- round(min(var, na.rm = TRUE), 4)
    maxValue <- round(max(var, na.rm = TRUE), 4)
    shiny::sliderInput(id,
        round = -4, step = 0.0001,
        paste("Minimum", variable),
        min = minValue,
        max = maxValue,
        value = minValue
    )
}

## Other functions =============================================================

arrangeMask <- function(from, to, aln) {
    alnLength <- ncol(aln)
    if (from == 0 && to == 0) {
        NULL
    } else {
        if (from < 1) from <- 1
        if (to < 1) to <- 1
        if (to > alnLength) to <- alnLength
        sort(from:to)
    }
}

filterOligos <- function(x,
                         fwdFrom = 0,
                         fwdTo = Inf,
                         revFrom = 0,
                         revTo = Inf,
                         prFrom = 0,
                         prTo = Inf,
                         fwdIdentity = 0,
                         fwdCoverage = 0,
                         revIdentity = 0,
                         revCoverage = 0,
                         prIdentity = 0,
                         prCoverage = 0) {
    x <- as.data.frame(x)
    emptyRow <- makeEmptyRow(x)
    if (any(x$type == "primer")) {
        primers <- x[x$type == "primer", ]
        if (any(primers$fwd)) {
            fwd <- primers$fwd &
                primers$start >= fwdFrom &
                primers$end <= fwdTo &
                primers$identity >= fwdIdentity &
                primers$coverage >= fwdCoverage
            primers$fwd <- fwd
        }
        if (any(primers$rev)) {
            rev <- primers$rev &
                primers$start >= revFrom &
                primers$end <= revTo &
                primers$identity >= revIdentity &
                primers$coverage >= revCoverage
            primers$rev <- rev
        }
    } else {
        primers <- NULL
    }
    if (any(x$type == "probe")) {
        probes <- x[x$type == "probe", ]
        pr <- probes$start >= prFrom &
            probes$end <= prTo &
            probes$identity >= prIdentity &
            probes$coverage >= prCoverage
        probes <- probes[pr, ]
    } else {
        probes <- NULL
    }
    all <- rbind(primers, probes)
    all <- all[all$fwd | all$rev, ]
    all <- all[order(all$start), ]
    if (nrow(all) == 0L) {
        all <- emptyRow
    }
    RprimerOligo(all)
}

roundDbls <- function(x) {
    x <- as.data.frame(x)
    y <- lapply(seq_len(ncol(x)), \(i) {
        out <- if (is.double(x[, i])) {
            round(x[, i], 4)
        } else if (is.list(x[, i])) {
            if (is.double(x[, i][[1]])) {
                list(round(x[, i][[1]], 4))
            } else {
                list(x[, i][[1]])
            }
        } else {
            x[, i]
        }
    })
    names(y) <- names(x)
    data.frame(do.call("cbind", y))
}

removeListColumns <- function(x) {
    lists <- vapply(seq_len(ncol(x)), \(i) {
        is.list(x[, i])
    }, logical(1L))
    x[, !lists]
}

splitAssayToList <- function(x) {
    x <- as.data.frame(x)
    fwd <- x[, grepl("Fwd", names(x))]
    fwd <- cbind(
        fwd,
        "roiStart" = x$roiStart, "roiEnd" = x$roiEnd
    )
    names(fwd) <- gsub("Fwd", "", names(fwd))
    rev <- x[, grepl("Rev", names(x))]
    rev <- cbind(
        rev,
        "roiStart" = x$roiStart, "roiEnd" = x$roiEnd
    )
    names(rev) <- gsub("Rev", "", names(rev))
    if (any(grepl("Pr", names(x)))) {
        pr <- x[, grepl("Pr", names(x))]
        pr <- cbind(
            pr,
            "roiStart" = x$roiStart, "roiEnd" = x$roiEnd
        )
        names(pr) <- gsub("Pr", "", names(pr))
        all <- list(fwd, rev, pr)
    } else {
        all <- list(fwd, rev)
    }
    all
}

convertToOligo <- function(x, rev = FALSE, type = "primer") {
    add <- data.frame(
        "type" = type, "fwd" = !rev, "rev" = rev, "score" = NA,
        "roiStart" = NA, "roiEnd" = NA, "iupacSequenceRc" = NA,
        "sequenceRc" = NA
    )
    x <- cbind(x, add)
    x$iupacSequenceRc <- makeRc(x$iupacSequence)
    x$sequenceRc <- makeRcList(x$sequence)
    if (rev) {
        old <- c("iupacSequence", "sequence", "iupacSequenceRc", "sequenceRc")
        new <- c("iupacSequenceRc", "sequenceRc", "iupacSequence", "sequence")
        names(x)[names(x) %in% old] <- new
    }
    x <- .beautifyOligos(x)
    RprimerOligo(x)
}

makeRc <- function(x) {
    x <- Biostrings::DNAString(x)
    x <- Biostrings::reverseComplement(x)
    as.character(x)
}

makeRcList <- function(x) {
    x <- unlist(x)
    x <- lapply(x, makeRc)
    x <- unlist(x)
    list(x)
}

makeListTable <- function(x) {
    x <- as.data.frame(x)
    list <- vapply(seq_len(ncol(x)), \(i) {
        is.list(x[, i])
    }, logical(1L))
    x <- x[, list]
    allCols <- lapply(seq_len(ncol(x)), \(i) {
        x[, i][[1]]
    })
    all <- data.frame(do.call("cbind", allCols))
    names(all) <- names(x)
    all$gcContent <- as.double(all$gcContent)
    all$tm <- as.double(all$tm)
    all$deltaG <- as.double(all$deltaG)
    all
}

makeEmptyRow <- function(x) {
    emptyRow <- x[1, , drop = FALSE]
    emptyRow[1, ] <- NA
    emptyRow$roiStart <- x$roiStart[[1]]
    emptyRow$roiEnd <- x$roiEnd[[1]]
    emptyRow$start <- 1
    emptyRow$end <- 1
    emptyRow$fwd <- emptyRow$rev <- TRUE
    emptyRow$type <- "primer"
    emptyRow$length <- NA
    emptyRow
}

#' rprimer Shiny application
#'
#' \code{runRprimerApp()} starts a Shiny application where
#' the workflow of the rprimer package can be run through a
#' graphical user interface.
#'
#' @return Opens the Shiny application.
#'
#' @export
#'
#' @import shiny
#'
#' @importFrom MultipleAlignment readDNAMultipleAlignment
#'
#' @importFrom MultipleAlignment colmask
#'
#' @importFrom MultipleAlignment rowmask
#'
#' @examples
#' ## Only run this in interactive R sessions:
#' if (interactive()) {
#'     runRprimerApp()
#' }
runRprimerApp <- function() {
    rprimerApp()
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/plotData.R
\name{plotData}
\alias{plotData}
\alias{plotData,RprimerProfile-method}
\alias{plotData,RprimerOligo-method}
\alias{plotData,RprimerAssay-method}
\alias{plotData,RprimerMatchOligo-method}
\alias{plotData,RprimerMatchAssay-method}
\title{Plot an Rprimer object}
\usage{
plotData(x, ...)

\S4method{plotData}{RprimerProfile}(x, type = "overview", highlight = NULL, rc = FALSE)

\S4method{plotData}{RprimerOligo}(x)

\S4method{plotData}{RprimerAssay}(x)

\S4method{plotData}{RprimerMatchOligo}(x)

\S4method{plotData}{RprimerMatchAssay}(x)
}
\arguments{
\item{x}{An \code{RprimerProfile}, \code{RprimerOligo} \code{RprimerAssay},
\code{rprimerMatchOligo} or \code{RprimerMatchAssay} object.}

\item{...}{Optional arguments for \code{RprimerProfile} objects.}

\item{type}{For \code{Rprimeroligo} objects:
Type of plot: \code{"overview"}, or
\code{"nucleotide"}, defaults to \code{"overview"}.}

\item{highlight}{For \code{Rprimeroligo} objects:
If a specific region within an overview plot should be highlighted.
A numeric vector indicating the start and end position,
e.g. \code{c(100, 1000)}, defaults to \code{NULL}
(i.e., no highlight).}

\item{rc}{For \code{Rprimeroligo} objects, and \code{type = "nucleotide"}:
If the plotted sequence should be displayed
as reverse complement or not.
\code{TRUE} or {FALSE}, defaults to \code{FALSE}.

See examples below.}
}
\value{
A plot.
}
\description{
\code{plotData} visualizes objects from all different
Rprimer classes.
}
\section{Methods (by class)}{
\itemize{
\item \code{RprimerProfile}: 

\item \code{RprimerOligo}: 

\item \code{RprimerAssay}: 

\item \code{RprimerMatchOligo}: 

\item \code{RprimerMatchAssay}: 
}}

\examples{
#### Plot an RprimerProfile object

data("exampleRprimerProfile")
prof <- exampleRprimerProfile

## Plot an overview
plotData(prof, highlight = c(500, 1000))

## Select a region of interest
roi <- prof[prof$position >= 500 & prof$position <= 550, ]

## Plot an overview of the roi
plotData(roi)

## Plot the nucleotide distribution of the roi
plotData(roi, type = "nucleotide")

#### Plot an RprimerOligo object

data("exampleRprimerOligo")
plotData(exampleRprimerOligo)

#### Plot an RprimerAssay object

data("exampleRprimerAssay")
plotData(exampleRprimerAssay)

#### Plot an RprimerMatchOligo object

data("exampleRprimerMatchOligo")
plotData(exampleRprimerMatchOligo)

#### Plot an RprimerMatchAssay object

data("exampleRprimerMatchAssay")
plotData(exampleRprimerMatchAssay)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/designOligos.R
\name{designOligos}
\alias{designOligos}
\title{Design primers and probes}
\usage{
designOligos(
  x,
  maxGapFrequency = 0.01,
  lengthPrimer = c(18, 22),
  maxDegeneracyPrimer = 4,
  gcClampPrimer = TRUE,
  avoidThreeEndRunsPrimer = TRUE,
  gcPrimer = c(0.4, 0.65),
  tmPrimer = c(50, 65),
  concPrimer = 500,
  designStrategyPrimer = "ambiguous",
  probe = TRUE,
  lengthProbe = c(18, 22),
  maxDegeneracyProbe = 4,
  avoidFiveEndGProbe = TRUE,
  gcProbe = c(0.4, 0.65),
  tmProbe = c(50, 70),
  concProbe = 250,
  concNa = 0.05
)
}
\arguments{
\item{x}{An \code{RprimerProfile} object.}

\item{maxGapFrequency}{Maximum allowed gap frequency at the primer and probe binding sites in
the target alignment.
A number [0, 1], defaults to \code{0.01}.}

\item{lengthPrimer}{Primer length range. A numeric vector [15, 30],
defaults to \code{c(18, 22)}.}

\item{maxDegeneracyPrimer}{Maximum number of variants of each primer. A number [1, 64],
defaults to \code{4}.}

\item{gcClampPrimer}{If primers must have a GC clamp.
A GC clamp
is identified as two to three G or
C:s within the last five bases (3' end) of the primer.
\code{TRUE} or \code{FALSE}, defaults to \code{TRUE}.}

\item{avoidThreeEndRunsPrimer}{If primers with more than two runs
of the same nucleotide at the terminal 3' end should be avoided.
\code{TRUE} or \code{FALSE}, defaults to \code{TRUE}.}

\item{gcPrimer}{GC-content range for primers.
A numeric vector [0, 1], defaults to \code{c(0.40, 0.65)}.}

\item{tmPrimer}{Tm range for primers (in Celcius degrees).
A numeric vector [30, 90], defaults to \code{c(55, 65)}.}

\item{concPrimer}{Primer concentration in nM, for tm calculation. A number
[20, 2000], defaults to \code{500}.}

\item{designStrategyPrimer}{\code{"ambiguous"} or \code{"mixed"}. Defaults to \code{"ambiguous"}
(see details below).}

\item{probe}{If probes should be designed. \code{TRUE} or \code{FALSE},
defaults to \code{TRUE}.}

\item{lengthProbe}{Probe length range. A numeric vector [15, 40],
defaults to \code{c(18, 22)}.}

\item{maxDegeneracyProbe}{Maximum number of variants of each probe. A number [1, 64],
defaults to \code{4}.}

\item{avoidFiveEndGProbe}{If probes with G
at the 5' end should be avoided. \code{TRUE} or \code{FALSE},
defaults to \code{TRUE}.}

\item{gcProbe}{GC-content range for probes. A numeric vector [0, 1],
defaults to \code{c(0.40, 0.65)}.}

\item{tmProbe}{Tm range for probes (in Celcius degrees).
A numeric vector [30, 90], defaults to \code{c(55, 70)}.}

\item{concProbe}{Primer concentration in nM, for tm calculation. A numeric vector
[20, 2000], defaults to \code{250}.}

\item{concNa}{Sodium ion (equivalent) concentration in the PCR reaction (in M).
For calculation of tm and delta G.
A numeric vector [0.01, 1], defaults to \code{0.05} (50 mM).}
}
\value{
An \code{RprimerOligo} object, containing the following information:

\describe{
  \item{type}{Whether the oligo is a primer or probe.}
  \item{fwd}{\code{TRUE} if the oligo is valid in forward direction,
    \code{FALSE} otherwise.}
  \item{rev}{\code{TRUE} if the oligo is valid in reverse direction,
    \code{FALSE} otherwise.}
  \item{start}{Start position of the oligo.}
  \item{end}{End positon of the oligo.}
  \item{length}{Oligo length.}
  \item{iupacSequence}{Oligo sequence in IUPAC format
  (i.e. with ambiguous bases).}
  \item{iupaSequenceRc}{The reverse complement of the iupacSequence.}
  \item{identity}{For ambiguous oligos: average identity of the oligo.
    For mixed oligos: average identity of the 5' (consensus) part of the
    oligo. The value can range from 0 to 1.}
  \item{coverage}{For ambiguous oligos: average coverage of the oligo.
    For mixed oligos: average coverage of the 3' (degenerate) part of the
    oligo. The value can range from 0 to 1.}
  \item{degeneracy}{Number of sequence variants of the oligo.}
  \item{gcContentMean}{Mean GC-content of all sequence variants of the oligo.
  }
  \item{gcContent}{Range in GC-content of all sequence variants of
    the oligo.}
  \item{tmMean}{Mean tm of all sequence variants of the oligo
  (in Celcius degrees).}
  \item{tm}{Range in tm of all sequence variants of the oligo
  (in Celcius degrees).}
  \item{deltaGMean}{Mean delta G of all sequence variants of the oligo
  (in kcal/mol).}
  \item{deltaG}{Range in delta G of all sequence variants of the oligo
  (in kcal/mol).}
  \item{sequence}{All sequence variants of the oligo.}
  \item{sequenceRc}{Reverse complements of all sequence variants.}
  \item{gcContent}{GC-content of all sequence variants.}
  \item{tm}{Tm of all sequence variants (in Celcius degrees).}
  \item{deltaG}{Delta G of all sequence variants (in kcal/mol).}
  \item{method}{Design method used to generate the oligo: "ambiguous",
  "mixedFwd" or "mixedRev".}
  \item{score}{Oligo score, the lower the better.}
  \item{roiStart}{First position of the input \code{RprimerProfile} object
    (roi = region of interest).}
  \item{roiEnd}{Last position of the input \code{RprimerProfile} object.}
}

An error message will return if no oligos are found. If so, a good idea
could be to re-run the design process with relaxed constraints.
}
\description{
\code{designOligos()} designs oligos (primers and probes)
from a consensus profile.
}
\details{
\strong{Valid oligos}

For an oligo to be considered as valid, all sequence variants must fulfill
all the specified design constraints.

Furthermore, oligos with at least one sequence variant containing
more than four consecutive runs
of the same
nucleotide (e.g. "AAAAA") and/or more than three consecutive runs
of the same di-nucleotide (e.g. "TATATATA") will be excluded
from consideration.

\strong{Calculation of tm and delta G}

Melting temperatures
are calculated for perfectly matching
DNA duplexes using the
nearest-neighbor
method (SantaLucia and Hicks, 2004), by using the following equation:

\loadmathjax
\mjsdeqn{Tm = (\Delta H ^o \cdot 1000) / (\Delta S ^o + R \cdot \log [\mathrm{oligo}]) - 273.15}

where \mjseqn{\Delta H ^o} is
the change in enthalpy (in cal/mol) and \mjseqn{\Delta S ^o} is the
change in entropy (in cal/K/mol) when an
oligo and a perfectly matching target sequence goes from random coil to
duplex formation.
\mjseqn{K} is the gas constant (1.9872 cal/mol \emph{K}).

Delta G is calculated at 37 Celcius degrees, for when an oligo and a
perfectly matching target
sequence goes from random coil to duplex state, by using  the following
equation:

\mjsdeqn{ \Delta G ^o _T = ( \Delta H ^o \cdot 1000 - T \cdot \Delta S ^o ) / 1000}{ASCII representation}
For both tm and delta G, the following salt correction method is used
for \mjseqn{ \Delta S^o }, as
described in SantaLucia and Hicks (2004):

\mjsdeqn{ \Delta S^o [\mathrm{Na^+}] = \Delta S^o [\mathrm{1 M NaCl}] + 0.368 \cdot N / 2 \cdot \log [\mathrm{Na^+}] }

where \mjseqn{N} is the total number of phosphates in the duplex, and [Na+] is
the total
concentration of monovalent cations.

Nearest neighbor table values for \mjseqn{\Delta S^o} and \mjseqn{\Delta H^o} are
from SantaLucia and Hicks, 2004, and can be
retrieved calling \code{rprimer:::lookup$nn}.

\strong{Primer design strategies}

Primers can be generated by using one of the two following strategies:

\itemize{
\item{The \strong{ambiguous strategy} (default) generates primers from the
IUPAC consensus sequence alone, which means that ambiguous bases can
occur at any position in the primer.}

\item{The \strong{mixed strategy} generates primers from both the majority
and the IUPAC consensus sequence. These primers consist of a shorter
degenerate part at the 3' end (approx. 1/3 of the primer, targeting a
conserved
region) and a longer consensus part at the 5' end (approx.
2/3 of the primer),
which instead of having ambiguous bases contains the most frequently occuring
nucleotide at each position.
This strategy resembles the widely-adopted Consensus-Degenerate Hybrid
Oligonucleotide Primer (CODEHOP) principle (Rose et al., 1998), and aims to
to allow amplification of highly variable targets using primers with
low degeneracy.
The idea is that the degenerate 3' end part will bind specifically to
the target sequence in the initial PCR cycles, and promote amplification
in spite of eventual mismatches at the 5' consensus part
(since 5' end mismatches are generally less detrimental than
3' end mismatches). In this way, the generated products
will match the 5' ends of all primers perfectly, which allows them
to be efficiently amplified in later PCR cycles.
To provide a sufficiently high tm in spite of mismatches, it is
recommended to design relatively long primers (at least 25 bases) when using
this strategy.}
}

Probes are always designed using the ambiguous strategy.

\strong{Scoring system for oligos}

All valid oligos are scored based on their identity, coverage and
average GC content. The scoring system is presented below.

\strong{Identity and coverage}

\tabular{lr}{
Value range \tab Score \cr
\eqn{(0.99, 1]} \tab 0 \cr
\eqn{(0.95, 0.99]} \tab 1 \cr
\eqn{(0.90, 0.95]} \tab 2 \cr
\eqn{\leq 0.90} \tab 3
}

\strong{Average GC-content}

This score is based on how much
the average GC-content deviates from 0.5 (in absolute value).

\tabular{lr}{
Value range \tab Score \cr
\eqn{[0, 0.05)}
\tab 0 \cr
\eqn{[0.05, 0.1)}
\tab 1 \cr
\eqn{[0.1,  0.2)}
\tab 2 \cr
\eqn{\geq 0.2} \tab 3
}

These scores are summarized. The weight of each individual score is 1, and
thus, the lowest and best
possible score for an oligo is 0, and the worst possible score is 9.
}
\examples{
data("exampleRprimerProfile")
x <- exampleRprimerProfile

## Design primers and probes with default values
designOligos(x)
}
\references{
Rose, TM., Schultz ER., Henikoff JG., Pietrokovski S.,
McCallum CM., and Henikoff S. 1998. Consensus-Degenerate
Hybrid Oligonucleotide Primers for
Amplification of Distantly Related Sequences. Nucleic Acids Research 26 (7):
1628-35.

SantaLucia Jr, J., & Hicks, D. (2004).
The thermodynamics of DNA structural motifs.
Annu. Rev. Biophys. Biomol. Struct., 33, 415-440.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/datasets.R
\docType{data}
\name{example-datasets}
\alias{example-datasets}
\alias{exampleRprimerAlignment}
\alias{exampleRprimerProfile}
\alias{exampleRprimerOligo}
\alias{exampleRprimerAssay}
\alias{exampleRprimerMatchOligo}
\alias{exampleRprimerMatchAssay}
\title{Example datasets}
\format{
An object of class \code{DNAMultipleAlignment} with 50 rows and 7597 columns.

An object of class \code{RprimerProfile} with 7597 rows and 12 columns.

An object of class \code{RprimerOligo} with 322 rows and 26 columns.

An object of class \code{RprimerAssay} with 4883 rows and 65 columns.

An object of class \code{RprimerMatchOligo} with 10 rows and 13 columns.

An object of class \code{RprimerMatchAssay} with 5 rows and 39 columns.
}
\usage{
data("exampleRprimerAlignment")

data("exampleRprimerProfile")

data("exampleRprimerOligo")

data("exampleRprimerAssay")

data("exampleRprimerMatchOligo")

data("exampleRprimerMatchAssay")
}
\description{
The purpose of these datasets is to illustrate the functionality of
rprimer. The following datasets are provided:

\itemize{
\item \code{exampleRprimerAlignment} - a \code{MultipleAlignment::DNAMultipleAlignment}
object (Carlson & Aboyoun, 2020)
containing an alignment of 50 hepatitis E virus
sequences collected from NCBI GenBank. See
"documentation_example_alignment.txt" within the inst/script folder of this
package for more details.
\item \code{exampleRprimerProfile} - an \code{RprimerProfile} object, generated
from the alignment above.
\item \code{exampleRprimerOligo} - an \code{RprimerOligo} object, generated from
the consensus profile above.
\item \code{exampleRprimerAssay} - an \code{RprimerAssay} object, generated from
the oligos above.
\item \code{exampleRprimerMatchOligo} - an \code{RprimerMatchOligo} object,
describing how well some oligos match with the sequences in
exampleRprimerAlignment.
\item \code{exampleRprimerMatchAssay} - an \code{RprimerMatchAssay} object,
describing how well some assays match with the seuqences in
exampleRprimerAlignment.
}
}
\references{
M. Carlson and P. Aboyoun (2026). MultipleAlignment:
Representation of multiple sequence alignments in Bioconductor.
R package version 0.99.5.
}
\keyword{datasets}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/rprimer-package.R
\docType{package}
\name{rprimer-package}
\alias{rprimer}
\alias{rprimer-package}
\title{rprimer: Design Degenerate Oligos from a Multiple DNA Sequence Alignment}
\description{
Functions, workflow, and a Shiny application for visualizing sequence conservation and designing degenerate primers, probes, and (RT)-(q/d)PCR assays from a multiple DNA sequence alignment. The results can be presented in data frame format and visualized as dashboard-like plots. For more information, please see the package vignette.
}
\details{
The package contains five functions: \code{consensusProfile()},
\code{designOligos()}, \code{designAssays()}, \code{checkMatch()} 
and \code{plotData()}
}
\seealso{
Useful links:
\itemize{
  \item \url{https://github.com/sofpn/rprimer}
  \item Report bugs at \url{https://github.com/sofpn/rprimer/issues}
}

}
\author{
\strong{Maintainer}: Sofia Persson \email{sofiapersson27@gmail.com} (\href{https://orcid.org/0000-0003-2611-3030}{ORCID})

}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/checkMatch.R
\name{checkMatch}
\alias{checkMatch}
\alias{checkMatch,RprimerOligo-method}
\alias{checkMatch,RprimerAssay-method}
\title{Check how oligos and assays match to their target sequences}
\usage{
checkMatch(x, target)

\S4method{checkMatch}{RprimerOligo}(x, target)

\S4method{checkMatch}{RprimerAssay}(x, target)
}
\arguments{
\item{x}{An \code{RprimerOligo} or \code{RprimerAssay} object.}

\item{target}{A \code{MultipleAlignment::DNAMultipleAlignment} alignment with
intended target sequences. Note that it must be
same alignment that was used for generating the oligos/assays in \code{x}.}
}
\value{
An \code{RprimerMatchOligo} or \code{RprimerMatchAssay} object, depending
on whether an \code{RprimerOligo} or \code{RprimerAssay} object was used
as input.

\code{RprimerMatchOligo} objects contain the following information:

\describe{
\item{iupacSequence}{The oligo sequence in IUPAC format.}
\item{perfectMatch}{Proportion of target sequences that matches perfectly
to the oligo within the intended binding region.}
\item{idPerfectMatch}{Names of all sequences that matches perfectly.}
\item{oneMismatch}{Proportion of target sequences with one mismatch to
the oligo
within the intended binding region.}
\item{idOneMismatch}{Names of all sequences that matches with one
mismatch.}
\item{twoMismatches}{Proportion of target sequences with two mismatches
to the oligo
within the intended binding region.}
\item{idTwoMismatches}{Names of all sequences that matches with two
mismatches.}
\item{threeMismatches}{Proportion of target sequences with
three mismatches to the oligo
within the intended binding region.}
\item{idThreeMismatches}{Names of all sequences that matches with three
mismatches.}
\item{fourOrMoreMismatches}{Proportion of target sequences with
four or more mismatches to the oligo
within the intended binding region.}
\item{idFourOrMoreMismatches}{Names of all sequences that matches with four
or more mismatches.}
\item{offTargetMatch}{Proportion of target sequences with maximum two
mismatches to at least one site outside the intended
oligo binding region in the input alignment.
}
\item{idOffTargetMatch}{Names of all off-target matching sequences.}

\code{RprimerMatchAssay} objects contain the following information:

\item{iupacSequenceFwd}{The forward primer sequence in IUPAC format.}
\item{perfectMatchFwd}{Proportion of target sequences that matches
perfectly
with the forward primer withing the intended binding region.}
\item{idPerfectMatchFwd}{Names of all sequences that matches perfectly.}
\item{oneMismatchFwd}{Proportion of target sequences with one mismatch to
the forward primer
within the intended binding region.}
\item{idOneMismatchFwd}{Names of all sequences that matches with one
mismatch.}
\item{twoMismatchesFwd}{Proportion of target sequences with two mismatches
to the forward primer
within the intended binding region.}
\item{idTwoMismatchesFwd}{Names of all sequences that matches with two
mismatches.}
\item{threeMismatchesFwd}{Proportion of target sequences with
three mismatches to the forward primer
within the intended binding region.}
\item{idThreeMismatchesFwd}{Names of all sequences that matches with three
mismatches.}
\item{fourOrMoreMismatchesFwd}{Proportion of target sequences with
four or more mismatches to the forward primer
within the intended binding region.}
\item{idFourOrMoreMismatchesFwd}{Names of all sequences that matches with
four or more mismatches.}
\item{offTargetMatchFwd}{Proportion of target sequences with maximum two
mismatches to at least one site outside the intended
forward primer binding region in the input alignment.}
\item{idOffTargetMatchFwd}{Names of all off-target matching sequences.}
\item{iupacSequenceRev}{The reverse primer sequence in IUPAC format.}
\item{perfectMatchRev}{Proportion of target sequences that matches
perfectly
with the reverse primer withing the intended binding region.}
\item{idPerfectMatchRev}{Names of all sequences that matches perfectly.}
\item{oneMismatchRev}{Proportion of target sequences with one mismatch to
the reverse primer
within the intended binding region.}
\item{idOneMismatchRev}{Names of all sequences that matches with one
mismatch.}
\item{twoMismatchesRev}{Proportion of target sequences with two mismatches
to the reverse primer
within the intended binding region.}
\item{idTwoMismatchesRev}{Names of all sequences that matches with two
mismatches.}
\item{threeMismatchesRev}{Proportion of target sequences with
three mismatches to the reverse primer
within the intended binding region.}
\item{idThreeMismatchesRev}{Names of all sequences that matches with three
mismatches.}
\item{fourOrMoreMismatchesRev}{Proportion of target sequences with
four or more mismatches to the reverse primer
within the intended binding region.}
\item{idFourOrMoreMismatchesRev}{Names of all sequences that matches with
four or more mismatches.}
\item{offTargetMatchRev}{Proportion of target sequences with maximum two
mismatches to at least one site outside the intended
reverse primer binding region in the input alignment.}
\item{idOffTargetMatchRev}{Names of all off-target matching sequences.}

If the input assay contains probes, the following information
is also added:

\item{iupacSequencePr}{The probe sequence in IUPAC format.}
\item{perfectMatchPr}{Proportion of target sequences that matches perfectly
with the probe withing the intended binding region.}
\item{idPerfectMatchPr}{Names of all sequences that matches perfectly.}
\item{oneMismatchPr}{Proportion of target sequences with one mismatch to
the probe
within the intended binding region.}
\item{idOneMismatchPr}{Names of all sequences that matches with one
mismatch.}
\item{twoMismatchesPr}{Proportion of target sequences with two mismatches
to the probe
within the intended binding region.}
\item{idTwoMismatchesPr}{Names of all sequences that matches with two
mismatches.}
\item{threeMismatchesPr}{Proportion of target sequences with
three mismatches to the probe
within the intended binding region.}
\item{idThreeMismatchesPr}{Names of all sequences that matches with three
mismatches.}
\item{fourOrMoreMismatchesPr}{Proportion of target sequences with
four or more mismatches to the probe
within the intended binding region.}
\item{idFourOrMoreMismatchesPr}{Names of all sequences that matches with
four or more mismatches.}
\item{offTargetMatchPr}{Proportion of target sequences with maximum two
mismatches to at least one site outside the intended
probe binding region in the input alignment.}
\item{idOffTargetMatchPr}{Names of all off-target matching sequences.}
}
}
\description{
\code{checkMatch()} investigates how well oligos or assays match with their
intended target sequences
within a multiple DNA sequence alignment.
}
\details{
The output provides information on the proportion and names of target
sequences
that match perfectly as well as with one, two, three, or four or more
mismatches to the oligo within the intended oligo binding region
in the input alignment (on-target match).
It also gives the proportion and names of target sequences that
match with a maximum of two mismatches to at least one sequence variant of
the oligo outside the
intended oligo binding region (off-target match).
The function is a wrapper to \code{Biostrings::vcountPDict()}
(Pages et al., 2020).
}
\section{Methods (by class)}{
\itemize{
\item \code{RprimerOligo}: 

\item \code{RprimerAssay}: 
}}

\section{Limitations}{

There are a few limitations with this function, which is important to be
aware of:
\itemize{
\item False negatives or positives may occur due to
poorly aligned sequences
\item The output does not tell which strand
(minus or plus) the oligo matches to.
This is important to consider when
assessing off-target matches to single-stranded
targets
\item Ambiguous bases and gaps in the
target sequences are identified as mismatches
\item The function checks strictly on- and off-target, and may therefore
miss off-target matches that partially overlap the intended target
}
}

\examples{
#### RprimerOligo objects

data("exampleRprimerOligo")
data("exampleRprimerAlignment")

x <- exampleRprimerOligo[1:2, ]
target <- exampleRprimerAlignment

checkMatch(x, target)

#### RprimerAssay objects

data("exampleRprimerAssay")
data("exampleRprimerAlignment")

x <- exampleRprimerAssay[1:2, ]
target <- exampleRprimerAlignment

checkMatch(x, target)
}
\references{
Pages, H., Aboyoun, P., Gentleman R., and DebRoy S. (2020). Biostrings:
Efficient manipulation of biological strings. R package version
2.57.2.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/AllClasses.R
\docType{class}
\name{Rprimer-classes}
\alias{Rprimer-classes}
\alias{RprimerProfile-class}
\alias{.RprimerProfile}
\alias{RprimerProfile}
\alias{RprimerOligo-class}
\alias{.RprimerOligo}
\alias{RprimerOligo}
\alias{RprimerAssay-class}
\alias{.RprimerAssay}
\alias{RprimerAssay}
\alias{RprimerMatchOligo-class}
\alias{.RprimerMatchOligo}
\alias{RprimerMatchOligo}
\alias{RprimerMatchAssay-class}
\alias{.RprimerMatchAssay}
\alias{RprimerMatchAssay}
\alias{coerce}
\title{S4 classes for representation of different Rprimer objects}
\usage{
RprimerProfile(...)

RprimerOligo(...)

RprimerAssay(...)

RprimerMatchOligo(...)

RprimerMatchAssay(...)
}
\arguments{
\item{...}{A data frame or list to be converted into an Rprimer-object.}
}
\value{
An Rprimer-object if validation succeeds, an error
message otherwise.
}
\description{
The rprimer package contains five different S4 classes. Each class is used
as input or output for the different functions within the oligo and assay
design workflow:

 \itemize{
 \item \code{RprimerProfile}: output from \code{consensusProfile()},
 input for \code{oligos()}.
 \item \code{RprimerOligo}: output from \code{oligos()}, input
 for \code{assays()} and \code{checkMatch()}.
 \item \code{RprimerAssay}: output from \code{assays()}, input
 for \code{checkMatch()}.
 \item \code{RprimerMatchOligo}: output from \code{checkMatch()}.
 \item \code{RprimerMatchAssay}: output from \code{checkMatch()}.
}

These classes extends the DFrame class from S4vectors
(Pages et al., 2020), without any additional slots, but with some
additional checks for validity.
}
\section{Coercion}{

Each class can be converted to a traditional data frame, by using either
\code{as()} or \code{as.data.frame()}.

Moreover, \code{as()} can also be used for converting oligo sequences
within an
\code{RprimerOligo} or \code{RprimerAssay} object into a
\code{Biostrings::DNAStringSet}
object (Pages et al., 2020).
Note that all oligo sequences will be written in the
same direction as the input alignment that was used to generate
the oligos.
}

\examples{
## Constructors

data("exampleRprimerProfile")
x <- as.data.frame(exampleRprimerProfile)
RprimerProfile(x)

data("exampleRprimerOligo")
x <- as.data.frame(exampleRprimerOligo)
RprimerOligo(x)

data("exampleRprimerAssay")
x <- as.data.frame(exampleRprimerAssay)
RprimerAssay(x)

data("exampleRprimerMatchOligo")
x <- as.data.frame(exampleRprimerMatchOligo)
RprimerMatchOligo(x)

data("exampleRprimerMatchAssay")
x <- as.data.frame(exampleRprimerMatchAssay)
RprimerMatchAssay(x)

## Coercion methods for RprimerOligo and RprimerAssay objects

## Convert an RprimerOligo object to a DNAStringSet
data("exampleRprimerOligo")

## Pick rows to convert
x <- exampleRprimerOligo[1:2, ]
as(x, "DNAStringSet")

## Convert an RprimerAssay object to a DNAStringSet
data("exampleRprimerAssay")

## Pick rows to convert
x <- exampleRprimerAssay[1:2, ]
as(x, "DNAStringSet")
}
\references{
Pages, H., Lawrence, M., and Aboyoun, R. (2020). S4Vectors:
Foundation of vector-like and list-like containers in
Bioconductor. R package version 0.28.0.

Pages, H., Aboyoun, P., Gentleman R., and DebRoy S. (2020). Biostrings:
Efficient manipulation of biological strings. R package version
2.57.2.
}
\seealso{
consensusProfile, oligos, assays, checkMatch
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/runRprimerApp.R
\name{runRprimerApp}
\alias{runRprimerApp}
\title{rprimer Shiny application}
\usage{
runRprimerApp()
}
\value{
Opens the Shiny application.
}
\description{
\code{runRprimerApp()} starts a Shiny application where
the workflow of the rprimer package can be run through a
graphical user interface.
}
\examples{
## Only run this in interactive R sessions:
if (interactive()) {
    runRprimerApp()
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/consensusProfile.R
\name{consensusProfile}
\alias{consensusProfile}
\title{Get sequence information from an alignment}
\usage{
consensusProfile(x, ambiguityThreshold = 0)
}
\arguments{
\item{x}{A \code{MultipleAlignment::DNAMultipleAlignment} object.}

\item{ambiguityThreshold}{"Detection level" for ambiguous bases.
All DNA bases that occur with a relative frequency higher than the
specified value will be included when the IUPAC consensus character
is determined.
Can range from 0 to 0.2, defaults to \code{0}.}
}
\value{
An \code{RprimerProfile} object, which contains the following information:

\describe{
  \item{position}{Position in the alignment.}
  \item{a}{Proportion of A.}
  \item{c}{Proportion of C.}
  \item{g}{Proportion of G.}
  \item{t}{Proportion of T.}
  \item{other}{Proportion of bases other than A, C, G, T.}
  \item{gaps}{Proportion of gaps (recognized as "-" in the alignment).}
  \item{majority}{Majority consensus sequence.
  Denotes the most frequently occurring nucleotide.
  If two or more bases occur with the same frequency,
  the consensus nucleotide will be randomly selected among these.}
  \item{identity}{Proportion of sequences, among all sequences with a
  DNA base (i.e., A, C, G or T), that has the majority consensus base.}
  \item{iupac}{The consensus sequence expressed in IUPAC format.
  The IUPAC consensus sequence only
  takes 'A', 'C', 'G', 'T' and '-' as input. Degenerate bases
  will be skipped. If a position only contains
  degenerate bases, the IUPAC consensus will be \code{NA} at that
  position.}
  \item{coverage}{Proportion of sequences in the target alignment,
  among all sequences with a DNA base, that are covered the IUPAC consensus
  character.
  The value will be 1 if there are no "remaining" DNA bases (and/or if
  \code{ambiguityThreshold = 0}).}
}
}
\description{
\code{consensusProfile()} takes a DNA multiple alignment as input and
returns all the data needed for subsequent primer and probe design process.
The function is a wrapper to
\code{Biostrings::consensusMatrix()} (Pages et al., 2020).
}
\examples{
data("exampleRprimerAlignment")
consensusProfile(exampleRprimerAlignment)

consensusProfile(exampleRprimerAlignment, ambiguityThreshold = 0.05)
}
\references{
Pages, H., Aboyoun, P., Gentleman R., and DebRoy S. (2020). Biostrings:
Efficient manipulation of biological strings. R package version
2.57.2.
}
